#!/usr/bin/python
import os
import sys
import struct
import time

import pymedia
import speex
import horde_api
import asyncio

#
# global constants
#
HORDE_SESSION_NAME = "audio"
HORDE_STREAM_NAME = "audio_stream"

AUDIO_SAMPLE_RATE = 11025 # sampling for sent audio
AUDIO_CHANNELS = 1        # number of channels in sent audio

SPEEX_QUALITY_LEVEL = 5  # encoding quality: 0 is worst, 8 is best

SAMPLES_PER_ADU = 500

PLAYBACK_TICK_PERIOD = ((1.0 / AUDIO_SAMPLE_RATE) / AUDIO_CHANNELS) * SAMPLES_PER_ADU * .9

#
# parse command line
#
if len(sys.argv) != 3:
    print "usage: ", sys.argv[0],"<filename> <destination host> 2> debug.log"
    sys.exit(2)

in_filename = sys.argv[1]
dest_host = sys.argv[2]

#
# initialize speex, pymedia and open input file
#
audio_file = open( in_filename, 'rb' )

ext = os.path.splitext(in_filename)[1]
if ext: ext = ext[1:] # skip period in extension
mp3_demux = pymedia.muxer.Demuxer( ext )

frames = mp3_demux.parse( audio_file.read() )
decoder = pymedia.audio.acodec.Decoder( mp3_demux.streams[ 0 ] )

speex_encoder = speex.new(SPEEX_QUALITY_LEVEL)

#
# create an audio resampler
#
decoded = decoder.decode( frames[0][1] )
resampler = pymedia.audio.sound.Resampler((decoded.sample_rate, decoded.channels),
                                          (AUDIO_SAMPLE_RATE, AUDIO_CHANNELS))

per_sample_sleep_time = (1.0 / AUDIO_SAMPLE_RATE) / AUDIO_CHANNELS

#
# miscellaneous
#
start_time = time.time()

def debug_timestamp():
    t = time.time() - start_time
    t *= 1000
    return "%8d\t" % int(t)

#
# horde-related globals and helper functions
#
hstream = 0
last_rtt = -1
send_rate = 0.0

def session_cb(x):
    if (x == horde_api.EVENT_SESSION_CLOSED):
        print "session closed. exiting.\n"
        horde_api.session_close(HORDE_SESSION_NAME)
        sys.exit(0)

def event_cb(event, stream_handle, a, b, c, d):
    global last_rtt
    if (event == horde_api.EVENT_ADU_FRAG_ACK):
        sys.stderr.write(debug_timestamp() + "ACK\t%d ms\n" % d)
        if (last_rtt < 32000): last_rtt = d    

def recv_cb(stream_handle, adu_id, totlen, offset, data):
    # simplex stream: should never receive data
    return

def rate_to_kbps(rate):
    kbps = (rate * 8.0) / 1024;
    return "%.1f kbps" % kbps

def stat_cb(total_send, total_recv, total_allowed_send, nchannels):
    global send_rate
    send_rate = (total_send * 8.0) / 1024;
    sys.stderr.write("statistics: send=%d (%s), allowed=%d (%s), nchannels=%d\n"\
          % (total_send, send_rate, total_allowed_send, rate_to_kbps(total_allowed_send), nchannels))

def horde_event_loop(time):
    while (horde_api.event_dispatch(time) > 0):
        pass
#
# sending data
#
adu_id = 0
current_frame = 0
current_frame_offset = 0
decoded_frames = []

def get_next_adu():
    """constructs the next adu using decoded_frames and last_rtt
       and returns (adu_id, adu_data), or (None, None)"""
    global current_frame, current_frame_offset, decoded_frames
    global hstream, adu_id, last_rtt

    n = len(decoded_frames)
    if (current_frame >= n): return (None,None)

    # get rtt
    rtt_short = struct.pack("h", last_rtt)
    last_rtt = -1 # only send rtt once

    # get input samples for adu payload
    read_size = SAMPLES_PER_ADU * 2
    data = decoded_frames[current_frame][current_frame_offset:current_frame_offset + read_size]
    if len(data) < read_size:
        current_frame += 1
        current_frame_offset = 0
        read_size = read_size - len(data)
        if current_frame < n:
            data += decoded_frames[current_frame][current_frame_offset:current_frame_offset + read_size]

    current_frame_offset += read_size
    if current_frame < n and current_frame_offset > len(decoded_frames[current_frame]):
        current_frame += 1
        current_frame_offset = 0

    # encode adu payload
    encoded_data = speex_encoder.encode(data, True)

    # return adu
    adu = rtt_short + encoded_data
    adu_id += 1

    return (adu_id, adu)
    
#
# I/O event handlers
# 
def read_avail_horde_ippc(fd, time_now):
    """
    Called when one or more IPPC messages are available
    on horde's IPPC socket.
    """
    n = 1
    while (n > 0):
        n = horde_api.event_dispatch(0)
        
    if n < 0: # error (daemon probably died)
        print "\nhorde IPPC read error: exiting program."
        exit(1)

def timer_playback_tick(time_now):
    """
    Called when it is time to encode and send another frame.
    """
    adu_id, adu = get_next_adu()

    if adu_id:
        horde_api.stream_send(hstream, adu_id, adu, fragment_allowed=0)
        sys.stderr.write(debug_timestamp() + "SENT\t%d\t%d\n" % (adu_id, len(adu)))
        
        # status
        pct = int(current_frame * 100.0 / len(decoded_frames))
        print "\r\t%3d%% completed [frame %d of %d]" % (pct, current_frame, len(decoded_frames)),
        sys.stdout.flush()        
    else:
        io_loop.stop() # no more adu's
                
#
# main program loop
#
if (horde_api.session_connect(dest_host, HORDE_SESSION_NAME)):
        print "connected to server '%s'." % dest_host
        h = horde_api.stream_connect(HORDE_STREAM_NAME)
        if (h > 0):
            hstream = h
            print "\tstream '%s' connected, uid:%d." % (HORDE_STREAM_NAME, h)
            
            # register callbacks
            horde_api.session_set_callback(session_cb)
            horde_api.stream_set_event_callback(h, event_cb)
            horde_api.stream_set_recv_callback(h, recv_cb)
            horde_api.statistics_set_callback(stat_cb)
            horde_api.statistics_enable(1)

            # decode entire file
            print "Decoding and resampling file..."
            decoded_frames = []
            n = len(frames)
            for i in range(n):
                # decode a frame
                frame = frames[i]
                decoded = decoder.decode(frame[1])

                # resample
                pcm_audio = resampler.resample(decoded.data)

                # save for later
                decoded_frames.append(pcm_audio)

                # status
                pct = int(i * 100.0 / n)
                print "\r\t%3d%% completed" % (pct),
                sys.stdout.flush()

            print "\r\t%3d%% completed" % (100)

            # encode and send data
            print "Encoding and sending..."
            timestamp = time.time()
            current_frame = 0

            io_loop = asyncio.SelectLoop()
            fd = horde_api.get_ippc_fd()
            
            io_loop.want_read(fd, read_avail_horde_ippc)
            io_loop.add_timeout(PLAYBACK_TICK_PERIOD, timer_playback_tick, run_once=False)
            io_loop.run()

            print
            print "Done. Total send time: %.1f sec" % (time.time() - timestamp)
            
            horde_api.session_close(HORDE_SESSION_NAME)
