/**
 * Program to run horde bandwidth tests. 
 *
 * Client connects to server and uploads ADUs. 
 * 
 * Client produces a statistics log on standard error. The log contains info
 * on overall statistics as the first group of columns, followed by
 * info for each channel active at that point. Each group of columns
 * consists of: 
 * 'group_name recv_rate send_rate allowed_send_rate loss_rate smoothed_rtt'
 *
 * group_name is TOTAL for the first column, and the channel identifier for
 * every other group.
 *
 * Also an example of the horde API.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <iostream>
#include <iostream>
#include <limits>
#include <fstream>
#include <set>
#include <cmath>
#include <ctime>
#include <cassert>
#include <string>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>


#include <string>
using namespace std;

#include "horde.h"

// 
// globals
//

#define DEFAULT_SESSION_ID "bwtest";
#define ADU_SIZE 2048

char *sessionid = DEFAULT_SESSION_ID;
char *streamid = "zeros";
bool horde_connected = false;
bool is_server = false;
int hstream = -1;
int total_lost_bytes = 0;
int total_sent_bytes = 0;


statistics_report last_report;

//
// utility functions
//

int horde_event_loop(int timeout=-1) {
  int k = 0;
  while ((k = horde_event_dispatch(timeout)) > 0) ;
  return k;
}

string fmtfloat(float num, const char* fmt="%f", int max_size=32) {
  char buf[max_size];
  snprintf(buf, max_size, fmt, num);
  return string(buf);
}

string fmtint(int num, const char* fmt="%d", int max_size=32) {
  char buf[max_size];
  snprintf(buf, max_size, fmt, num);
  return string(buf);
}

string to_kbps(int bytes_per_sec) {
  float kbps = (bytes_per_sec * 8.0) / 1024;
  return fmtfloat(kbps, "%.1f");
}

string timestr() {
  timeval now;
  gettimeofday(&now, NULL);
  char buf[32];
  snprintf(buf,32,"%ld.%06ld", (unsigned long)now.tv_sec, (unsigned long)now.tv_usec);
  return string(buf);
}

//
// signal handler stuff
//

static void signal_handler(int s) {
  cout << "\nInterrupted by signal: " << s << endl;
  if (horde_connected) {
    cout << "\tClosing horde session.\n";

    horde_stream_close(streamid, hstream);
    horde_session_close(sessionid);    
  }
  exit(EXIT_SUCCESS);
}

void setup_signal() {
  //
  // setup signal handlers
  //
  signal(SIGINT, signal_handler);
  signal(SIGTERM, signal_handler);
  signal(SIGHUP, signal_handler);
  signal(SIGSEGV, signal_handler);
  
  siginterrupt(SIGINT, 1);
  siginterrupt(SIGTERM, 1);
  siginterrupt(SIGHUP, 1);
  siginterrupt(SIGSEGV, 1);
}

//
// callbacks for horde
//

void stream_event_cb(horde_stream_event event_code, int stream_uid, int adu_id, int offset, int len, int unused, void* unusedptr) {
  if (event_code == ADU_FRAG_LOSS) {
    total_lost_bytes += len;       
  }
}

void statistics_cb (statistics_report* report) {
  string out;

  out += timestr() + "\t";

  // overall stats
  out += "ALL\t" + to_kbps(report->total_recv_rate) + "\t" + to_kbps(report->total_send_rate) + "\t" +
    to_kbps(report->total_allowed_send_rate) + "\t" + fmtint(report->loss_rate) + "\t" + fmtint(report->ave_rtt);

  out += "\t";

  // for each channel
  for (int i = 0; i < last_report.nchannels; i++) {               
    statistics_channel *creport = &(last_report.channel_stats[i]);

    out += string(creport->channel_id) + "\t" + to_kbps(creport->recv_rate) + "\t" + to_kbps(creport->send_rate) + "\t" +
      to_kbps(creport->allowed_send_rate) + "\t" + fmtint(creport->loss_rate) + "\t" + fmtint(creport->ave_rtt);

    if (i < last_report.nchannels-1) out += "\t";
  }

  // output to stderr
  cerr << out << endl;
  
  // remember this report in case someone else wants it
  last_report = *report;
}

//
// server
//
void run_server() {
  cout << "Waiting for client..." << flush;
  if (horde_session_accept(sessionid)) {
    cout << "session up..." << flush;
    hstream = horde_stream_accept(streamid);
    if (hstream > 0) {
      horde_connected = true;
      cout << "stream connected [" << hstream  << "]\n" << flush;

      // Enable statistics reporting
      horde_statistics_set_callback(statistics_cb);
      horde_statistics_enable(TRUE);

      // don't want any stream events
      horde_stream_set_options(hstream, STREAM_OPT_NONE);

      // just sit and wait until horde exits.
      horde_event_loop();

      // done
      horde_stream_close(streamid, hstream);
    }
    
    horde_session_close(sessionid);  
  }
}

//
// client
//
void update_stat_display() { 
  string stats;

  // collect from each channel
  string sendstr;
  string recvstr;
  string lossstr;
  string rttstr;
  string allowedstr;
  for (int i = 0; i < last_report.nchannels; i++) {         
    sendstr += last_report.channel_stats[i].channel_id;
    sendstr += ":";
    sendstr += to_kbps(last_report.channel_stats[i].send_rate) +"k";

    //recvstr += last_report.channel_stats[i].channel_id;
    //recvstr += ":";
    //recvstr += to_kbps(last_report.channel_stats[i].recv_rate) +"k";

    //allowedstr += last_report.channel_stats[i].channel_id;
    //allowedstr += ":";
    //allowedstr += to_kbps(last_report.channel_stats[i].allowed_send_rate) +"k";

    lossstr += last_report.channel_stats[i].channel_id;
    lossstr += ":";
    lossstr += fmtint(last_report.channel_stats[i].loss_rate) + "%";

    rttstr += last_report.channel_stats[i].channel_id;
    rttstr += ":";
    rttstr += fmtint(last_report.channel_stats[i].ave_rtt);

    if (i < last_report.nchannels-1) { 
      sendstr += ", ";
      recvstr += ", ";
      lossstr += ", ";
      rttstr += ", ";
    }
  }

  // calculate overall cumulative loss rate
  float overall_loss = (100.0 * total_lost_bytes) / (total_sent_bytes == 0 ? 1 : total_sent_bytes);

  // put the output string together...
  stats += "send: ";
  stats += to_kbps(last_report.total_send_rate);
  stats += " [" + sendstr + "]";
  stats += ", ";

  stats += "loss: ";
  stats += fmtfloat(overall_loss, "%.1f%%");
  stats += "/" + fmtint(last_report.loss_rate) + "%";
  stats += " [" + lossstr + "]";
  stats += ", ";

  stats += "rtt: ";
  stats += fmtint(last_report.ave_rtt) + " ms";
  stats += " [" + rttstr + "]";
  stats += ", ";

  stats += "allowed: ";
  stats += to_kbps(last_report.total_allowed_send_rate) +"k";
  //  stats += " [" + allowedstr + "]";
  stats += ", ";

  stats += "recv: ";
  stats += to_kbps(last_report.total_recv_rate)+"k";
  //  stats += " [" + recvstr + "]";

  stats += "    ";

  // output
  cout << "\r" << stats << flush;
}

void run_client(char* server) {
  cout << "Contacting server..."  << flush;
    if (horde_session_connect(server, sessionid)) {
    cout << "session up..."  << flush;
    hstream = horde_stream_connect(streamid);
    if (hstream > 0) {
      horde_connected = true;
      cout << "stream connected [" << hstream  << "]\n"  << flush;

      // Enable statistics reporting
      horde_statistics_set_callback(statistics_cb);
      horde_statistics_enable(TRUE);

      // want to count losses and want I/O back-pressure
      horde_stream_set_options(hstream, STREAM_OPT_EVENT_LOSS | STREAM_OPT_IO_PRESSURE);
      horde_stream_set_event_callback(hstream, stream_event_cb, NULL);


      // send loop
      horde_objective_tags tags;
      char *zeros = (char*)calloc(ADU_SIZE, 1);
      int adu_id = 0;
      
      for (;;) {
	horde_stream_send(hstream, adu_id, zeros, ADU_SIZE, tags);
	adu_id++;
	total_sent_bytes += ADU_SIZE;

	if (horde_event_loop(0) < 0) break;

	update_stat_display();
      }
      cout << endl;

      // done
      horde_stream_close(streamid, hstream);
    }
    
    horde_session_close(sessionid);  
  }
}

//
// main
//

int main(int argc, char **argv)
{ 
  if ((argc > 4) || (argc < 2)  // not too many or too few?
      || (!(string(argv[1]) == "-s" || string(argv[1]) == "-c")) // first arg must be -c or -s
      || ((string(argv[1]) == "-c") && (argc < 3))) { // client must have server id
    cerr << "Correct usage:" << endl
	 << "\tServer: " << argv[0] << " -s [optional: session/stream id]" << endl
	 << "\tClient: " << argv[0] << " -c <server host name> [optional: session/stream id]"  << endl
	 << "Standard out will have the statistics log." << endl;
    exit(EXIT_FAILURE);
  }

  setup_signal();

  if (string(argv[1]) == "-s") {
    if (argc > 2) sessionid = argv[2];
    is_server = true;
    
    run_server();

  } else if (string(argv[1]) == "-c") { 
    char *server = argv[2];
    if (argc > 3) sessionid = argv[3];
    is_server = false;

    run_client(server);
  }
}
