/**
 * The Python interface to the horde daemon. 
 *
 * Implemented as a set of wrappers around the C-language interface.
 */

#include <Python.h>

#include "horde.h"

//
// wrappers
//
static PyObject *py_session_callback = NULL;
static PyObject *py_exception_callback = NULL;
static PyObject *py_statistics_callback = NULL;
static bool error_flag = false;

static
void session_callback(horde_session_event event) {
  if (py_session_callback) {
    PyObject *arglist;
    PyObject *result;

    /* Time to call the callback */
    arglist = Py_BuildValue("(i)", (int)event);
    result = PyEval_CallObject(py_session_callback, arglist);
    Py_DECREF(arglist);

    // get rid of result
    if (result == NULL)
      error_flag = true;   // so that event dispatch raises an exception
    else
      Py_DECREF(result);
  }
}

static void exception_callback(horde_exception_code excode, const char* msg) {
  if (py_exception_callback) {

    PyObject *arglist;
    PyObject *result;
    
    /* Time to call the callback */
    arglist = Py_BuildValue("(i,s)", (int)excode, msg);
    result = PyEval_CallObject(py_exception_callback, arglist);
    Py_DECREF(arglist);
    
    // get rid of result
    if (result == NULL)
      error_flag = true;   // so that event dispatch raises an exception
    else
      Py_DECREF(result);
  }
}

static void statistics_callback(statistics_report* report) {
  if (py_statistics_callback) {    
    PyObject *result = NULL;
    PyObject *arglist;
    
    /* Time to call the callback */
    arglist = Py_BuildValue("(i, i, i, i)", report->total_send_rate, report->total_recv_rate,
			    report->total_allowed_send_rate, report->nchannels);
    result = PyEval_CallObject(py_statistics_callback, arglist);
    Py_DECREF(arglist);
        
    // get rid of result
    if (result == NULL)
      error_flag = true;   // so that event dispatch raises an exception
    else
      Py_DECREF(result);
  }
}

static
void stream_recv_callback(int stream_uid, int adu_id, int total_len, int frag_offset, 
			   int frag_length, const char* frag_data, 
			   void* ptr_context) {
    PyObject *arglist;
    PyObject *result;

    PyObject *cb = (PyObject*)ptr_context;

    //    frag_data[frag_length] = '\0'; // make null terminated for Python

    if (cb != NULL) {

      /* Time to call the callback */
      arglist = Py_BuildValue("(i,i,i,i,s#)", stream_uid, adu_id, total_len, frag_offset, frag_data, frag_length);
      result = PyEval_CallObject(cb, arglist);
      Py_DECREF(arglist);
      
      // get rid of result
      if (result == NULL)
	error_flag = true;   // so that event dispatch raises an exception
      else
	Py_DECREF(result);
    } else {
      printf("\tcb is NULL\n");
    }
}

static
void stream_event_callback(horde_stream_event event_code, int stream_uid, int a, int b, int c, int d,
			   void* ptr_context) {
    PyObject *arglist;
    PyObject *result;

    PyObject *cb = (PyObject*)ptr_context;

    if (cb != NULL) {
      //	printf("event callback is: %08x\n", (int)((void*)cb));

      /* Time to call the callback */
      arglist = Py_BuildValue("(i,i,i,i,i,i)", (int)event_code, stream_uid, a, b, c, d);
      result = PyEval_CallObject(cb, arglist);
      Py_DECREF(arglist);
      
      // get rid of result
      if (result == NULL)
	error_flag = true;
      else
	Py_DECREF(result);
    } else {
      printf("\tcb is NULL\n");
    }
}

static PyObject*
hapi_session_set_callback(PyObject *self, PyObject* args) {
    PyObject *result = NULL;
    PyObject *temp;

    if (PyArg_ParseTuple(args, "O", &temp)) {
        if (!PyCallable_Check(temp)) {
            PyErr_SetString(PyExc_TypeError, "parameter must be callable");
            return NULL;
        }
        Py_XINCREF(temp);              /* Add a reference to new callback */

	if (py_session_callback != NULL)
	  Py_XDECREF(py_session_callback);  /* Dispose of previous callback */

        py_session_callback = temp;    /* Remember new callback */
	horde_session_set_callback(session_callback);

        /* Boilerplate to return "None" */
        Py_INCREF(Py_None);
        result = Py_None;
    }
    return result;
}

static PyObject*
hapi_exception_set_callback(PyObject *self, PyObject* args) {
    PyObject *result = NULL;
    PyObject *temp;

    if (PyArg_ParseTuple(args, "O", &temp)) {
        if (!PyCallable_Check(temp)) {
            PyErr_SetString(PyExc_TypeError, "parameter must be callable");
            return NULL;
        }
        Py_XINCREF(temp);              /* Add a reference to new callback */

	if (py_exception_callback != NULL)
	  Py_XDECREF(py_exception_callback);  /* Dispose of previous callback */

        py_exception_callback = temp;    /* Remember new callback */
	horde_exception_set_callback(exception_callback);

        /* Boilerplate to return "None" */
        Py_INCREF(Py_None);
        result = Py_None;
    }
    return result;
}

static PyObject*
hapi_statistics_set_callback(PyObject *self, PyObject* args) {
    PyObject *result = NULL;
    PyObject *temp;

    if (PyArg_ParseTuple(args, "O", &temp)) {
        if (!PyCallable_Check(temp)) {
            PyErr_SetString(PyExc_TypeError, "parameter must be callable");
            return NULL;
        }
        Py_XINCREF(temp);              /* Add a reference to new callback */

	if (py_statistics_callback != NULL)
	  Py_XDECREF(py_statistics_callback);  /* Dispose of previous callback */

        py_statistics_callback = temp;    /* Remember new callback */
	horde_statistics_set_callback(statistics_callback);

        /* Boilerplate to return "None" */
        Py_INCREF(Py_None);
        result = Py_None;
    }
    return result;
}

static PyObject*
hapi_stream_set_recv_callback(PyObject *self, PyObject* args) {
    PyObject *result = NULL;
    PyObject *new_cb;
    int hstream;

    if (PyArg_ParseTuple(args, "iO", &hstream, &new_cb)) {
        if (!PyCallable_Check(new_cb)) {
            PyErr_SetString(PyExc_TypeError, "parameter must be callable");
            return NULL;
        }
        Py_XINCREF(new_cb);              /* Add a reference to new callback */
	
	PyObject* stream_cb = (PyObject*)horde_stream_get_recv_cb_context(hstream);

	//	printf("setting recv callback: %08x\n", (int)((void*)new_cb));

	if (stream_cb != NULL)
	  Py_XDECREF(stream_cb);  /* Dispose of previous callback */
        
	/* Remember new callback */
	horde_stream_set_recv_callback(hstream, stream_recv_callback, (void*)new_cb);

	//	printf("recv callback is: %08x\n", (int)((void*)horde_stream_get_event_cb_context(hstream)));

        /* Boilerplate to return "None" */
        Py_INCREF(Py_None);
        result = Py_None;
    }
    return result;
}

static PyObject*
hapi_stream_set_event_callback(PyObject *self, PyObject* args) {
    PyObject *result = NULL;
    PyObject *new_cb;
    int hstream;

    if (PyArg_ParseTuple(args, "iO", &hstream, &new_cb)) {
        if (!PyCallable_Check(new_cb)) {
            PyErr_SetString(PyExc_TypeError, "parameter must be callable");
            return NULL;
        }
        Py_XINCREF(new_cb);              /* Add a reference to new callback */
	
	PyObject* stream_cb = (PyObject*)horde_stream_get_event_cb_context(hstream);

	if (stream_cb != NULL)
	  Py_XDECREF(stream_cb);  /* Dispose of previous callback */

	//	printf("setting event callback: %08x\n", (int)((void*)new_cb));

	/* Remember new callback */
	horde_stream_set_event_callback(hstream, stream_event_callback, (void*)new_cb);

	//	printf("event callback is: %08x\n", (int)((void*)horde_stream_get_event_cb_context(hstream)));

        /* Boilerplate to return "None" */
        Py_INCREF(Py_None);
        result = Py_None;
    }
    return result;
}

static PyObject *
hapi_event_dispatch(PyObject *self, PyObject *args) {
  int timeout = -1;  
  int ok = PyArg_ParseTuple(args, "|i", &timeout); // no arguments
  
  if (!ok) {
    return NULL; // raise exception
  }
  
  int ret;
  ret = horde_event_dispatch(timeout);

  if (error_flag) {
    error_flag = false;
    return NULL;        // raise the exception
  } else {     
    return Py_BuildValue("i", ret);
  }
}

static PyObject *
hapi_session_accept(PyObject *self, PyObject *args) {
  const char* name;
    
  if (!PyArg_ParseTuple(args, "s", &name))
    return NULL;
  
  int ret;
  ret = horde_session_accept(name);
  
  return Py_BuildValue("i", ret);
}

static PyObject *
hapi_session_connect(PyObject *self, PyObject *args) {
  const char* address;
  const char* name;
    
  if (!PyArg_ParseTuple(args, "ss", &address, &name))
    return NULL;
  
  int ret;
  ret = horde_session_connect(address, name);
  
  return Py_BuildValue("i", ret);
}

static PyObject *
hapi_session_close(PyObject *self, PyObject *args) {
  const char* name;
    
  if (!PyArg_ParseTuple(args, "s", &name))
    return NULL;
  
  horde_session_close(name);
  
  return Py_BuildValue(""); // return None
}

static PyObject *
hapi_stream_close(PyObject *self, PyObject *args) {
  const char* name;
  int handle;
    
  if (!PyArg_ParseTuple(args, "si", &name, &handle))
    return NULL;
  
  horde_stream_close(name, handle);
  
  return Py_BuildValue(""); // return None
}

static PyObject *
hapi_stream_clear(PyObject *self, PyObject *args) {
  int handle;
    
  if (!PyArg_ParseTuple(args, "i", &handle))
    return NULL;
  
  horde_stream_clear(handle);
  
  return Py_BuildValue(""); // return None
}

static PyObject *
hapi_stream_connect(PyObject *self, PyObject *args) {
  const char* name;
    
  if (!PyArg_ParseTuple(args, "s", &name))
    return NULL;
  
  int retval = horde_stream_connect(name);
  
  return Py_BuildValue("i", retval); // return None
}

static PyObject *
hapi_stream_accept(PyObject *self, PyObject *args) {
  const char* name;
    
  if (!PyArg_ParseTuple(args, "s", &name))
    return NULL;
  
  int retval = horde_stream_accept(name);
  
  return Py_BuildValue("i", retval);
}

static PyObject * hapi_get_ippc_fd(PyObject *self, PyObject *args) {
  int retval = horde_get_ippc_fd();
  
  return Py_BuildValue("i", retval);
}

static PyObject*
hapi_stream_send(PyObject *self, PyObject *args, PyObject *keywds) {
  // parse arguments
  int hstream, adu_id, data_sz, 
    priority=0, fragment_allowed=1, corr_group=0, multiple_tx=0, 
    loss_thresh=100, rtt_threshold=0, rel_loss_threshold=0, rel_rtt_threshold=0;
  const char* data;

  static char *kwlist[] = {"stream_handle", "adu_id", "data", "priority", "fragment_allowed", "correlation_group", 
			   "multiple_tx", "loss_threshold", "rtt_threshold", "relative_loss_threshold",
			   "relative_rtt_threshold", NULL};
  
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "iis#|iiiiiiii", kwlist,  &hstream, &adu_id, &data, &data_sz,
				   &priority, &fragment_allowed, &corr_group, &multiple_tx,
				   &loss_thresh,  &rtt_threshold, &rel_loss_threshold, &rel_rtt_threshold))
    return NULL;

  // call horde function
  horde_stream_send(hstream, adu_id, data, (size_t)data_sz, 
		    horde_objective_tags(priority, fragment_allowed, corr_group, multiple_tx, loss_thresh, 
					 rtt_threshold, rel_loss_threshold, rel_rtt_threshold));

  // return None
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
hapi_stream_send_fragment(PyObject *self, PyObject *args, PyObject *keywds) {
  // parse arguments
  int hstream, adu_id, offset, totlen, data_sz, 
    priority=0, corr_group=0, loss_thresh=100, fragment_allowed=1, multiple_tx=0, 
    rtt_threshold=0, rel_loss_threshold=0, rel_rtt_threshold=0;
  const char* data;

  static char *kwlist[] = {"stream_handle", "adu_id", "offset", "total_len", "data",
			   "priority", "fragment_allowed", "correlation_group", 
			   "multiple_tx", "loss_threshold", "rtt_threshold", "relative_loss_threshold",
			   "relative_rtt_threshold", NULL};
  
  if (!PyArg_ParseTupleAndKeywords(args, keywds, "iiiis#|iiiiiiii", kwlist, 
				   &hstream, &adu_id, &offset, &totlen, &data, &data_sz, 
				   &priority, &fragment_allowed, &corr_group, &multiple_tx,
				   &loss_thresh,  &rtt_threshold, &rel_loss_threshold, &rel_rtt_threshold))
    return NULL;

  // call horde function
  horde_stream_send_fragment(hstream, adu_id, offset, totlen, data, (size_t)data_sz, 
		    horde_objective_tags(priority, fragment_allowed, corr_group, multiple_tx, 
					 loss_thresh, rtt_threshold, rel_loss_threshold, rel_rtt_threshold));

  // return None
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
hapi_set_log_level(PyObject *self, PyObject *args) {
  int level;
    
  if (!PyArg_ParseTuple(args, "i", &level))
    return NULL;

  horde_set_log_level(level);
  
  return Py_BuildValue("");
}

static PyObject *
hapi_statistics_enable(PyObject *self, PyObject *args) {
  int enabled=1;
    
  if (!PyArg_ParseTuple(args, "|i", &enabled))
    return NULL;

  horde_statistics_enable((enabled) ? TRUE: FALSE);
  
  return Py_BuildValue("");
}

static PyObject *
hapi_stream_set_options(PyObject *self, PyObject *args) {
  int hstream;
  int opts=STREAM_OPT_ALL;
    
  if (!PyArg_ParseTuple(args, "i|i", &hstream, &opts))
    return NULL;

  horde_stream_set_options(hstream, opts);
  
  return Py_BuildValue("");
}

static PyObject *
hapi_stream_accept_multiple(PyObject *self, PyObject *args) {

  int nstreams;
  PyObject *list_id, *list_handles;

  // parse arguments
  if (!PyArg_ParseTuple(args, "iOO", &nstreams, &list_id, &list_handles))
    return NULL;

  if (!(PyList_Check(list_id) && (PyList_Check(list_handles)))) {
    PyErr_SetString(PyExc_TypeError, "last two arguments must be lists.");
    return NULL;
  }

  int nid = PyList_Size(list_id);
  int hid = PyList_Size(list_handles);

  if ((nid != hid) || (nid != nstreams)) {
    PyErr_SetString(PyExc_TypeError, "list arguments must be same length as nstreams.");
    return NULL;
  }

  // extract stream id's
  const char* id[32];
  for (int i = 0; i < nstreams; i++) {
    PyObject* str = PyList_GetItem(list_id, i);
    if (PyString_Check(str)) {
      id[i] = PyString_AsString(str);
    } else {
      PyErr_SetString(PyExc_TypeError, "all stream id's must be strings.");
      return NULL;
    } 
  }  

  // call horde code
  int handles[32];
  int retval = horde_stream_accept_multiple(nstreams, id, handles);

  // fill in list_handles
  for (int i = 0; i < nstreams; i++) {
    PyObject* old = PyList_GetItem(list_handles, i);
    PyList_SetItem(list_handles, i, Py_BuildValue("i", handles[i]));
    Py_XDECREF(old);
  }

  // return
  return Py_BuildValue("i", retval);
}

static PyObject *
hapi_stream_connect_multiple(PyObject *self, PyObject *args) {

  int nstreams;
  PyObject *list_id, *list_handles;

  // parse arguments
  if (!PyArg_ParseTuple(args, "iOO", &nstreams, &list_id, &list_handles))
    return NULL;

  if (!(PyList_Check(list_id) && (PyList_Check(list_handles)))) {
    PyErr_SetString(PyExc_TypeError, "last two arguments must be lists.");
    return NULL;
  }

  int nid = PyList_Size(list_id);
  int hid = PyList_Size(list_handles);

  if ((nid != hid) || (nid != nstreams)) {
    PyErr_SetString(PyExc_TypeError, "list arguments must be same length as nstreams.");
    return NULL;
  }

  // extract stream id's
  if (nstreams > 32) nstreams = 32;
  const char* id[32];
  for (int i = 0; i < nstreams; i++) {
    PyObject* str = PyList_GetItem(list_id, i);
    if (PyString_Check(str)) {
      id[i] = PyString_AsString(str);
    } else {
      PyErr_SetString(PyExc_TypeError, "all stream id's must be strings.");
      return NULL;
    } 
  }  

  // call horde code
  int handles[32];
  int retval = horde_stream_connect_multiple(nstreams, id, handles);

  // fill in list_handles
  for (int i = 0; i < nstreams; i++) {
    PyList_SetItem(list_handles, i, Py_BuildValue("i", handles[i]));
    // todo: what about the ref count of the item we're overwriting...?
  }

  // return
  return Py_BuildValue("i", retval);
}

//
// other python code
//

static PyMethodDef HordeAPIMethods[] = { 
    {"event_dispatch",  hapi_event_dispatch, METH_VARARGS,
"Blocks, waits for an event for some maximum amount of time.\n"
"If an event occurs during that period, it calls the appropriate\n"
"callback, and returns immediately.\n\n"

"int event_dispatch(int timeout_ms=-1)\n\n"

"RETURNS: <0 on error; 0 if no events were dispatched;\n"
"         or 1 if an event was dispatched.\n"
"PARAM timeout_ms: the maximum amount of time to wait for an event,\n"
"                  in milliseconds.\n"
"EXCEPTIONS: if an event handler raises an exception, that exception\n"
"            may be raised by this method."},

    {"session_accept",  hapi_session_accept, METH_VARARGS, 
"Wait for and accept a session request from a remote host, if\n"
"the name presented by the remote host matches this name. This\n"
"method blocks until a connection request is received and resolved,\n" 
"or an error occurs (e.g., connection broken with horde backend).\n\n"

"boolean session_accept(string session_name)\n\n"

"RETURNS: False if the session couldn't be established due to an error,\n"
"         True otherwise."},

    {"session_connect",  hapi_session_connect, METH_VARARGS,    
"Attempts to connect to a remote host, using the given session name.\n"
"The remote host must be waiting to accept a session. This blocks until\n" 
"it is able to connect, or an error occurs.\n\n"  

"boolean session_connect(string connect_address, string session_name)\n\n"

"RETURNS: False if the session couldn't be established due to an error;\n"
"         True otherwise.\n"
"PARAM connect_address: remote host address (can be hostname, or IP in\n"
"                       dotted-quad notation)"},

    {"session_close",  hapi_session_close, METH_VARARGS,    
"Closes the named session, if it is the current session. Or does nothing.\n"
"This must be the process that owns the session to be able to close it.\n\n"
"void session_close(string session_name)"},

    {"session_set_callback",  hapi_session_set_callback, METH_VARARGS,    
"Sets the callback to be invoked on session events.\n\n"

"void session_set_callback(callback)\n\n"

"PARAM callback is of the form: func(int event_code), and event_code\n"
"      is one of: EVENT_SESSION_CLOSED."},

    {"exception_set_callback",  hapi_exception_set_callback, METH_VARARGS,    
"Sets the callback to be invoked on exception events.\n\n" 
"void exception_set_callback(callback)\n\n" 
"PARAM callback is of the form: func(int exception_code, string message),\n"
"      and exception_code is one of the EXCEPTION_ values."}, 

    {"statistics_set_callback",  hapi_statistics_set_callback, METH_VARARGS,    
"Sets the callback to be invoked on statistics events.\n\n" 
"void statistics_set_callback(callback)\n\n" 
"PARAM callback is of the form:\n"
"        func(total_send_rate, total_recv_rate,\n"
"             total_allowed_send_rate, nchannels)\n"
"      and all rates are in bytes/second."}, 

    {"statistics_enable",  hapi_statistics_enable, METH_VARARGS,    
"Enables or disable statistics. If enabled, statistics callback is\n"
"invoked periodically.\n\n"
"void statistics_enable(int enabled=1)"},

    {"stream_close",  hapi_stream_close, METH_VARARGS,
"Closes the named stream, if given the correct handle. Or does nothing.\n\n"
"void stream_close(string stream_name, int stream_handle)"}, 

    {"stream_accept",  hapi_stream_accept, METH_VARARGS,
"Waits for and handles a stream connect request. Blocks until a stream\n"
"with the given identifier is accepted or an error occurs.\n\n" 

"int stream_accept(string stream_name)\n\n" 

"RETURNS: a handle for the open stream >0; or <0, and sets errno."}, 

    {"stream_connect",  hapi_stream_connect, METH_VARARGS,
"Sends a stream connect request. Blocks until the stream is connected\n"
"or an error occurs.\n\n"

"int stream_connect(string stream_name)\n\n"

"RETURNS: a handle to the open stream >0; or <0 and sets errno."},

    {"stream_send",  (PyCFunction)hapi_stream_send, METH_VARARGS|METH_KEYWORDS,
"Send an ADU.\n\n"

"void stream_send(stream_handle, int adu_id, string adu_data,\n"
"                 int priority=0, int fragment_allowed=1, \n"
"                 int correlation_group=0, int multiple_tx=0,\n"
"                 int loss_threshold=100, rtt_threshold=0,\n"
"                 int relative_loss_threshold=0,\n"
"                 int relative_rtt_threshold=0)\n\n"
"PARAM stream_handle: a valid stream handle.\n"
"PARAM adu_id: A unique 24-bit identifier for the ADU.\n\n"

"All arguments after adu_data are optional QoS objective tags. One or\n"
"more of these can be specified using keyword calling. These parameters\n"
"control how the data is scheduled, as described below.\n\n"

"PARAM priority: Lower numbers are higher priorities. Higher priority\n"
"         data is sent in preference to lower priority data.\n"
"PARAM correlation_group: Ignored if equal to zero. If non-zero, Horde\n"
"         will try to schedule two ADU's with the same correlation group\n"
"         in packets that will not experience correlated losses.\n"
"PARAM loss_threshold: Maximum allowed loss probability (as a percentage\n"
"         0-100). This data will not be sent in slots with a higher\n"
"         loss probablity. A low value may result in head-of-line\n"
"         blocking.\n"
"PARAM fragment_allowed: Whether or not this ADU can be fragmented.\n"
"         If this is zero, the ADU will never be fragmented. Be careful:\n"
"         if the ADU is larger than the maximum packet size, the stream\n"
"         may get blocked up, and nothing after this ADU will be sent.\n"
"         If this is >0, the ADU may be fragmented.\n"
"PARAM multiple_tx: If this is non-zero, the data may be transmitted\n"
"         multiple times by Horde, if the available bandwidth is\n"
"         more than the overall offered load.\n"
"PARAM rtt_threshold: If non-zero, no slots with an expected RTT\n"
"         greater than this will be used.\n"
"PARAM relative_loss_threshold: If non-zero, then slots with expected\n"
"         loss greater than the average loss (over all channels) plus\n"
"         this threshold will not be used.\n"
"PARAM relative_rtt_threshold: If non-zero, then slots with expected\n"
"         RTT's larger than the stream's mean RTT plus this\n"
"         threshold will not be used.\n"
"NOTE: May block. See STREAM_OPT_IO_PRESSURE notes for stream_set_options"},

    {"stream_send_fragment",  (PyCFunction)hapi_stream_send_fragment, METH_VARARGS|METH_KEYWORDS,    
"Send an ADU fragment. Useful for retransmitting lost data.\n\n"
"void stream_send_fragment\n"
"             (stream_handle, int adu_id, int offset, int total_length,\n"
"              string adu_data, int priority=0, ... )\n\n" 
"NOTE: All arguments after adu_data are optional QoS objective tags.\n"
"      One or more of these can be specified using keyword calling.\n"
"SEE ALSO: stream_send"},

    {"stream_clear",  hapi_stream_clear, METH_VARARGS,
"Clear's the given stream's send queue.\n\nvoid stream_clear(int stream_handle)"},

    {"stream_set_recv_callback",  hapi_stream_set_recv_callback, METH_VARARGS,     
"Sets the callback to be invoked on stream data-receive events.\n\n"
"void stream_set_recv_callback(stream_handle, callback)\n\n"
"PARAM stream_handle: a valid stream handle.\n"
"PARAM callback: a function of the form:\n"
"    func(stream_handle, int adu_id, int total_length,\n" 
"         int frag_start, string frag_data)"},

    {"stream_set_event_callback",  hapi_stream_set_event_callback, METH_VARARGS,
"Sets the callback to be invoked on all stream events other than data\n"
"receives.\n\n"

"void stream_set_event_callback(stream_handle, callback)\n\n"

"PARAM stream_handle: a valid stream handle.\n"
"PARAM callback: a function of the form:\n"
"                    func(event, stream_handle, int, int, int, int)\n"
"  when event is\n" 
"     EVENT_STREAM_CLOSED:\n"
"       func(event, stream_handle, 0, 0, 0, 0)\n"
"     EVENT_ADU_FRAG_ACK:\n"
"       func(event, stream_handle, adu_id, frag_start, frag_len, rtt_ms)\n" 
"     EVENT_ADU_FRAG_LOSS:\n"
"       func(event, stream_handle, adu_id, frag_start, frag_len, 0)"},

    {"stream_set_options",  hapi_stream_set_options, METH_VARARGS,    
"Sets stream options.\n\n"

"When a stream is created, by default all options are set. Most of the\n"
"stream options are event masks (when they are not set, that type of\n"
"event is not delivered to the stream event handler). See the STREAM_OPT_\n"
"constants. STREAM_OPT_IO_PRESSURE is the only complex option. When this\n"
"option is set (default), sending too much data will cause the IPPC\n"
"socket to block, to not accept any new IPPC sends, until the data\n"
"already queued inside the horde daemon has been sent on the network.\n"
"When STREAM_OPT_IO_PRESSURE is not set, the IPPC layer will accept\n"
"data as fast as it can. The horde daemon may silently drop packets\n"
"internally if the client is sending at a rate much higher than the\n"
"allowed rate.\n\n"

"void stream_set_options(int stream_handle, int opts=STREAM_OPT_ALL)\n\n"

"PARAM opts: A set of option flags. A set of STREAM_OPT_ values OR'd\n"
"            together. Any STREAM_OPT_ that has not been added to opts,\n"
"            will be disabled, and options set in opts will be enabled.\n" 
"            If you want to turn off a single option, start with\n"
"            STREAM_OPT_ALL and XOR out any options you don't want.\n"
"DEFAULTS: everything except STREAM_OPT_EVENT_FLUSHED" }, 

    {"set_log_level",  hapi_set_log_level, METH_VARARGS,
"Set the API's debug log's verbosity level. Messages are logged to\n"
"standard error. Setting the level to LOG_LEVEL_NONE turns off logging\n"
"completely; LOG_LEVEL_ERROR only displays error messages;\n"
"LOG_LEVEL_DEBUG turns on everything.\n\n"
"void set_log_level(int level)"},
    
    {"get_ippc_fd",  hapi_get_ippc_fd, METH_VARARGS,
"Get the file descriptor for the socket being used to communicate\n"
"with the horde IPPC daemon. Nothing should be written to or read from\n"
"this descriptor. This function is provided so that the descriptor can\n"
"be used inside a select loop outside the horde code. Whenever the\n"
"descriptor is ready to read from, call event_dispatch(0)\n\n"
"int get_ippc_fd()\n\n"
"RETURN: int, -1 or a valid file descriptor."},

    {"stream_connect_multiple", hapi_stream_connect_multiple, METH_VARARGS,
"Roughly equivalent to call horde_stream_connect N-times, but potentially\n"
"far more efficent.\n\n"
"int stream_connect_multiple(int nstreams, list stream_id, list handles)\n\n"
"PARAM nstreams: number of entries in stream_id and handles.\n"
"PARAM stream_id: a list of stream identifiers (all unique).\n"
"PARAM handles: list of size nstreams where the handle for the\n"
"          corresponding stream_id or a -1 will be filled in.\n"
"RETURNS: The number of streams successfully connected. This is the\n"
"          number of entries in handles > 0\n"
"NOTE: nstreams cannot be greater than 32."},

    {"stream_accept_multiple", hapi_stream_accept_multiple, METH_VARARGS,
"Roughly equivalent to call horde_stream_accept N-times, but potentially\n"
"far more efficent.\n\n"
"int stream_accept_multiple(int nstreams, list stream_id, list handles)\n\n"
"PARAM nstreams: number of entries in stream_id and handles.\n"
"PARAM stream_id: a list of stream identifiers (all unique).\n"
"PARAM handles: list of size nstreams where the handle for the\n"
"          corresponding stream_id or a -1 will be filled in.\n"
"RETURNS: The number of streams successfully connected. This is the\n"
"          number of entries in handles > 0\n" 
"NOTE: nstreams cannot be greater than 32."},
    //*/    
    {NULL, NULL, 0, NULL}        /* Sentinel */ 
}; 

PyMODINIT_FUNC
inithorde_api(void)
{
    PyObject* module = Py_InitModule("horde_api", HordeAPIMethods);

    // add integer constants
    PyModule_AddIntConstant(module, "EVENT_SESSION_CLOSED", (long)SESSION_CLOSED);
    PyModule_AddIntConstant(module, "EVENT_STREAM_CLOSED",  (long)STREAM_CLOSED);
    PyModule_AddIntConstant(module, "EVENT_STREAM_FLUSHED",  (long)STREAM_FLUSHED);
    PyModule_AddIntConstant(module, "EVENT_ADU_FRAG_ACK", (long)ADU_FRAG_ACK);
    PyModule_AddIntConstant(module, "EVENT_ADU_FRAG_LOSS", (long)ADU_FRAG_LOSS);

    PyModule_AddIntConstant(module, "EXCEPTION_BAD_IPPC", (long)EX_BAD_IPPC);
    PyModule_AddIntConstant(module, "EXCEPTION_ILLEGAL_OPERATION", (long)EX_ILLEGAL_OPERATION);
    PyModule_AddIntConstant(module, "EXCEPTION_INSUFFICIENT_PERMISSION", (long)EX_INSUFFICIENT_PERMISSION);
    PyModule_AddIntConstant(module, "EXCEPTION_BAD_SESSION_ID", (long)EX_BAD_SESSION_ID);
    PyModule_AddIntConstant(module, "EXCEPTION_BAD_STREAM_ID", (long)EX_BAD_STREAM_ID);
    PyModule_AddIntConstant(module, "EXCEPTION_INTERNAL", (long)EX_INTERNAL);


    PyModule_AddIntConstant(module, "STREAM_OPT_ALL", (long)STREAM_OPT_ALL);
    PyModule_AddIntConstant(module, "STREAM_OPT_NONE", (long)STREAM_OPT_NONE);
    PyModule_AddIntConstant(module, "STREAM_OPT_EVENT_RECV", (long)STREAM_OPT_EVENT_RECV);
    PyModule_AddIntConstant(module, "STREAM_OPT_EVENT_ACK", (long)STREAM_OPT_EVENT_ACK);
    PyModule_AddIntConstant(module, "STREAM_OPT_EVENT_LOSS", (long)STREAM_OPT_EVENT_LOSS);
    PyModule_AddIntConstant(module, "STREAM_OPT_EVENT_FLUSHED", (long)STREAM_OPT_EVENT_FLUSHED);
    PyModule_AddIntConstant(module, "STREAM_OPT_IO_PRESSURE", (long)STREAM_OPT_IO_PRESSURE);

    PyModule_AddIntConstant(module, "LOG_LEVEL_DEBUG", (long)3);
    PyModule_AddIntConstant(module, "LOG_LEVEL_INFO", (long)2);
    PyModule_AddIntConstant(module, "LOG_LEVEL_ERROR", (long)1);
    PyModule_AddIntConstant(module, "LOG_LEVEL_NONE", (long)0);
}
