// -*- mode: C++; -*-

#ifndef __HORDE_DAEMON_NET_H__
#define __HORDE_DAEMON_NET_H__

#include <nmstl/netioevent>

#include "string_hashmap.h"

#include "daemon.h"
#include "packet_io.h" // argh, better way?

namespace horde {
  class interface_pool_manager;
  class interface_iohandler;

  // hooks for daemon code to handle connect request/ack events at a level of
  // abstraction above the net wire protocol...
  void handle_recv_connect_request(inet_address peer_address, string session_name);
  void handle_recv_connect_ack(inet_address peer_address, string session_name);

  // requires: only called once.
  void create_poolman(vector<string>& managed_ifaces, uint16_t port, 
		    uint32_t period, io_event_loop& loop);
  
  interface_pool_manager* get_poolman();

  //
  // interface I/O manager
  //
  class interface_iohandler : public udp_iohandler {
  private:
    void handle_packet_recv(horde_net_packet& packet, inet_address& addr);
    
  public:
    interface_iohandler(io_event_loop& ioloop,
			inet_address bind_addr) : udp_iohandler(ioloop, bind_addr, true, false, false) 
      { }
    
    void send_control_packet(inet_address peer_address, string session_name, horde_packet_type type);
    
    void shutdown() {
      get_ioh().shutdown(true, true);
    }
  protected:
    void ravail();
    //void wavail(); assume, since UDP socket, that we can always write...
  };

  //
  // interface pool manager
  //
  class interface_pool_manager {
  private:
    // the scheduling period
    uint32_t period_msecs;
    
    // is this poolman running?
    bool running;
    
    // list of managed interfaces
    vector<string> ifaces;
    
    // the I/O loop to use
    io_event_loop* ioloop;

    // the port to use on every interface (TODO: is there a better way?)
    uint16_t port;
    
    // interface information snapshot
    STRING_HASHMAP(inet_address) iface_addresses;

    // interace index -> I/O handler map
    hash_map<int, ptr<interface_iohandler> > iohandler_map;
    
    // internal functions to handle pool management...
    
    // wakeup, examine state of interfaces, determine what needs to be done
    // and do it...
    void wakeup();
    
    // note that an interface came online, init the io handlers
    void iface_up (int idx, inet_address addr);
    
    // note that an interface went off, destroy the io handlers
    void iface_down(int idx);

    void send_control_packet(inet_address peer_address, string session_name, horde_packet_type type);
    
  public:
    // constructor
    interface_pool_manager(vector<string>& managed_ifaces, uint16_t port, 
			   uint32_t period, io_event_loop& loop) 
      : period_msecs(period), running(false), ioloop(&loop),
      ifaces(managed_ifaces), port(port), iface_addresses()
      { }
      
      bool is_running() { return running; }
      
      // starts pool manager, unless is_running().
      void start();  
      
      // stops, and deactivates all channel managers and io handlers
      void stop();
      
      // Function to send a connect request.
      // - the connect request can be dropped, so this function might need to be
      //   multiple times, until a connect_ack is received.
      void send_connect_request(inet_address peer_address, string session_name);
      void send_connect_nack(inet_address peer_address, string session_name);
  };

};

#endif
