// -*- mode: C++; -*-

/** @file scheduler_impl.h
 *
 * Packet scheduler implementations.
 *
 * The \a create_scheduler() factory accepts the following arguments:
 * "firehose", "roundrobin", "hoard", or "default".  "default" is the
 * same as "roundrobin". Depending on the \a type_id string, the 
 * factory uses one of the three scheduler classes defined here.
 */
#ifndef __HORDE_SESSION_SCHEDULER_IMPL_H__
#define __HORDE_SESSION_SCHEDULER_IMPL_H__

#include "scheduler.h"

#include "session/stream.h"

#include "packet_io.h"
#include "net/txslot.h"
#include "net/manager.h"
#include "horde_log.h"

#include <map>

#include "coder.h"

HORDE_NAMESPACE_BEGIN;

/** 
 * Consumes every TX slot, filling it with garbage. Useful for
 * testing/debugging the congestion control layer.
 */
class scheduler_firehose : public scheduler {
 public:
  scheduler_firehose() {
    LOG_EVENT("schedule") << "created 'firehose' scheduler.";
  }

  void schedule_tx_slot(ptr<tx_slot> slot);
  vector<data_fragment_info> get_packet_contents(int packet_uid);
};

/** 
 * A partial scheduler implementation with useful utility functions
 * to help implement working schedulers.
 *
 * Provides helper functions to aid packing packets with data, and 
 * functions to help deal with packet UID's. See the subclasses to 
 * see how these functions are used.
 */
class skeletal_scheduler_impl : public scheduler {
 private:
  map<int, vector<data_fragment_info> > send_history;

  horde_net_packet current;
  vector<data_fragment_info> current_contents;
  size_t current_used, current_size, current_overhead;
  bool reuse_current;

  unsigned int peer_address_round; // useful for 1-to-many data sends
  bool peer_address_round_robin; // whether to change the address round on every tx slot consumption  
  
 protected:
  int current_uid; // current packet uid

  /// turn off peer address round-robin cycling in consume_tx_slot
  void disable_address_round_robin() { peer_address_round_robin=false; }
  void enable_address_round_robin() { peer_address_round_robin=true; }

  /// Start working on a new packet that will fit in the given slot, and return the UID. 
  int packet_new(ptr<tx_slot> slot);  

  /// Start working on a new packet but restrict to given size, and return the UID. 
  int packet_new(ptr<tx_slot> slot, size_t size);  

  /// get a buffer to the current packet's contents
  constbuf packet_get_data();

  /// don't discard the current packet, try to reuse it in the next round...
  void packet_reset() { 
    reuse_current = true;
  }

  /** Read some data into the current packet, and return true if the packet is full. 
   *  @pre assumes stream's sendq is not empty. */  
  bool packet_read_into(stream* stream);

  /**
   * Reads an entire data buffer into the packet, if there is enough space, and
   * returns true if the packet is full. Gets stream info (UID, remote UID) from
   * the pointer passed in.
   */
  bool packet_read_into(int adu_id, dynbuf& data, stream* stream);

  bool packet_write_int(int x); // HACK for coder

  // returns the percent fraction of the packet that consists of horde control structures...
  int packet_get_overhead() { return ((current_overhead * 100) / current_used); }
  int packet_get_num_fragments() { return current_contents.size(); }
  size_t packet_get_used() { return current_used; }
  size_t packet_get_unused() { return (current_size - current_used); }
  bool packet_is_full(); ///< Returns true iff the current packet is full.
  bool packet_has_data(); ///< Returns true iff there is some send-able data in the packet.
  
  /// Send the current packet and remember the contents
  void packet_send(ptr<tx_slot> slot);

  skeletal_scheduler_impl() : current_uid(-1), send_history(), 
			      current_contents(), current(), current_used(0), current_size(0),
			      reuse_current(false), peer_address_round(0),
			      peer_address_round_robin(false)
  { }

 public:
  vector<data_fragment_info> get_packet_contents(int packet_uid);
};

/* Assigns TX slots by a round-robin servicing of the streams. */
class scheduler_roundrobin : public skeletal_scheduler_impl {
 protected:
  int round;

 public:
  scheduler_roundrobin() : round(0), skeletal_scheduler_impl() {
    LOG_EVENT("schedule") << "created 'roundrobin' scheduler.";
  }
  void schedule_tx_slot(ptr<tx_slot> slot);
};

/* Hoard. Assigns TX slots based on priority/correlation-group/loss-threshold ADU tags. */
class scheduler_hoard : public skeletal_scheduler_impl {
 protected:
  int round;

  struct mtx_set_elem_t {
    dynbuf data;
    int adu_id;
    stream* pstream;

    mtx_set_elem_t() : data(), adu_id(0), pstream() {} 
    mtx_set_elem_t(dynbuf& x, int id, stream* ps) : data(x), adu_id(id), pstream(ps) { }
  };

  int mtx_set_seq;
  map<int,mtx_set_elem_t> mtx_set;         ///< set of recent ADU's that have been tagged multi TX
  static const int MTX_SET_MAX_SIZE = 16;

  void mtx_set_remove_worst();
  void mtx_set_add(dynbuf& data, int adu, stream* pstream);
  bool mtx_set_find_best_match(int length, dynbuf& buf, int& adu_id, stream* pstream);

 public:
  scheduler_hoard() : skeletal_scheduler_impl(), round(0), mtx_set_seq(0), mtx_set() {
    LOG_EVENT("schedule") << "created 'hoard' scheduler.";
  }
  void schedule_tx_slot(ptr<tx_slot> slot);
};


/* Cleaner rewrite of the hoard algorithm, plus support for rtt_threshold objective. */
class scheduler_grunt : public skeletal_scheduler_impl {
 protected:
  int round;

  // packet related helpers and state
  set<int> groups_in_packet;
  map<int,int> stream_srtt;
  
  int grunt_packet_new(ptr<tx_slot> slot);
  bool grunt_packet_read_into(ptr<tx_slot> slot, stream* src, horde_objective_tags& src_tags);
  bool grunt_packet_test_constraints(ptr<tx_slot> slot, stream* src, int data_len, int offset, horde_objective_tags& tags);
  void grunt_packet_fill_unused_space();

  // multiple TX related stuff
  struct mtx_set_elem_t {
    dynbuf data;
    int adu_id;
    stream* pstream;

    mtx_set_elem_t() : data(), adu_id(0), pstream() {} 
    mtx_set_elem_t(dynbuf& x, int id, stream* ps) : data(x), adu_id(id), pstream(ps) { }
  };

  int mtx_set_seq;
  int mtx_set_max_size;
  map<int,mtx_set_elem_t> mtx_set;         ///< set of recent ADU's that have been tagged multi TX
  static const int MTX_SET_MAX_SIZE = 16;
  bool mtx_streams; ///< are there any streams with multiple_tx < 0 at their head?

  void mtx_set_remove_worst();
  void mtx_set_add(dynbuf& data, int adu, stream* pstream);
  bool mtx_set_find_best_match(int length, dynbuf& buf, int& adu_id, stream*& pstream);  

  bool mtx_streams_read_into(int maxsize);
 public:
  scheduler_grunt();

  void schedule_tx_slot(ptr<tx_slot> slot);
};

/* Round-robin scheduler with (3+1)XOR packet coding. Only works with 1024-byte packets. Assumes packet-size = 1024.*/
class scheduler_coder : public skeletal_scheduler_impl {
 protected:
  int round;
  BlockCoder coder;
  horde_net_packet *coded;   

  void _coded_create(pair<unsigned char*, size_t> p);
  void _coded_send(ptr<tx_slot> slot);

 public:
  scheduler_coder();
  void schedule_tx_slot(ptr<tx_slot> slot);
};

HORDE_NAMESPACE_END;

#endif
