// -*- mode: C++; -*-

/** @file channel_sim.h
 *
 * Channel simulation layer. Provides a subclass of \a channel_iohandler 
 * that simulates network latency and losses. 
 */

#ifndef __HORDE_CHANNEL_SIMULATION_H__
#define __HORDE_CHANNEL_SIMULATION_H__

//
// channel simulation code
//
#include <nmstl/ntime>

#include "horde_namespace.h"

#include "manager_impl.h"
#include "properties.h"


using namespace horde;
using namespace nmstl;
using namespace std;

HORDE_NAMESPACE_BEGIN;

/**
 * A simulated I/O handler. 
 *
 * This will actually send and receive UDP packets like a standard \a channel_iohandler, 
 * but this may add simulated latency and losses before transmitting packets.
 *
 * Only TX path is simulated.
 *
 * @see channel_iohandler
 */
class simulated_channel_iohandler : public channel_iohandler {
protected:  
  int rtt_mean;       ///< Simulated RTT's mean.
  int rtt_sdev;       ///< Simulated RTT's standard deviation.
  float loss;         ///< Simulated loss rate, 0.0-1.0
  
  ntime last_send;         ///< Last time simulated_send() was called.
public:
  /**
   * Create a simulated I/O handler. Use \a params to set the simulated channel
   * model's parameters:
   *
   * \code
   * sim.rtt.mean   int   : milliseconds
   * sim.rtt.sdev   int
   * sim.loss       float : random loss rate (0.0-1.0)
   * sim.maxburst   int   : maximum size of a burst, before packets are dropped.
   * \endcode   
   */
  simulated_channel_iohandler(properties& params,
			      io_event_loop& ioloop, string channel_name, inet_address bind_addr,
			      callback<void, horde_net_packet, inet_address> recv_handler,
			      callback<void> stat_handler=NULL, size_t recv_buffer_sz=2048);
  
  void async_send(horde_net_packet dest_buf, inet_address addr) {    
    simulated_send(dest_buf, addr);
  }

  void start_io() {
    channel_iohandler::start_io();
  }

  void send_now(horde_net_packet packet, inet_address addr) {
    channel_iohandler::send_now(packet, addr);
  }
  
protected:
  /// send with a simulated latency and loss
  void simulated_send(horde_net_packet dest_buf, inet_address addr);           
};

HORDE_NAMESPACE_END;

// debug helpers
string float_to_str(double data, const char* format_str="%0.2lf");
string int_to_str(int data, const char* format_str="%08x");

#endif
