// -*- mode: C++; -*-

#ifndef __HORDE_CHANNEL_MANAGER_H__
#define __HORDE_CHANNEL_MANAGER_H__

#include <vector>

#include <nmstl/netioevent>
#include <nmstl/ptr>

#include "string_hashmap.h"
#include "horde_namespace.h"
#include "txslot.h"
#include "stats.h"

using namespace std;
using namespace nmstl;

HORDE_NAMESPACE_BEGIN;

class horde_net_packet;

//
// time32
//
typedef uint32_t time32;

/// Returns the number of milliseconds that have elapsed from an arbitrary point in time
time32 time32_now();  

/// Convert a time32 value into a conventional timestamp.
ntime time32_to_ntime(time32 t);

/**
 * Manager for a single network channel.
 *
 * The pool manager maintains the invariant that there is at most one
 * channel manager per active interface. The specific \a
 * network_channel_manager object is created by the \a
 * create_network_channel_manager() factory function.
 *
 * Subclasses should implement a congestion control algorithm that
 * appropriately determines the send-data rate on the underlying
 * network channel. Whenever a a new transmission slot becomes
 * available, the notify_core_new_tx_slot function should be
 * called. The manager's consume_tx_slot function will be called when
 * that transmission slot is used up by the packet scheduler.
 *
 * Whenver data arrives on the channel, a manager should call the
 * notify_core_packet_received function, unless that data is a network
 * acknowledgement, or some other control packet used by the channel
 * manager implementation.
 *  
 * Subclasses should export a \a channel_statistics subclass that
 * allows read-only access to the state of this channel.
 */
class network_channel_manager {
public:
  /** 
   * Get the channel identifier of the underlying channel. This is the identifier that is
   * in horde network packet headers. 
   */
  virtual int get_channel_id() = 0;

  /** Get a unique string identifier for the underlying channel. */
  virtual string get_channel_name() = 0;

  /// get an observer to keep track of the channel's statistics.
  virtual ptr<channel_statistics> get_statistics() = 0;
    
  /// starts the channel manager, allows it to start generating TX slots.
  virtual bool startup() = 0; 

  /// cleanly shuts down the channel manager, stops it from generating TX slots.
  virtual bool shutdown() = 0;  // needs to destroy timers etc, before returning

  /// returns the bind address of this channel manager
  virtual inet_address get_address() = 0;

  /**
   * Transmits data on this channel using the given TX slot capability iff this slot was
   * created by this channel manager, and iff the slot is still valid.
   *
   * @post <tt>slot.is_valid == false</tt>
   */
  virtual void consume_tx_slot(ptr<tx_slot> slot, horde_net_packet data, int packet_uid, inet_address send_address) = 0;

  /// send a control packet, preferably bypass congestion control rate-logic.
  virtual void send_control_packet(inet_address peer_address, horde_net_packet packet, bool no_delay=false) = 0;
};

/**
 * Transmits data using the given TX slot capability iff the slot is still valid.
 *
 * @post <tt>slot.is_valid == false</tt>
 */
inline void consume_tx_slot(ptr<tx_slot> slot, horde_net_packet data, int packet_uid, inet_address send_address) {
  slot->get_channel()->consume_tx_slot(slot, data, packet_uid, send_address);
}

/**
 * Factory function to create channel managers from string type arguments.
 *
 * @param channel_id A number 0-255 that identifies the channel to the pool manager.
 * @param manager_type A string containg type and initialization information (see sample properties file)
 * @param bind_address A local bind address, without the port. 
 */
ptr<network_channel_manager> create_network_channel_manager(string channel_name, int channel_id, string manager_type, 
							    inet_address bind_address, io_event_loop& ioloop);

HORDE_NAMESPACE_END;

#endif
