// -*- mode: C++; -*-

/** @file manager_impl.h
 * 
 * Various implementations of the network channel manager interface.
 */

#ifndef __HORDE_CHANNEL_MANAGER_IMPL_H__
#define __HORDE_CHANNEL_MANAGER_IMPL_H__

#include <map>
#include <list>

#include "horde_namespace.h"
#include "properties.h"
#include "packet_io.h"
#include "manager.h"
#include "txslot.h"
#include "horde_log.h"

using namespace horde;
using namespace nmstl;
using namespace std;

HORDE_NAMESPACE_BEGIN;

void notify_core_new_tx_slot(ptr<tx_slot> slot);
void notify_core_ack_received(int packet_uid, int rtt);
void notify_core_loss_detected(int packet_uid);
void notify_core_packet_received(horde_net_packet packet, inet_address from_address, int channel_id);

/**
 * A channel I/O handler. A helper class for channel manager
 * implementations, to manage the low-level UDP send-recv, and to keep
 * track of I/O statistics.
 */
class channel_iohandler : public udp_iohandler {
protected:
  string channel_name;       ///< string identifier for the underlying channel
  size_t recv_buffer_sz;                                     ///< maximum receive packet size
  callback<void,horde_net_packet,inet_address> recv_handler; ///< callback invoked on a packet receive
  callback<void> stat_update_handler;  ///< callback invoked once every second, right after the local statistics have been updated. 

  int count_recv_bytes;                        ///< a count of received bytes, since start_io()
  int count_sent_bytes;                        ///< a count of sent bytes, since start_io()
  int ave_send_rate;                           ///< average send rate (bytes/sec)
  int ave_recv_rate;                           ///< average receive rate (bytes/sec)  

  int last_count_sent;                         ///< last value of count_sent_bytes (at end of last second)
  int last_count_recv;                         ///< last value of count_recv_bytes (at end of last second)

  bool running;                                ///< Is this I/O handler running.
  callback_timer* stat_update_timer;             ///< a periodic timer that updates the send/recv rate averages

public:
  channel_iohandler(io_event_loop& ioloop, string channel_name, inet_address bind_addr,
		    callback<void, horde_net_packet, inet_address> recv_handler,
		    callback<void> stat_handler=NULL, size_t recv_buffer_sz=2048)
    : udp_iohandler(ioloop, bind_addr, false, false, false), channel_name(channel_name), 
      recv_buffer_sz(recv_buffer_sz), recv_handler(recv_handler), stat_update_handler(stat_handler), running(false),
      count_recv_bytes(0), count_sent_bytes(0), ave_send_rate(0), ave_recv_rate(0), 
      last_count_sent(0), last_count_recv(0), stat_update_timer(NULL)
  { 
    LOG_EVENT_DEBUG("net") << "created I/O handler for: " << get_name();
  }

  /// free resources
  ~channel_iohandler() { stop_io(); }

  string get_name() {
    return channel_name;
  }

  /// send a packet to the specified destination address, without blocking.
  virtual void async_send(horde_net_packet packet, inet_address addr) {
    send_now(packet, addr);
  }

  /// send a packet immediately.
  void send_now(horde_net_packet packet, inet_address addr) {
    if (running) {    
      LOG_NET("send") << channel_name << "\t" << addr.as_string() << "\t" << packet;
      ssize_t n = send_packet(packet.get_buf(), addr);
      if (n <= 0) { // send failed
	LOG_NET("error") << channel_name << "\t send failed: '" << strerror(errno) << "', errno=" << errno << ", n=" << n;
      } else { // sent some bytes
	count_sent_bytes += n;
      }
    }
  }

  /** 
   * Stop I/O on this channel. Sends will not work after this, and no
   * more receive events will be generated by this manager.
   *
   * close()'s this I/O handler.
   */
  virtual void stop_io() {
    running = false;

    want_write(false);
    want_read(false);

    if (stat_update_timer != NULL) stat_update_timer->disarm();

    close();
  }

  /** 
   * Start I/O on this channel. 
   * 
   * @pre stop_io() was not previously called.
   */
  virtual void start_io() {
    running = true;

    want_write(false);   // assume i can send whenever, since this is a UDP socket.
    want_read(true);

    count_sent_bytes = 0;
    count_recv_bytes = 0;
    
    stat_update_clock_tick();
  }

  /// returns number of sent bytes since since start_io() (not including horde headers)
  int get_sent_count() { return count_sent_bytes; }

  /// returns number of received bytes since start_io() (not including horde headers)
  int get_recv_count() { return count_recv_bytes; }

  /// returns average send rate in bytes/sec (not including horde headers; EWMA recalculated every second)
  int get_ave_send_rate() { return ave_send_rate; }

  /// returns average recv rate in bytes/sec (not including horde headers; EWMA recalculated every second)
  int get_ave_recv_rate() { return ave_recv_rate; }

  /// returns the number of sent bytes since the last clock tick (clock ticks once per second)
  int get_sent_count_current_second() { 
    return (count_sent_bytes - last_count_sent);
  }

protected:    
  /// timer function that recalculates average send/recv rates once per second. 
  void stat_update_clock_tick() {
    //// update averages
    //ave_send_rate = (ave_send_rate + (count_sent_bytes - last_send_count)) / 2;
    //ave_recv_rate = (ave_recv_rate + (count_recv_bytes - last_recv_count)) / 2;

    ave_send_rate = (count_sent_bytes - last_count_sent);
    ave_recv_rate = (count_recv_bytes - last_count_recv);

    last_count_sent = count_sent_bytes;
    last_count_recv = count_recv_bytes;

    LOG_EVENT_DEBUG("net") << "stat_update_clock_tick: ave_send_rate=" << ave_send_rate 
			   << ", ave_recv_rate=" << ave_recv_rate
			   << ", last_count_sent=" << last_count_sent << ", last_count_recv=" << last_count_recv;

    // do any other related work
    if (stat_update_handler) stat_update_handler();

    // ... and call myself again after 1 sec.
    stat_update_timer= (new callback_timer
		      (*(get_loop()), wrap(this, &channel_iohandler::stat_update_clock_tick)));
    stat_update_timer->arm(ntime::now_plus_msecs(1000));
  }
  
  /// handles a packet receive
  void ravail() { 
    horde_net_packet recv_buf(recv_buffer_sz);
    inet_address recv_addr;

    size_t n = recv_packet(recv_buf.get_buf(), recv_addr);
    if (n <= 0) { // recv failed
      LOG_NET("error") << channel_name << "\t recv failed: '" << strerror(errno) << "', errno=" << errno;
    } else { // got a valid packet
      LOG_NET("recv") << channel_name << "\t" << recv_addr.as_string() << "\t" << recv_buf;
      count_recv_bytes += recv_buf.get_length() - horde_net_packet::header_sz;     // update count
      recv_handler(recv_buf, recv_addr);   // pass packet on to appropriate recv handler    
    }
  }
};

/**
 * Skeleton implementation of the network_channel_manager interface.
 * 
 * Abstract class; not fully functional; has some pure virtual
 * functions. Never generates TX slots and doesn't provide a \a
 * consume_tx_slot() implementation.
 *
 * The main functionality provided here is related to the network ACK
 * mechanism. This allows for ack feedback, loss detection, and
 * maintainence of average RTT and loss rate.
 *
 * Notes for subclasses:
 *
 * ALWAYS call this class's send_data() method to send data. Don't
 * call the async_send() method of the I/O handler
 * directly. send_data() implements ACK support, and this convention
 * is needed to ensure the ACK/LOSS detection works properly.
 *
 * Conversely, async_send() should be used to send all control packets.
 * This is what the default implementation in this class does.
 */
class skeleton_channel_manager_impl : public network_channel_manager {
  // 
  // NOTES: 
  //  - Keep a list/map of unack'd packets. 
  //  - Ignore ACK for a packet not in the list.
  //  - Basically, timeouts for _every_ packet sent. If I don't get
  //    an ACK for that packet within the timeout period, flag that
  //    as a loss, and ignore an ACK if it arrives later...
  //  - use a 500msec heartbeat timer to implement coarse timeouts...
  //  - TODO: impl a better condition for loss detection. (really?)
  //  
  //
private: 
  /** The unique channel identifier, known by pool manager. Packets are tagged with this. */
  int channel_id;

  /** Current send sequence number. */
  uint32_t send_seq;
    
  /** Smoothed rtt in milliseconds, from measurements. */
  int srtt_msec;    

  /** A measured estimate of the RTT variance. */
  int rttvar_msec;

  /** round-trip-timeout estimate. */
  uint32_t rto_msec;

  /// recalculate srtt, rttvar, rto and return rtt
  uint32_t update_rtt_estimators(uint32_t time_echo);

  uint32_t ack_count;     ///< number of packets ack'd by peer
  uint32_t loss_count;    ///< number of losses detected
  uint32_t total_sent;    ///< total sent

  int count_lost_bytes;   ///< number of bytes lost
  int last_count_lost;    ///< value of count_lost_bytes last time stat_update_beat was called
  float loss_rate;        ///< ewma of loss rate as a percentage

  static const int MAX_ACK_DELAY = 10;
  static const int MAX_ACKS_PER_PACKET = 4;
  int max_ack_delay; // if > 0 then ack delays enabled. 
  list< pair<inet_address, packet_ack> > ackq;
  callback_timer *ack_send_timer;

  //
  // data structures etc for ack/loss detection mechanism
  //

  /**
   * Relative timeouts for different packets.
   *
   * The first element of the pair is the packet sequence number; the second
   * element is the timeout time (a time32 value).
   */
  list< pair<int,int> > timeouts;

  /**
   * Map to get the global packet UID from the local sequence number
   * of a previous send. Contains info for outstanding packets that
   * are unacknowledged, and are not detected to have been lost.
   */
  map<int,int> map_packet_uid;

  /// map to get packet size in bytes from packet UID.
  map<int,int> map_packet_size;

  callback_timer* heartbeat_timer;

  void loss_detector_beat(); //< function called by the heartbeat timer

  void schedule_heartbeat(); ///< schedule the heartbeat timer
  void stop_heartbeat();     ///< stop the heartbeat timer, if it is scheduled to run

  void loss_detected(int seq, int packet_uid, bool timeout_loss);

  void process_incoming_net_ack(packet_ack* ack); ///< process the received ack and raise any relevant events
  void send_net_ack(horde_net_packet packet, inet_address dest_addr); ///< send an ack for the given packet

  void process_incoming_composite_ack(horde_net_packet& packet);
  void send_composite_ack_now();

  void stat_update_beat(); ///< called roughly once-per-second when channel stats are updated.
protected:
  /** Is this channel manager running? */
  bool running;

  /** The I/O event loop used by this channel manager */
  io_event_loop *io_loop;

  /** The associated channel I/O handler */
  ptr<channel_iohandler> iohandler;


  /** Sets up iohandler etc, but leaves running=false.*/
  skeleton_channel_manager_impl(string channel_name, int channel_id, properties& params, 
				inet_address bind_addr, io_event_loop& iol);

  /** 
   * Called whenever the I/O handler receives a packet. If the packet
   * is an acknowledgement, it is handled, otherwise it is forwarded
   * on to notify_core_packet_received().
   */
  void handle_recv(horde_net_packet recv_data, inet_address recv_addr);

  /** 
   * Called whenever a loss is detected. Exactly once for every detected loss.
   *
   * notify_core_loss_detected() will be called immediately after this returns.
   */
  virtual void notify_loss (int channel_seq, int packet_uid, bool timeout_loss) { }

  /** 
   * Called whenever an ack is received. Exactly once for every non-duplicate received ack.
   *
   * notify_core_ack_received() will be called immediately after this returns.
   */
  virtual void notify_ack (int channel_seq, int packet_uid, int rtt_msec) { }

  virtual void statistics_update() { }; ///< called roughly once-per-second when channel stats are updated.

  /** 
   * Subclasses MUST use this to send data. This fills in the channel_id, channel_seq and 
   * timestamp in the packet's header.
   */
  void send_data(horde_net_packet data, int packet_uid, inet_address send_address);

  int get_average_rtt_msec() { 
    if (srtt_msec < 0)
      return 0;
    else
      return srtt_msec; 
  }
    
  int get_average_send_rate() { 
    if (iohandler == NULL)
      return 0;
    else
      return iohandler->get_ave_send_rate(); 
  }
  int get_average_recv_rate() { 
    if (iohandler == NULL)
      return 0;
    else
      return iohandler->get_ave_recv_rate(); 
  }

  int get_loss_rate() { /// @todo make this a near-past average...?
    return (int)(loss_rate+0.5); // round up
  }

  virtual int get_allowed_send_rate() { 
    return get_average_send_rate(); 
  }

  class skeletal_statistics : public channel_statistics {
  private:
    skeleton_channel_manager_impl* parent;
  public:
    skeletal_statistics(skeleton_channel_manager_impl* parent) : parent(parent) { }
    
    string get_channel_name()   { return parent->get_channel_name();      }
    int get_average_rtt_msec()  { return parent->get_average_rtt_msec();  }
    int get_loss_rate()         { return parent->get_loss_rate();         }
    int get_average_send_rate() { return parent->get_average_send_rate(); }
    int get_average_recv_rate() { return parent->get_average_recv_rate(); }
    int get_allowed_send_rate() { return parent->get_allowed_send_rate(); }
  };

public:
  int get_channel_id() { return channel_id; }
  string get_channel_name() { return iohandler->get_name(); }
  inet_address get_address() { return iohandler->get_sockname(); }

  bool startup();
  bool shutdown();

  bool is_running() { return running; }

  ptr<channel_statistics> get_statistics() {
    return ptr<channel_statistics>(new skeletal_statistics(this));
  }

  /**
   * @note Other than the packet type, the other header fields in a
   * control packet are not initialized. They're garbage.
   */
  void send_control_packet(inet_address send_address, horde_net_packet data, bool no_delay=false);
};

//
// CBR channel manager
//
class cbr_channel_manager_impl : public skeleton_channel_manager_impl {
protected:
  int bitrate;
  int packet_size;
  int period_msecs;

  callback_timer *slot_gen_timer;

  void slot_gen_timer_fire();

  int get_allowed_send_rate();
public:
  cbr_channel_manager_impl(string name, int channel_id, properties& params, inet_address bind_addr, io_event_loop& loop);
    
  bool startup();
  bool shutdown();
    
  void consume_tx_slot(ptr<tx_slot> slot, horde_net_packet data, int packet_uid, inet_address send_address);
};

//
// Basic AIMD channel manager
//
class aimd_channel_manager_impl : public skeleton_channel_manager_impl {
protected:
  int packet_size;
  int cwnd;
  int fract_cwnd;
  int max_cwnd;
  int last_loss_seq;
  callback_timer *slot_gen_timer;

  static const int DEFAULT_MAX_CWND = 100;

  enum state_t { SLOW_START, AIMD } state;

  static const int SLOT_GEN_TIMER_PERIOD = 1000; // at least generate one packet per second...

  void notify_loss (int channel_seq, int packet_uid, bool timeout_loss);
  void notify_ack (int channel_seq, int packet_uid, int rtt_msec);

  void generate_tx_slot();
  void slot_gen_timer_fire();

  int get_allowed_send_rate();
public:
  aimd_channel_manager_impl(string name, int channel_id, properties& params, inet_address bind_addr, io_event_loop& loop);
    
  bool startup();
  bool shutdown();
    
  void consume_tx_slot(ptr<tx_slot> slot, horde_net_packet data, int packet_uid, inet_address send_address);
};

//
// VERB channel manager (based on Boston IMC06 code) -- based on AIMD code
//
class verb_channel_manager_impl : public skeleton_channel_manager_impl {
protected:
  int packet_size;
  float cwnd;              ///< congestion window
  float max_cwnd;
  float probe_factor;    
  int in_flight;           ///< approx number of packets in-transit over the network
  int last_loss_seq;
  int min_rtt;            ///< minimum RTT seen so far, updated on each ACK
  ntime last_gen_time;    ///< time when the last tx slot was generated
  bool slow_start;        ///< in slow start state

  callback_timer *long_timer, *short_timer, *slot_gen_timer;

  static const int SLOT_GEN_TIMER_PERIOD = 1000; // at least generate one packet per second...
  static const int LONG_TIMER_PERIOD = 5000; ///< long timeout five seconds
  static const int MAX_SLOT_GEN = 10;
  static const int MAXIMUM_MIN_RTT = 550; ///< maximum min_rtt before cwnd-reduce rule kicks in
  static const int INITIAL_MIN_RTT = 2000;

  static const float MAX_PROBE_FACTOR = 0.9;
  static const float MIN_PROBE_FACTOR = 0.9;
  static const float PROBE_FACTOR_STEP = 0.05;

  static const int MAX_SEND_PER_SECOND = 15090;
  static const float DEFAULT_MAX_CWND = 5.5;

  int long_timer_period;
  int maximum_min_rtt;
  float max_probe_factor;
  float min_probe_factor;
  float probe_factor_step;
  int allowed_rate_ewma;
  int max_send_per_second;


  void notify_loss (int channel_seq, int packet_uid, bool timeout_loss);
  void notify_ack (int channel_seq, int packet_uid, int rtt_msec);
  void statistics_update();

  void long_timer_fire();     ///< fires roughly every five seconds
  void short_timer_fire();    ///< fires roughly every 2*SRTT
  void slot_gen_timer_fire(); ///< makes sure at least one slot per second is generated.

  void schedule_short_timer();
  void schedule_long_timer();

  void generate_tx_slot();

  int get_allowed_send_rate();
public:
  verb_channel_manager_impl(string name, int channel_id, properties& params, inet_address bind_addr, io_event_loop& loop);
    
  bool startup();
  bool shutdown();
    
  void consume_tx_slot(ptr<tx_slot> slot, horde_net_packet data, int packet_uid, inet_address send_address);
};

//
// Verb2 channel manager
//
class verb2_channel_manager_impl : public skeleton_channel_manager_impl {
protected:
  int packet_size;
  float cwnd;              ///< congestion window
  float max_cwnd;
  float probe_factor;    
  int in_flight;           ///< approx number of packets in-transit over the network
  int last_loss_seq;
  ntime last_gen_time;    ///< time when the last tx slot was generated
  bool slow_start;        ///< in slow start state

  callback_timer *long_timer, *slot_gen_timer;

  static const int SLOT_GEN_TIMER_PERIOD = 1000; // at least generate one packet per second...
  static const int LONG_TIMER_PERIOD = 5000; ///< long timeout five seconds
  static const int MAX_SLOT_GEN = 10;

  static const float PROBE_FACTOR = 0.9;

  static const int MAX_SEND_PER_SECOND = 16384;
  static const float DEFAULT_MAX_CWND = 5.5;

  int long_timer_period;
  int allowed_rate_ewma;
  int max_send_per_second;

  void notify_loss (int channel_seq, int packet_uid, bool timeout_loss);
  void notify_ack (int channel_seq, int packet_uid, int rtt_msec);
  void statistics_update();

  void long_timer_fire();     ///< fires roughly every five seconds
  void slot_gen_timer_fire(); ///< makes sure at least one slot per second is generated.

  void schedule_long_timer();

  void generate_tx_slot();

  int get_allowed_send_rate();
public:
  verb2_channel_manager_impl(string name, int channel_id, properties& params, inet_address bind_addr, io_event_loop& loop);
    
  bool startup();
  bool shutdown();
    
  void consume_tx_slot(ptr<tx_slot> slot, horde_net_packet data, int packet_uid, inet_address send_address);
};


HORDE_NAMESPACE_END;

#endif
