#include <sys/socket.h>
#include <net/if.h> 
#include <ifaddrs.h>
#include <string.h>

#include <nmstl/net>
#include <nmstl/ntime>

#include "packet_io.h"
#include "horde_log.h"

#include "poolman.h"
#include "manager.h"

using namespace nmstl;
using namespace std;
using namespace horde;

HORDE_NAMESPACE_BEGIN;

void notify_core_iface_up(inet_address address);
void notify_core_iface_down(inet_address address);

/// need a reference to fold down time to a 32-bit val
ntime time32_start_ref; 

// --------- time32 methods ------------
uint32_t time32_now() {
  return (uint32_t)((ntime::now() - time32_start_ref).to_msecs());
}

ntime time32_to_ntime(uint32_t t) {
  return (time32_start_ref + ntime::msecs(t));
}

void time32_init() {
  time32_start_ref = ntime::now();
}

//
// pool manager 
//
network_pool_manager* the_poolman = NULL;

void 
network_pool_manager::create_poolman(vector< pair<string, string> >& managed_ifaces, int period, io_event_loop& loop) {
  if (the_poolman) {
    if (the_poolman->is_running()) {
      the_poolman->stop();
    }
    delete the_poolman;
  }

  the_poolman = new network_pool_manager(managed_ifaces, period, loop);
}

network_pool_manager* 
network_pool_manager::get_poolman() { 
  return the_poolman; 
} 

// --------- utility functions ----------

// bridge between nmstl and socket structs
// @param port: port to use, or if zero, port is read from addr
inet_address parse_sockaddr_in(sockaddr_in* addr, int port) { 
  char buf[INET_ADDRSTRLEN+1];  
  inet_ntop(AF_INET, &(addr->sin_addr), buf, INET_ADDRSTRLEN);

  if (port == 0) 
    port = ntohs(addr->sin_port);

  return inet_address(string(buf), port);
}

// looks up an inet4 interface.
// @return: true if an interface with the given name was found; false otherwise.
// @param addr: if the iface_name was found, copies contents of
//             address of first record that matched to this param.
// @param port: port to use, or if zero, port is read from addr
bool lookup_interface(string iface_name, inet_address& addr, int port = 0) {
  bool retval=false;

  const char *iface_id = iface_name.c_str();
  if (iface_id[0] == '+')        // simulated interface: id is of the form '+#interface'
    iface_id = iface_id + 2;

  struct ifaddrs *iface = NULL, *ifaces = NULL;
  getifaddrs(&iface);
  ifaces = iface;

  for (; iface != NULL; iface = iface->ifa_next) {
    if ((iface->ifa_addr != NULL) && (iface->ifa_addr->sa_family == AF_INET)) {     

      if (strcmp(iface_id, iface->ifa_name) == 0) { // found it...
	// copy contents
	addr = parse_sockaddr_in((sockaddr_in*)iface->ifa_addr, port);
	retval=true; 
	break;
      }
    }
  }

  freeifaddrs(ifaces); // avoid mem leak.

  //LOG_NET("debug") << iface_name << ", " << addr << ", " << port << ", " << xstr << ", " << retval;

  return retval;
}

// --------- constructor / start / stop --------------

network_pool_manager::network_pool_manager(vector< pair<string, string> >& managed_ifaces,		       
					   int period, io_event_loop& loop) 
  : period_msecs(period), running(false), ioloop(&loop), ifaces(), iface_manager_types(),
    iface_addresses(), iohandler_map() {
  for (vector< pair<string, string> >::const_iterator iter = managed_ifaces.begin();
       iter != managed_ifaces.end(); iter++) {
    string iface = iter->first;
    string manager_type = iter->second;
    
    ifaces.push_back(iface);
    iface_manager_types[iface] = manager_type;
  }
}

void
network_pool_manager::start() {
  if (!running) {
    // debug
    string ifacestr("");
    for (vector<string>::iterator iter = ifaces.begin(); iter != ifaces.end(); iter++) {
      ifacestr += *iter;
      ifacestr += ' ';
    }

    LOG_EVENT("poolman") << "net poolman started, monitoring: " << ifacestr;

    // make sure time32 is initialized
    time32_init();

    // start poolman
    running = true;
    wakeup();
  } // else: wakeup will be called soon, so do nothing...
}

void
network_pool_manager::stop() {

  // shut down every interface...
  if (running)
    for (int i = 0; i < ifaces.size(); i++)
      iface_down(i);

  LOG_EVENT("poolman") << "net poolman stopped.";
  running = false;
}

// ---------- wakeup --------------------
void
network_pool_manager::wakeup() {
  if (running) {
    int idx = 0;
    //LOG_EVENT_DEBUG("poolman") << "poolman woke up.";
    for (vector<string>::const_iterator iter = ifaces.begin();
	 iter != ifaces.end(); iter++, idx++) {
      string iface = *iter;

      // find last round IP of this interface...
      STRING_HASHMAP(inet_address)::iterator if_addr = iface_addresses.find(iface);
      bool was_active = (if_addr != iface_addresses.end());
      
      // is interface up?
      inet_address address;
      if (lookup_interface(iface, address)) {
	// interface is currently up
	if (was_active) { // was this active in last round?
	  // we only care if the address has changed
	  if (if_addr->second != address) { 
	    LOG_EVENT("poolman") << "rebooting interface <" << iface << ">, address change.";
	    iface_down(idx);
	    iface_up(idx, address);
	  } else {
	    //LOG_EVENT_DEBUG("poolman") << "unchanged interface <" << iface << ">";
	  }
	} else {
	  // newly activated interface...
	  LOG_EVENT("poolman") << "interface up <" << iface << ">";
	  iface_up(idx, address);
	}
      } else {
	// interface is not currently up, but was active...
	if (was_active) {
	  LOG_EVENT("poolman") << "interface down <" << iface << ">";
	  iface_down(idx); // active in last round, so clean up...
	} else {
	  //LOG_EVENT_DEBUG("poolman") << "interface never active <" << iface << ">";
	}
      }     
    }

    // flush logs 
    logging_flush(); // as good a place as any to do this...

    // reschedule wakeup...
    (new callback_timer
     (*ioloop, wrap(this, &network_pool_manager::wakeup)))->arm(ntime::now_plus_msecs(period_msecs));
  }
}

// ------------- iface_up / iface_down ---------------

// requires: # of iface_up calls - # of iface_down calls for this idx = 0
void
network_pool_manager::iface_up(int idx, inet_address address) {
  string iface = ifaces[idx];

  // lookup manager type
  string manager_type = iface_manager_types[iface];

  // create and start an network manager for this interface
  ptr<network_channel_manager> ioh = create_network_channel_manager(iface, idx, manager_type, 
								    address, *ioloop);
  if (ioh) {    
    // add to set of active addresses
    iface_addresses[iface] = address;

    iohandler_map[idx] = ioh;  
    
    ioh->startup();
    
    // notify the core
    notify_core_iface_up(ioh->get_address());
  } else {
    LOG_EVENT("warning") << "poolman::iface_up failed to create a channel manager for '" << iface << "'";
  }
}

// requires: # of iface_up calls - # of iface_down calls for this idx = 1
void
network_pool_manager::iface_down(int idx) {
  string iface = ifaces[idx];
  inet_address sockaddr;

  // shut down the io handler for this interface
  hash_map<int, ptr<network_channel_manager> >::const_iterator iter = iohandler_map.find(idx);
  if (iter != iohandler_map.end()) {
    sockaddr = iter->second->get_address();
    iter->second->shutdown(); // shut down io handler
    iohandler_map.erase(idx); // remove from table   
  }

  // remove this interface from the set of iface_addresses set
  STRING_HASHMAP(inet_address)::iterator if_addr = iface_addresses.find(iface);
  if (if_addr != iface_addresses.end()) { // if found...
    inet_address address = if_addr->second;

    iface_addresses.erase(if_addr); // remove from set of active addresses

    // notify the core
    if (sockaddr) notify_core_iface_down(sockaddr);
  }
}

void
network_pool_manager::synchronize_core() {
  for (hash_map<int, ptr<network_channel_manager> >::const_iterator iter = iohandler_map.begin();
       iter != iohandler_map.end();
       iter++) {
    notify_core_iface_up(iter->second->get_address());
  }
}

// --------------- misc --------------------------------

ptr<network_channel_manager>
network_pool_manager::get_manager_from_id(int channel_id) {
  if ((channel_id >= 0) && (channel_id < ifaces.size())) {
    // find last round IP of this interface...
    hash_map<int, ptr<network_channel_manager> >::const_iterator iface = iohandler_map.find(channel_id);
    if (iface != iohandler_map.end()) return (iface->second);
  }

  return ptr<network_channel_manager>();
}


vector< ptr<channel_statistics> > 
network_pool_manager::get_statistics() {
  vector< ptr<channel_statistics> > stats;

  for (hash_map<int, ptr<network_channel_manager> >::const_iterator iter = iohandler_map.begin();
       iter != iohandler_map.end(); iter++) {
    stats.push_back(iter->second->get_statistics());
  }

  return stats;
}

// ---------------- packet I/O ------------------------

int control_send_round=0;

void 
network_pool_manager::send_control_packet(inet_address peer_address, 
					  horde_net_packet packet,
					  bool resilient, bool no_delay) {
  // pick one interface
  int round = control_send_round++;
  int n = ifaces.size();
  for (int i = 0; i < n; i++) {
    int ifx = (i + round) % n;
    hash_map<int, ptr<network_channel_manager> >::iterator iter = iohandler_map.find(ifx);
    if (iter != iohandler_map.end()) { // this interface is valid.

      ptr<network_channel_manager> ioh = iter->second;

      ioh->send_control_packet(peer_address, packet, no_delay); // send the request
      if (!resilient) return; // return after first send, if not in resilient mode.
    }
  }
}

HORDE_NAMESPACE_END;
