#include <strings.h>

#include <vector>

#include "string_hashmap.h"
#include "horde_log.h"
#include "event_handlers.h"
#include "core/control_netio.h"
#include "stream.h"
#include "session_daemon.h"

HORDE_NAMESPACE_BEGIN;

//
// table of saved stream requests
//
struct stream_request_t {
  string stream_id;
  int local_uid;
  int remote_uid;
};

vector<stream_request_t> received_stream_requests;

/// retrieves and forgets the oldest connection request for the given stream_id, or returns false.
bool get_old_stream_request(string stream_id, int& local_uid, int& remote_uid) {
  for(vector<stream_request_t>::iterator iter = received_stream_requests.begin();
      iter < received_stream_requests.end(); iter++) {
    if (iter->stream_id == stream_id) { // found it.
      local_uid = iter->local_uid;
      remote_uid = iter->remote_uid;
      received_stream_requests.erase(iter);
      return true;
    }
  }
  return false;
}

/// saves the given stream connection request for later use
void remember_stream_request(string stream_id, int local_uid, int remote_uid) {
  // remove a previous entry for this stream_id, if it exists
  int ignore;
  get_old_stream_request(stream_id, ignore, ignore);

  // add a new entry
  stream_request_t req;
  req.stream_id = stream_id;
  req.local_uid = local_uid;
  req.remote_uid = remote_uid;
  
  received_stream_requests.push_back(req);
}

//
// misc stuff
//
string 
stream::to_debug_string() {
  return "stream{ id=\'" + stream_id + "\', uid=" + (int)stream_uid 
    + ", uid'=" + (int)remote_stream_uid + ", state=" + (int)stream_state
    + "}";
}

//
// UID stuff
//

// (note: should implement this initialization in a better way...)
// invariant uid:0 never unused
bool uid_table[256] = {true,  false, false, false, false, false, false, false, // 1
		       false, false, false, false, false, false, false, false, 
		       false, false, false, false, false, false, false, false, 
		       false, false, false, false, false, false, false, false, // 4
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false, // 8
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false, // 12
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false, // 16
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false, 
		       false, false, false, false, false, false, false, false, // 20
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false, // 24
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false, // 28
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false,
		       false, false, false, false, false, false, false, false }; // 32
#define MAX_UID 255

void stream::uid_free(int uid) {
  if ((uid > 0) && (uid <= MAX_UID)) {
    uid_table[uid] = false;
  }
}

// note: because of randomization, not gauranteed to find a free UID even if one exists.  
/// @todo better implementation...
int stream::uid_alloc() {
  for (int i = 1; i <= MAX_UID; i++) {
    int j = rand() % (MAX_UID+1);
    if (!uid_table[j]) { // found a free uid
      uid_table[j] = true; // note that this uid is used
      return j;
    }
  }

  return 0; // no free uid's found, return the 'null' uid 
}

//
// event handlers for network control I/O and their helpers
//
void 
notify_stream_connect_request(string stream_id, int local_uid, int remote_uid) {
  stream* pstream = stream::lookup_stream(stream_id);

  if ((pstream) && (pstream->stream_state == stream::READY_TO_ACCEPT)) {
    // stream is connected

    // send network ack
    control_netio::send_stream_connect_ack(pstream->stream_id, pstream->stream_uid, remote_uid);

    // the above send may be delayed, so delay my local response to this connect request...
    (new callback_timer(io_loop, wrap((stream*)pstream, &stream::connected, remote_uid)))->arm(ntime::now_plus_msecs(control_netio::get_max_send_delay()+1));

  } else if ((pstream) && (pstream->stream_state == stream::CONNECTED)) {
    // send another network ack (maybe my last ack got dropped)
    control_netio::send_stream_connect_ack(pstream->stream_id, pstream->stream_uid, remote_uid);    
  } else {
    // // ignore connect request
    // LOG_EVENT("stream") << "ignoring stream connect request: '" << stream_id << "'";
    remember_stream_request(stream_id, local_uid, remote_uid);
    LOG_EVENT("stream") << "remembering stream connect request: '" << stream_id 
			<< "', luid=" << local_uid << ", ruid=" << remote_uid;
  }
}

void 
notify_stream_connect_ack(string stream_id, int local_uid, int remote_uid) {
  stream* pstream = stream::lookup_stream(stream_id);

  if ((pstream) && (pstream->stream_state == stream::TRYING_TO_CONNECT)) {
    // stream is connected
    pstream->connected(remote_uid);
  } else {
    // ignore connect request
    LOG_EVENT("stream") << "ignoring stream connect request: '" << stream_id << "'";
  }
}

void 
notify_stream_close_request(string stream_id, int local_uid, int remote_uid) {
  stream* pstream = stream::lookup_stream(stream_id);

  if ((pstream) && (pstream->stream_uid == local_uid)) {
    stream::close(stream_id, pstream->owner, false);
  } else {
    // ignore close request
    LOG_EVENT("stream") << "ignoring stream close request: '" << stream_id << "':" << local_uid;
  }
}

void 
notify_stream_data_ack(int stream_id, int adu_id, int frag_start, int frag_len, int rtt) {
  stream* pstream = stream::lookup_stream(stream_id);

  if ((pstream) && (pstream->event_mask & stream::EVENT_MASK_ACK)) {   
    notify_syscall_stream_data_ack(pstream->owner, stream_id, adu_id, frag_start, frag_len, rtt);
  }
}

void 
notify_stream_data_loss(int stream_id, int adu_id, int frag_start, int frag_len) {
  stream* pstream = stream::lookup_stream(stream_id);
  if ((pstream) && (pstream->event_mask & stream::EVENT_MASK_LOSS)) {
    notify_syscall_stream_data_loss(pstream->owner, stream_id, adu_id, frag_start, frag_len);
  }
}

void 
notify_stream_data_recv(int stream_id, int adu_id, int frag_start, int frag_len, int tot_len, const char* data) {
  stream* pstream = stream::lookup_stream(stream_id);
  if ((pstream) && (pstream->event_mask & stream::EVENT_MASK_RECV)) {     
    notify_syscall_stream_data_recv(pstream->owner, stream_id, adu_id, frag_start, tot_len, data, frag_len);
  }
}

//
// sendq
//
void stream::sendq_push(sendq_elem_t& e) {
  if (sendq_size < SENDQ_MAX_SIZE) {
    sendq.push_back(e);
    sendq_size += e.data.length();
    
    LOG_EVENT_DEBUG("stream") << "push: stream <" << stream_uid << "> sendq_size=" << sendq_size << ", emask: " << (event_mask & EVENT_MASK_CHOKE) << ", choked=" << source_choked;
    
    if (sendq_size > SENDQ_CHOKE_THRESHOLD) {
      if (event_mask & EVENT_MASK_CHOKE) {
	notify_syscall_stream_choke(owner, stream_uid, true);  // back-pressure: tell source to stop sending data...
	source_choked = true;
      }
    }
  } else {
    LOG_EVENT("stream") << "stream <" << stream_uid << "> sendq is full, discarding data.";

    if (event_mask & stream::EVENT_MASK_LOSS)
      notify_syscall_stream_data_loss(this->owner, stream_uid, e.adu_id, e.adu_offset, e.data.length());
  }
}

void stream::sendq_pop() {
  int adu_id = sendq.front().adu_id;
  size_t adu_len = sendq.front().data.length();
  sendq_size -= adu_len;
  sendq.pop_front();

  LOG_EVENT_DEBUG("stream") << "pop: adu=" << adu_id << ", adu_len=" <<  adu_len
			    << ", sendq_size=" << sendq_size << ", emask: " << (event_mask & EVENT_MASK_CHOKE) << ", choked=" << source_choked;

  if (sendq_size <= 0) {
    sendq_size = 0;

    // unchoke?
    if ((source_choked) && (event_mask & EVENT_MASK_CHOKE)) {
      notify_syscall_stream_choke(owner, stream_uid, false);  // tell source it's okay to send data again...    
      source_choked = false;
    }
  
    // notify on flush..?
    if (event_mask & EVENT_MASK_FLUSHED) {
       notify_syscall_stream_flushed(owner, stream_uid);
    }
  }
}

void stream::sendq_clear(io_principal caller) {
  if (owner == caller) {  
    sendq.clear();
    sendq_size = 0;
    
    // unchoke?  
    if (source_choked) {
      if (event_mask & EVENT_MASK_CHOKE) {
	notify_syscall_stream_choke(owner, stream_uid, false); 
	source_choked = false;
      }
    }
    
    // notify on flush..?
    if  (event_mask & EVENT_MASK_FLUSHED) {
      notify_syscall_stream_flushed(owner, stream_uid);
    }
  } else {
    // not the owner io_principal    
    notify_syscall_stream_exception(caller, stream_id, EX_STREAM_INSUFFICIENT_PERMISSION, "Only stream owner is allowed to clear send queue.");
  }
}

void 
stream::send_adu(int adu_id, char* data, size_t data_length, horde_objective_tags tags,
		 io_principal caller) {
  
  if (owner == caller) {  
    sendq_elem_t e(adu_id, data, data_length, false, 0, data_length, tags);    
    sendq_push(e);
  } else {
    // not the owner io_principal    
    notify_syscall_stream_exception(caller, stream_id, EX_STREAM_INSUFFICIENT_PERMISSION, "Only stream owner is allowed to send data.");
    LOG_EVENT_DEBUG("exception") << "owner: " << (void*)(owner) << ", caller=" << (void*)caller;
  }
}

void 
stream::send_fragment(int adu_id, int adu_offset, size_t adu_length, 
		      char* fragment, size_t fragment_length, horde_objective_tags tags,
		      io_principal caller) {

  if (owner == caller) {  
    sendq_elem_t e(adu_id, fragment, fragment_length, true, adu_offset, adu_length, tags);
    sendq_push(e);
  } else {
    // not the owner io_principal    
    notify_syscall_stream_exception(caller, stream_id, EX_STREAM_INSUFFICIENT_PERMISSION, "Only stream owner is allowed to send data.");
  }
}

void 
stream::sendq_read_metadata(int *adu_id, int *current_offset, int *length,
			    bool *is_fragment, int *adu_offset, int *adu_length,
			    horde_objective_tags *tags) {
  
  if (!sendq.empty()) {
    sendq_elem_t e = sendq.front();
    if (adu_id) *adu_id = e.adu_id;
    if (current_offset) *current_offset = e.current_offset;
    if (length) *length = e.data.length();
    if (is_fragment) *is_fragment = e.is_fragment;
    if (adu_offset) *adu_offset = e.adu_offset;
    if (adu_length) *adu_length = e.adu_len;
    if(tags) *tags = e.tags;
  } else { 
    // bad read: nothing in send queue
    if (adu_id) *adu_id = -1;
    if (current_offset) *current_offset = -1;
    if (length) *length = -1;
    if (is_fragment) *is_fragment = -1;
    if (adu_offset) *adu_offset = -1;
    if (adu_length) *adu_length = -1;
    if(tags) *tags = horde_objective_tags();
  }
}

int
stream::sendq_read_data(char* dest_buf, size_t max_read) {
  if (!sendq.empty()) {
    sendq_elem_t *e = &(sendq.front());
    char *pdata = e->data.data() + e->current_offset;
    size_t datalen = e->data.length();
    ssize_t nread = datalen - e->current_offset;

    if (nread >= 0) {
      if (nread > max_read) nread = max_read;
      memcpy(dest_buf, pdata, nread);
      e->current_offset += nread;
      if (e->current_offset >= datalen) sendq_pop();

      return nread;
    } else {
      // if we get here, there's a bug...
      sendq_pop();
      return -1;
    }

  } else {
    // bad read: nothing in send queue
    return -1;
  }
}

dynbuf
stream::sendq_front(int* adu_id) {
  if (sendq.empty()) {
    if (adu_id) *adu_id = -1;
    return dynbuf();
  } else {
    sendq_elem_t e = sendq.front();
    if (adu_id) *adu_id = e.adu_id;
    return e.data;
  }
}

// 
// set of active streams
// 
vector<stream*> streams;
vector<stream*> connected_streams;
STRING_HASHMAP(stream*) streams_id_map;
hash_map<int, stream*>  streams_uid_map;

vector< stream* > stream::get_all_streams() {
  return streams;
}

vector< stream* > stream::get_connected_streams() {
  return connected_streams;
}

stream* stream::lookup_stream(string id) {
  STRING_HASHMAP( stream* )::const_iterator iter = streams_id_map.find(id);
  if (iter != streams_id_map.end()) {
    // found it
    return iter->second;
  } else {
    // didn't find it
    return NULL;
  }
}

stream* stream::lookup_stream(int uid) {
  hash_map<int, stream* >::const_iterator iter = streams_uid_map.find(uid);
  if (iter != streams_uid_map.end()) {
    // found it
    return iter->second;
  } else {
    // didn't find it
    return NULL;
  }
}

//
// close / accept / connect
//

void stream::close_all_streams_owned_by(io_principal caller, bool notify_peer) {

  // collect the id's of the streams that need to be closed...
  vector<string> dead; 
  for (vector<stream*>::const_iterator iter = streams.begin();
       iter != streams.end();
       iter++) {
    if ((*iter)->owner == caller) {
      dead.push_back((*iter)->stream_id);
    }
  }

  // close the streams
  for (vector<string>::const_iterator iter = dead.begin(); iter != dead.end(); iter++) {
    stream::close(*iter, caller, notify_peer);
  }
}

void stream::close(string stream_id, io_principal caller, bool notify_peer) {
  stream* pstream = stream::lookup_stream(stream_id);
  
  if (pstream) {
    if (pstream->owner == caller) {
      pstream->stream_state = stream::CLOSED; // note that it is closed, while we clean up...
      
      // notify the peer...
      if (notify_peer) 
	control_netio::send_stream_close(stream_id, pstream->stream_uid, pstream->remote_stream_uid);
      
      // remove from the active stream set
      STRING_HASHMAP(stream*)::iterator iter_id = streams_id_map.find(stream_id);
      if (iter_id != streams_id_map.end()) streams_id_map.erase(iter_id);     // remove from id:stream map
      
      hash_map<int, stream*>::iterator iter_uid = streams_uid_map.find(pstream->stream_uid);
      if (iter_uid != streams_uid_map.end()) streams_uid_map.erase(iter_uid); // remove from uid:stream map
      
      for (vector<stream*>::iterator iter = streams.begin(); iter != streams.end(); iter++) {
	if ((*iter) == pstream) {           // find and remove from stream vector
	  streams.erase(iter);
	  break;
	}
      }    

      for (vector<stream* >::iterator iter = connected_streams.begin(); iter != connected_streams.end(); iter++) {
	if ((*iter) == pstream) {           // find and remove from connected_stream vector
	  connected_streams.erase(iter);
	  break;
	}
      }
      
      // call the associated close handler     
      notify_syscall_stream_closed(pstream->owner, pstream->stream_id);

      // deallocate the stream object
      delete pstream;

    } else {
      // not the owner io_principal     
      notify_syscall_stream_exception(caller, stream_id, EX_STREAM_INSUFFICIENT_PERMISSION, "Only stream owner is allowed to close the stream.");
    }
  } else {
    // stream doesn't exist...    
    notify_syscall_stream_exception(caller, stream_id, EX_STREAM_BAD_ID, "A stream with this identifier does not exist.");
  }
}

void stream::accept(string stream_id, io_principal caller) {
  if (stream::lookup_stream(stream_id)) {
    // stream already exists...   
    notify_syscall_stream_exception(caller, stream_id, EX_STREAM_BAD_ID, "A stream with this identifier already exists.");
  } else {
    int uid = uid_alloc();
    if (uid) {
      // create a stream object and add it to the stream set
      stream* pstream = new stream(stream::READY_TO_ACCEPT, stream_id, uid, 0, caller);
      streams.push_back(pstream);
      streams_id_map[stream_id] = pstream;
      streams_uid_map[uid] = pstream;

      // if a stream connection request has already been received, 
      // act as if we've just received that request, otherwise wait
      // for a request.
      int luid, ruid;
      if (get_old_stream_request(stream_id, luid, ruid)) {
	notify_stream_connect_request(stream_id, luid, ruid);
      }
				
    } else {     
      notify_syscall_stream_exception(caller, stream_id, EX_STREAM_INTERNAL_ERROR, "Could not allocate a UID.");
    }
  }  
}

void stream::repeated_send_connect_request(stream* pstream) {
  if (pstream->stream_state == TRYING_TO_CONNECT) {
    // send a connection request
    control_netio::send_stream_connect_request(pstream->stream_id, pstream->stream_uid, 0);
    
    // try again after a second
    (new callback_timer(io_loop, wrap(&stream::repeated_send_connect_request, pstream)))->arm(ntime::now_plus_msecs(1000));
  }
}

void stream::connect(string stream_id, io_principal caller) {
  if (stream::lookup_stream(stream_id)) {
    // stream already exists...    
    notify_syscall_stream_exception(caller, stream_id, EX_STREAM_BAD_ID, "A stream with this identifier already exists.");
  } else {
    int uid = stream::uid_alloc();
    if (uid) {
      // create a stream object and add it to the stream set
      stream* pstream = new stream(stream::TRYING_TO_CONNECT, stream_id, uid, 0, caller);
      streams.push_back(pstream);
      streams_id_map[stream_id] = pstream;
      streams_uid_map[uid] = pstream;

      // send a connection request (w/ retry timeout)
      repeated_send_connect_request(pstream);

    } else {      
      notify_syscall_stream_exception(caller, stream_id, EX_STREAM_INTERNAL_ERROR, "Could not allocate a UID.");
    }
  }
}

void 
stream::set_event_mask(bool event_recv, bool event_ack, 
		       bool event_loss, bool event_choke, bool event_flushed,
		       io_principal caller) {

  if (owner == caller) {
    int new_mask = 0;
    if (event_recv) new_mask |= EVENT_MASK_RECV;
    if (event_ack) new_mask |= EVENT_MASK_ACK;
    if (event_loss) new_mask |= EVENT_MASK_LOSS;
    if (event_choke) new_mask |= EVENT_MASK_CHOKE;
    if (event_flushed) new_mask |= EVENT_MASK_FLUSHED;

    LOG_EVENT_DEBUG("stream") << "new event mask: " << new_mask << ", old mask: " << event_mask;

    event_mask = new_mask;
  } else {
    // not the owner io_principal    
    notify_syscall_stream_exception(caller, stream_id, EX_STREAM_INSUFFICIENT_PERMISSION, "Only stream owner is allowed to change event mask.");
  }
}

//
// misc
//
void
stream::connected(int remote_uid) {
  if ((stream_state == stream::READY_TO_ACCEPT) 
      || (stream_state == stream::TRYING_TO_CONNECT)) {

    // prep logging string
    string conntype;
    if (stream_state == stream::READY_TO_ACCEPT) {
      conntype = "ACCEPT:";
    } else {
      conntype = "CONNECT:";
    } 

    // shift this to the connected state
    stream_state = stream::CONNECTED;
    remote_stream_uid = remote_uid;
    connected_streams.push_back(this);

    // notify local event handler
    notify_syscall_stream_connected(owner, stream_id, stream_uid);

    // log
    LOG_EVENT("stream") << conntype << " stream connected: id='" << stream_id 
			<< "', uid=" << stream_uid  << ", ruid=" << remote_stream_uid;
  } 
}

HORDE_NAMESPACE_END;
