/** @file ippc.h
 *
 *  Primitives for the inter-process-procedure-call (IPPC) mechanism
 *  used by Horde.
 *
 *  <b>Channels</b>: IPPC relies on \em channels and \em messages
 *  sent/received over those channels. A channel must be connected
 *  before messages can be sent/received. The current IPPC
 *  implementation uses Unix domain sockets, so when one end of the
 *  channel closes, the other end is notified.
 *
 *  <b>Non-blocking Channels</b>: By default, channels are created in
 *  blocking mode. The file descriptor used by a channel can be put
 *  into non-blocking mode (async I/O), but in this mode the maximum
 *  message size is \a IPPC_MAX_ATOMIC_MESSAGE_SIZE for a send.
 *
 *  <b>Messages</b>: Messages are delivered in order, but some may be
 *  lost atomically (ie.  partial messages will not be received).
 *  Messages have a mandatory type argument, a length, and a data
 *  payload. The payload can be any \em simple object. The payload
 *  object should not contain pointers that point outside the payload.
 *
 *  <b>Multi-part Messages</b>:
 *  A message with a payload larger than \a IPPC_MAX_ATOMIC_MESSAGE_SIZE
 *  will be split into multiple messages during a send. Multi-part send
 *  functionality only works with blocking channels.
 *
 *  <b>RPC vs IPPC</b>: IPPC is a simple RPC-style mechanism implemented
 *  on top of IPC. Since everything is being processed locally, there
 *  is no need to marshall/unmarshall messages: as long as there are
 *  no pointers, data structures can be copied without complications.
 *
 *  @note <b>June 2006</b>: refactored, moved to \a SEQPACKET sockets, and added
 *        multipart functionality.
 *
 *  @author Asfandyar Qureshi (asfand)
 *
 *  @todo Fix the -4 hack in the implementation (right now compensates for \c char \c data[1] fields).
 */
#ifndef __HORDE_IPPC_H__
#define __HORDE_IPPC_H__

#include <limits.h>


struct ippc_channel_context;

/**
 * An IPPC communication channel. A channel is either valid or equal
 * to \a INVALID_CHANNEL. A valid channel is like a valid file
 * descriptor: you shouldn't try to manipulate its value directly;
 * just pass the channel to the functions that need it.
 */
typedef struct ippc_channel_context* ippc_channel;

/// An invalid channel.
#define INVALID_CHANNEL NULL

//
// channel operations
//

int ippc_channel_get_fd(ippc_channel channel);

/**
 * Open a channel using the given path. A server must be ready to
 * accept a client on that path. This function may block.
 *
 * @return Valid channel on success, or \a INVALID_CHANNEL on failure
 * (\c errno has failure code).
 */
ippc_channel ippc_channel_open(const char* pipe_path);

/**
 * Accept a channel on the given path. Waits for a client to
 * try to connect to that path and opens a channel to that
 * client. This function may block.
 *
 * @return Valid channel on success, or \a INVALID_CHANNEL on failure
 * (\c errno has failure code).
 */
ippc_channel ippc_channel_accept(const char* pipe_path);

/**
 * Closes an open channel.
 *
 * @param remove_from_path If non-zero, and the channel has a valid path, the 
 *                         path is unlink'd from the file namespace.
 */
void ippc_channel_close(ippc_channel channel, int remove_from_path=1);

/**
 * Create a channel from a blocking or non-blocking file descriptor
 * for a \a SEQPACKET pipe, and the full filename of that pipe.
 *
 * @param path \a NULL or the valid filename. If not \a NULL, the filename
 *             will be unlinked when this channel is closed.
 */
ippc_channel ippc_channel_create_from_fd(int fd, const char* path=NULL);

/**
 * Puts the underlying file descriptor into non-blocking mode.
 *
 * Non-blocking channels may loose messages and do not support
 * multi-part sends.
 */
void ippc_channel_make_nonblock(ippc_channel channel);

/**
 * Tests whether a channel is open.
 *
 * @return 0 if \a channel is closed or 1 if \a channel is open.
 */
int ippc_channel_is_open(ippc_channel channel);

//
// send/receive
//

/// Maximum IPPC message size
#define IPPC_MAX_MESSAGE_SIZE 65536

/// Maximum IPPC message size for an atomic send
#define IPPC_MAX_ATOMIC_MESSAGE_SIZE PIPE_BUF - 64

/// Return code for the case when \a ippc_message_receive() has
/// received part of a multi-part message on a non-blocking
/// file descriptor and no more data is currently available.
#define IPPC_RECV_INCOMPLETE (char*)(1)

/**
 * Send a message. 
 *
 * If the message is large, it may be fragmented and sent as multiple
 * smaller messages. If any fragment is lost, the entire message is discarded, so you
 * can treat messages as atomic entities, in spite of the fragmentation.
 *
 * @param type The type of the message. This is a user-defined code that allows the
 *             process reveiving a message to determine the structure of the payload.
 *
 * @param msg  Pointer to a message structure to send. The structure should not contain
 *             pointers that point outside the data block. Internal pointers are okay.
 *             The message structure will be copied as-is to the process on the other
 *             end of the IPPC channel.
 *
 * @param msg_length 
 *             Number of bytes top copy from msg.
 *
 * @warning If \a msg_length is greater than \a IPPC_MAX_ATOMIC_MESSAGE and the channel
 *          is in non-blocking mode, the send will not succeed.
 *
 * @return Either <0 in case of error (\c errno has failure code), or msg_length on success.
 *
 * @note If a send fails, and \c errno is equal to \c EBADF then the channel has been closed.
 */
ssize_t ippc_message_send(ippc_channel channel, int type, const void* msg, size_t msg_length);

/**
 * Receive a message, or part of a large multi-part message. When a
 * complete message has been received a pointer to a buffer containing
 * the message payload is returned.
 *
 * @param type   A pointer to a variable where the received message's
 *               type code will be stored.  

 * @param length A pointer to a variable where the received message's
 *               length will be stored.
 *
 * @return       A pointer to the recv'd message payload; or \a NULL (on error, see \c errno);
 *               or \a IPPC_RECV_INCOMPLETE (and \c errno is \c EAGAIN) if a partial 
 *               multi-part message has been received and the channel is in non-blocking mode.
 *
 * @note Don't call \c free on the returned buffer. It may be reused internally for the 
 *       next receive. The contents of the buffer are guaranteed to remain unchanged between 
 *       calls to \a ippc_message_receive().
 */
char* ippc_message_receive(ippc_channel channel, int* type, size_t* length);

/**
 * Receive a message, but don't block longer than \a timeout_ms
 * milliseconds.  See ippc_message_receive(). A non-blocking channel
 * may block up to \a timeout_ms when this function is called.
 */
char* ippc_message_receive_timed(ippc_channel channel, int* type, size_t* length, int timeout_ms);

/**
 * Make an inter-process-procedure call. Call-and-return semantics:
 * send a message, and wait for either a specific type of reply or an
 * exception, ignoring any other messages.
 *
 * @param send_type Type of message to send.
 * @param send_msg  Payload of message to send.
 * @param send_msg_length Length of send payload.
 *
 * @param acceptable_recv_type 
 *            This function only returns when (a) there is an I/O error; (b) an exception 
 *            message is received; or (c) a message whose type is \a acceptable_recv_type
 *            is received.
 *
 * @param recv_type Pointer to var where received message's type is stored.
 * @param recv_length Pointer to var where received message's length is stored.
 * 
 * @return A pointer to the received payload, or \a NULL on error (see \c errno)
 *
 * @note Doesn't work with non-blocking channels.
 */
char* ippc_call(ippc_channel channel, 
		int send_type, const void* send_msg, size_t send_msg_length,
		int acceptable_recv_type,
		int *recv_type, size_t* recv_length);

#include <string>
using namespace std;

/**
 * Debugging helper. Returns a hexdump of the given memory location. 
 */
string to_hexstr(const char* data, size_t datalen);

/**
 * Debugging helper. Takes a printf style format string and a double
 * argument and returns the formatted output from snprintf.
 */
string float_to_str(double data, const char* format_str="%0.2lf");
string int_to_str(int data, const char* format_str="%08x");

#endif
