#include <iostream>
#include <fstream>
#include <cstdlib>

#include "properties.h"

const char FIELD_VALUE_SEPARATOR = ':';
const char COMMENT_CHARACTER = '#';

// helper functions etc
// - these could be more efficient, but they're called infrequently...

string 
to_lowercase(const string& input) {
  string work = input;

  for (int i = 0; i < work.length(); i++) {
    char ch = work[i];
    if ((ch >= 'A') && (ch <= 'Z')) {
      work[i] = 'a' + (ch - 'A');      
    }
  }

  return work;
}

string 
to_uppercase(const string& input) {
  string work = input;

  for (int i = 0; i < work.length(); i++) {
    char ch = work[i];
    if ((ch >= 'a') && (ch <= 'z')) {
      work[i] = 'A' + (ch - 'a');      
    }
  }

  return work;
}

string
trim_spaces(const string& input) {
  // remove trailing/leading spaces...
  string retval;

  int n = input.length();

  int i;
  for (i = 0; i < n; i++) // leading spaces...
    if (input[i] != ' ') break;
  
  if (i < n) { // if something other than spaces...
    int j;
    for (j = n-1; j > i; j--) // leading spaces...
      if (input[j] != ' ') break;

    return string(input, i, j - i + 1);
  } else return "";
}

string
encode_value_string(const string& input) {
  return input; // nothing for now...
}

string
decode_value_string(const string& input) {
  return trim_spaces(input);
}

string
trim_commented_part(const string& input) {
  int n = input.length();  

  for (int i = 0; i < n; i++) // leading spaces...
    if (input[i] == COMMENT_CHARACTER) {
      n = i;
      break;
    }

  return string(input, 0, n);
}

bool
propfile_decode_line(const string& line, bool cerr_output, pair<string,string>& retval, int linenum) {
  //if (cerr_output) cerr << "DECODING: " << line;

  string trimmed = trim_commented_part(line);

  if (trimmed.length() > 0) {
    string::size_type sep_idx = trimmed.find(FIELD_VALUE_SEPARATOR);
    if (sep_idx != string::npos) { // found the separator
      retval.first = trimmed.substr(0, sep_idx);
      retval.second = decode_value_string(trimmed.substr(sep_idx+1, trimmed.length() - sep_idx));

      //if (cerr_output) cerr << " -> DECODED\n";

      return true;
    } else {

      if (cerr_output) cerr << "\033[1;31mline " << linenum 
			    << "\033[0m: ERROR, couldn't find separator\n"
			    << "> \033[1;31m" << line << "\033[0m\n";
      return false;
    }
  } else {
    //if (cerr_output) cerr << " -> COMMENT\n";
    return false;
  }

  //  if (cerr_output)   cerr << endl;
}

// --
vector<string>
properties::list_properties() const {
  vector<string> retval;
  for (hash_map_t::const_iterator iter = property_set.begin();
       iter != property_set.end();
       iter++) {
    retval.push_back((*iter).first);
 }

  return retval;
}

vector<string> 
properties::decode_array_value(const string& value) {
  vector<string> retval;

  int start = 0;
  string::size_type sep_idx = value.find(',');
  while (sep_idx != string::npos) {
    retval.push_back(trim_spaces(value.substr(start, sep_idx - start)));
    start = sep_idx + 1;
    sep_idx = value.find(',', start);
  }
  retval.push_back(trim_spaces(value.substr(start, value.length() - start)));

  return retval;
}

string 
properties::encode_array_value(vector<string> in) {
  string retval;
  bool comma=false;
  for (vector<string>::iterator iter = in.begin();
       iter != in.end(); iter++) {
    if (comma) retval.append(",");
    retval.append(*iter);
    comma = true;
  }

  return retval;
}

const char* value_type_description[] = { 
    "UNKNOWN", 
    "STRING[]",
    "STRING",
    "INTEGER[]",
    "INTEGER",
    "FLOAT[]",
    "FLOAT"
};

const char* 
properties::get_vtype_description(property_value_type t) {
  return (value_type_description[t]);
}

properties::property_value_type 
properties::infer_value_type(const string& value) {
  // hack :(

  int num_chars=0, other_chars=0, commas=0, periods=0;
  int n = value.length();

  for (int i = 0; i < n; i++) {
    char ch = value[i];
    if (ch == ',') commas++;
    else if (ch == '.') periods++;
    else if (((ch - '0') >= 0) && ((ch - '0') <= 9)) num_chars++;
    else other_chars++;	  
  }

  if ((other_chars == 0) && (periods == 0) && (num_chars > 0)) {
    if (commas == 0)
      return TYPE_INTEGER;
    else 
      return TYPE_INTEGER_ARRAY;
  }
  else if ((other_chars == 0) && (periods == commas+1) && (num_chars > 0)) {
    if (commas == 0)
      return TYPE_FLOAT;
    else 
      return TYPE_FLOAT_ARRAY;
  } else {
    if (commas > 0)
      return TYPE_STRING_ARRAY;
    else
      return TYPE_STRING;
  }
}

string
properties::to_string_x() const {
  string retval;
  char buffer[128];

  snprintf(buffer, 128, "properties [%d] {\n", property_set.size());
  retval.append(buffer);

  for (hash_map_t::const_iterator iter = property_set.begin();
       iter != property_set.end();
       iter++) {
    hash_map_t::value_type prop = *iter;
    snprintf(buffer, 128, "    %s: \'%s\' (%s)\n", 
	     prop.first.c_str(), prop.second.c_str(),
	     get_vtype_description(infer_value_type(prop.second)));
    retval.append(buffer);
  }
  retval.append("}");

  return retval;
}

string
properties::to_string() const {
  string retval;
  char buffer[128];

  snprintf(buffer, 128, "properties [%d] {\n", property_set.size());
  retval.append(buffer);

  for (hash_map_t::const_iterator iter = property_set.begin();
       iter != property_set.end();
       iter++) {
    hash_map_t::value_type prop = *iter;
    snprintf(buffer, 128, "    %s: \'%s\'\n", 
	     prop.first.c_str(), prop.second.c_str());
    retval.append(buffer);
  }
  retval.append("}");

  return retval;
}

int
properties::read_from_file(const string& filename, bool errors_to_cerr) {
    // open file...
  ifstream ifile(filename.c_str());

  // return if error...
  if (!ifile.is_open()) return -1;

  // read in variables...
  char buf[512];
  int n = 0, linenum = 1;

  pair<string,string> prop;

  while (!ifile.eof()) {
    ifile.getline(buf, 512);
    if (propfile_decode_line(string(buf), errors_to_cerr, prop, linenum)) {
      this->set(prop.first, prop.second);
      n++;
    }
    linenum++;
  }
  ifile.close();

  return n;
}

int
properties::write_to_file(const string& filename) const {
  
  // open file...
  ofstream ofile(filename.c_str());

  // return if error...
  if (!ofile.is_open()) return -1;

  // write data to file...
  int n = 0;
  for (hash_map_t::const_iterator iter = property_set.begin();
       iter != property_set.end();
       iter++) {
    hash_map_t::value_type prop = *iter;
    ofile << prop.first << FIELD_VALUE_SEPARATOR << ' ' << prop.second << endl;
    n++;
  }

  ofile.close();

  return n;
}

bool
properties::contains(const string& property_name) const {
  hash_map_t::const_iterator val = property_set.find(property_name);

  return (val != property_set.end());
}

string
properties::get(const string& property_name) const {
  hash_map_t::const_iterator val = property_set.find(property_name);

     if (val != property_set.end()) 
       return (*val).second;
     else 
       return "";       // doesn't exist; return a valid string.
}

string
properties::get(const string& property_name, const string& default_value) {
  hash_map_t::const_iterator val = property_set.find(property_name);

     if (val != property_set.end()) 
       return (*val).second;
     else {
       property_set.insert(hash_map_t::value_type(property_name, default_value));
       return default_value;
     }
}

void 
properties::set(const string& property_name, const string& value, bool overwrite_existing) {
  if ((!overwrite_existing) && (this->contains(property_name))) {
    return; // already exists && told not to overwrite, so return...
  }

  property_set[property_name] = value;
}

// returns: num of params parsed from string
int
properties::parse_parameter_string(const string& params) {
  vector<string> vals = decode_array_value(params);
  
  // each element of vals may be of the following forms:
  //  <field>
  //  <field>=
  //  <field>=<value>

  for (vector<string>::const_iterator iter = vals.begin();
       iter != vals.end(); iter++) {
    string param = *iter;
    string::size_type sep_idx = param.find('=');
    if (sep_idx != string::npos) {    
      string field = param.substr(0, sep_idx);
      string value = param.substr(sep_idx+1, param.length() - sep_idx);

      set(field, value);
    } else {
      set(param, "");
    }
  }

  return vals.size();
}
