// -*- mode: C++; -*-

/**
 * @file tribe_msg.h
 *
 * Tribe message protocol used to communicate control messages
 * between leaders and followers.
 */
// NOTE:
//   This would be so much easier with a reliable datagram protocol,
//   but we have to run this over TCP...
#ifndef __TRIBE_MESSAGING__
#define __TRIBE_MESSAGING__

#include <list>
#include <string>
#include <nmstl/netioevent>

using namespace std;
using namespace nmstl;

#define MESSAGE_MAXLEN 512
#define MESSAGE_DEFAULT_PORT 9970


#define TRIBE_MESSAGE_INIT_MESSAGE_PRE  ">>>TRIBE<<<"
#define TRIBE_MESSAGE_INIT_MESSAGE_POST "<<<TRIBE>>>"
#define TRIBE_MESSAGE_KEEP_ALIVE ">ALIVE<"

#define TRIBE_KEEP_ALIVE_PERIOD_MSEC  2500
#define TRIBE_KEEP_ALIVE_TIMEOUT_MSEC 5000

#define TRIBE_MESSAGE_IFACE_UP "+up"
#define TRIBE_MESSAGE_IFACE_DOWN "-down"

extern io_event_loop io_loop;

class tribe_message_iohandler;

typedef void (*message_receive_handler)(tribe_message_iohandler*, string msg);
typedef void (*message_disconnected_handler)(tribe_message_iohandler*);
typedef void (*message_connected_handler)(tribe_message_iohandler*);

typedef void (*accept_handler)(tribe_message_iohandler*);

/**
 * I/O handler to manage the Tribe messaging protocol over a TCP
 * socket. Each message consists of a string, followed by a newline.
 *
 * Has support for a rudimentary connection protocol that won't allow
 * this message handler to become 'connected' until it receives some
 * messages from the tribe peer implying that the peer is actually
 * running tribe.
 */
class tribe_message_iohandler : public io_handler {
private:
  tcpsocket sock;
  enum state_t { ACTIVE, CONNECTED, DISCONNECTED } state;
  string peer_host;

  // write-related
  list<string> out_msg_queue; ///< message queue
  int part_written;           ///< number of bytes already written for head of queue

  // read-related
  static const int READ_BUF_MAXLEN = MESSAGE_MAXLEN + 4;
  char read_buffer[READ_BUF_MAXLEN]; ///< input buffer
  int read_buf_used;                 ///< read_buffer[0:read_buffer_used] contains real data
  
  // event handlers
  message_receive_handler recv_handler;
  message_disconnected_handler disconnected_handler;
  message_connected_handler connected_handler;

  // keep-alive related
  callback_timer* keep_alive_timer;
  ntime last_recv_message_time;

  // helpers to handle initialization etc

  void really_send_message(string msg) {
    out_msg_queue.push_back(msg + "\n");
    DEBUG << "tribe message / send: '" << msg << "'";
    want_write(true);
  }

  void disconnect() {
    state = DISCONNECTED;
    want_read(false);
    ::close(sock.get_fd());
    close();
    if (disconnected_handler) disconnected_handler(this); // callback may delete this object
  }

  void send_init_message() {
    really_send_message(TRIBE_MESSAGE_INIT_MESSAGE_PRE);
  }

  void recv_message(string msg) {
    last_recv_message_time = ntime::now();
    DEBUG << "tribe message / recv: '" << msg << "'";

    if (msg == TRIBE_MESSAGE_KEEP_ALIVE) {
	// ignore keep alive messages (just keep track of when they got here)
    } else if (state == ACTIVE) {      
      if (msg == TRIBE_MESSAGE_INIT_MESSAGE_PRE) {
	// if this is the _PRE message send a _POST
	really_send_message(TRIBE_MESSAGE_INIT_MESSAGE_POST);
      } else if (msg == TRIBE_MESSAGE_INIT_MESSAGE_POST) {
	// if this is a _POST message, switch to the CONNECTED state
	state = CONNECTED;
	if (recv_handler == NULL) want_read(false); // don't care about any more reads
	if (connected_handler != NULL) connected_handler(this);
      } else {
	// otherwise disconnect! unexpected message.
	disconnect();
      }
    } else if (recv_handler != NULL) { // just pass message on to receive handler    
      recv_handler(this, msg);
    }
  }

  void keep_alive_timer_schedule() {
    keep_alive_timer= (new callback_timer
		      (io_loop, wrap(this, &tribe_message_iohandler::keep_alive_timer_fire)));
    keep_alive_timer->arm(ntime::now_plus_msecs(TRIBE_KEEP_ALIVE_PERIOD_MSEC));
  }

  void keep_alive_timer_destroy() {
    if (keep_alive_timer != NULL) {
      keep_alive_timer->disarm();
      keep_alive_timer = NULL;
    }
  }

  void keep_alive_timer_fire() {   
    // if i haven't received a message in a while, timeout and disconnect
    int delta = (ntime::now() - last_recv_message_time).to_msecs();
    if (delta > TRIBE_KEEP_ALIVE_TIMEOUT_MSEC) {      
      disconnect();                 // disconnect message handler
      keep_alive_timer_destroy();   // destory timer
    } else {
      // otherwise send a keep-alive message and run this timer again
      really_send_message(TRIBE_MESSAGE_KEEP_ALIVE);
      keep_alive_timer_schedule();
    }
  }
  
public:
  tribe_message_iohandler(tcpsocket sock, 
			  message_receive_handler hrecv=NULL, 
			  message_connected_handler hconnected=NULL,
			  message_disconnected_handler hdisconnect=NULL) : 
    io_handler(io_loop, sock), sock(sock), state(ACTIVE), 
    read_buf_used(0), out_msg_queue(), part_written(0), 
    recv_handler(hrecv), disconnected_handler(hdisconnect), connected_handler(hconnected), 
    keep_alive_timer(NULL), last_recv_message_time(ntime::now_plus_msecs(2 * TRIBE_KEEP_ALIVE_PERIOD_MSEC)) {
    get_ioh().set_blocking(false);
    want_read(true);
    want_write(false);

    peer_host = ((inet_address)sock.getpeername()).get_host();

    send_init_message(); // try to move from ACTIVE to CONNECTED state
    keep_alive_timer_schedule();
  }  

  ~tribe_message_iohandler() {
    keep_alive_timer_destroy();
    ::close(sock.get_fd());
    close();
  }

  /// send the string as a message (msg shouldn't contain any newlines) and return
  /// true, or return false in case of failure.
  bool send_message(string msg) {
    if (is_connected() && msg.length() < MESSAGE_MAXLEN) {
      really_send_message(msg);
      return true;
    } else {
      return false;
    }
  }

  tcpsocket get_socket() { return sock; }

  string get_peer_host() { return peer_host; }

  bool is_connected() { return (state == CONNECTED); }
protected:
  void ravail() {
    iohandle ioh = get_ioh();
    int n;
    char data;
    while ((n = ioh.read(&data, 1)) > 0) {
      // append char to read buffer...      
      //if (data != '\n') {
      read_buffer[read_buf_used] = data;
      read_buf_used++;
      //}
      
      // do we have a complete line? 
      // or is there going to be a buffer overflow...
      //   equiv: is this char a newline?
      if ((data == '\n') || (read_buf_used >= READ_BUF_MAXLEN-1)) {
	read_buffer[read_buf_used] = '\0'; // make sure null terminated
	if (data == '\n')  read_buffer[read_buf_used-1] = '\0'; // dump newline      
       
	recv_message(string(read_buffer));

	read_buf_used = 0;
      }
    }
    
    // handle disconnection...
    if (n == 0) {
      disconnect();
    } else if ((n < 0) && (errno != EAGAIN)) {
      INFO << "message socket read error: " << strerror(errno);
      disconnect();
    }
  }

  void wavail() {
    for (int i = 0; i < 10; i++) {

      // handle empty queue case.
      if (out_msg_queue.empty()) {
	want_write(false);
	return;
      }

      // get head of message queue
      string msg = out_msg_queue.front();
      int len = msg.length();
      if (part_written < len) { // if still something to write...
	// write it...
	int remain = len - part_written;
	int n = get_ioh().write(msg.c_str() + part_written, remain);
	
	if (n == remain) {  // we done with this message?
	  part_written = 0;
	  out_msg_queue.pop_front();
	  
	} else if ((n > 0) && (n < remain)) { // partial write
	  part_written += n;
	  return; // next wavail will continue writing
	  
	} else if ((n < 0) && (errno == EAGAIN)) { // error, next wavail will handle this...
	  return; 
	  
	} else { // some other error, close connection...
	  part_written = 0;
	  
	  disconnect();
	}
      }
    }
  }
};

// acceptor
class tribe_message_acceptor : public io_handler {
private:
  // event handlers
  message_receive_handler recv_handler;
  message_disconnected_handler disconnected_handler;
  message_connected_handler connected_handler;
  accept_handler haccept;

  // socket
  tcpsocket      sock;

  void make_handler(tcpsocket& client) {
    tribe_message_iohandler *newh =  new tribe_message_iohandler(client, recv_handler, connected_handler, disconnected_handler);
    if (haccept != NULL) haccept(newh); // someone might be interested that a new handler was just created
  }

public:
  tribe_message_acceptor(inet_address bind_address, 
			 accept_handler accept_handler=NULL,
			  message_receive_handler hrecv=NULL, 
			  message_connected_handler hconnected=NULL,
			  message_disconnected_handler hdisconnect=NULL) :
    io_handler(io_loop),
    recv_handler(hrecv), disconnected_handler(hdisconnect), connected_handler(hconnected),
    haccept(accept_handler), 
    sock(bind_address, tcpsocket::acceptor) { 
    set_ioh(sock);
    sock.set_blocking(false);
    want_read(true);
    sock.listen(1);
    DEBUG << "message-protocol is accepting connections on " << bind_address;
  }

protected:
  void ravail() {
    address a;
    tcpsocket client = sock.accept(a);
    if (client) {
      INFO << "message-protocol accepted new connection from " << client.getpeername();
      make_handler(client);
    }
  }
};

#endif
