#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <string>
#include <vector>
using namespace std;

#include "util.h"

//
// spawn stuff
//

int spawn(char *command, char* argv[], char* envp[], int redirect_fd) {
  pid_t child = fork();
  
  if (child == 0) { 
    // (in child)    
    if (redirect_fd) {
      close(1);
      close(2);
      dup(redirect_fd);
      dup(redirect_fd);
    }
    
    int ret = execve(command, argv, envp); // run the command.

    if (redirect_fd) {
      close(1);
      close(2);
      close(redirect_fd);
    }
  
    WARN << "spawn: failed to exec '" << strerror(errno) << "' [" << errno << "]";    
    ::exit(0);
  } else if (child > 0) { 
    // (in parent)
    return child;
  } else { 
    // (in parent)
    WARN << "spawn: failed to fork '" << strerror(errno) << "' [" << errno << "]";
    return -1; 
  }
}

int spawn(string& command) {
  INFO << "running command: '" << command << "'";
  char **argv = split_words(command);
  char **envp = {NULL};

  int ret = spawn(argv[0], argv, envp, 0);
  free_words(argv);

  //int ret = system(command.c_str());

  return ret;
}

char** split_words(string& src) {
  // parse string, splitting it on space boundaries
  vector<string> elements;
  int n = 0;
  string current="";
  bool in_space = true;
  for (int i = 0; i < src.size(); i++) {
    bool is_whitespace = (src[i] == ' ') || (src[i] == '\t') || (src[i] == '\n') || (src[i] == '\r');
    if (is_whitespace && !in_space) { // first space in a space-block
      in_space = true;
      elements.push_back(current);
      current = "";
    } else {
      current += src[i];

      if (in_space) { // space-block just ended
	in_space = false;
      }
    }
  }
  if (!in_space && (current != "")) {
    elements.push_back(current);
  }

  // convert to a char[]
  n = elements.size();
  char** argv = (char**)calloc(sizeof(char*), n+1);
  for (int i = 0; i < n; i++) {
    size_t sz = elements[i].size()+1;
    argv[i] = new char[sz];
    strncpy(argv[i], elements[i].c_str(), sz);    
  }
  argv[n] = NULL;

  return argv;
}

int count_words(char** x) {
  int n = 0;
  while (x[n] != NULL) n++;
  return n;
}

vector<string> parse_list(const char* liststr, char separator) {
  vector<string> list;

  char* str = strdup(liststr);
  
  char *comma = NULL;
  char *remains = str;
  while ((comma = index(remains, separator)) != NULL) {
    // element := remains to comma-1
    *comma = '\0';
    //DEBUG << "parse_list: adding '" << remains << "'" << (unsigned long)remains;
    list.push_back(string(remains));
    
    // remains := everything after this comma
    remains = comma+1;
  }

  if (strlen(remains) > 0) list.push_back(string(remains));

  //DEBUG << "parse_list: " << list.size() << " elements.";

  free(str);
  return list;
}

void free_words(char** words) {
  char **x = words;
  while (*x != NULL) {
    delete [] *x;
    x++;
  }

  free(words);
}

string to_escaped_str(const unsigned char* data, size_t length) {
  string ret="";
  char cbuf[8];
  for (int i = 0; i < length; i++) {
    unsigned char ch = data[i];
    if ((ch < 32) || (ch > 126)) {
      if (ch == '\r') {
	cbuf[0] = '\\';
	cbuf[1] = 'r';
	cbuf[2] = '\0';
      } else if (ch == '\n') {
	cbuf[0] = '\\';
	cbuf[1] = 'n';
	cbuf[2] = '\0';
      } else if (ch == '\t') {
	cbuf[0] = '\\';
	cbuf[1] = 't';
	cbuf[2] = '\0';
      } else {
	snprintf(cbuf, 8, "\\%02x", (int)ch);      
      }
    } else {
      cbuf[0] = ch;
      cbuf[1] = '\0';
    }
    ret += cbuf;
  }
  return ret;
}

//
// command line arguments
//
argument_table_element END_OF_TABLE = { ARG_TYPE_NONE, NULL, NULL, false, false, NULL };

void arguments_print_usage(const char* progname, argument_table_element args[]) {
  int nargs = 0, nrequired = 0;
  string required, optional;
  argument_table_element *parg = args;
  while (parg->name != NULL) {
    if (parg->required) {
      required += "--";
      required += parg->name;
      if (parg->type != ARG_TYPE_BOOLEAN) required += "=...";
      required += " ";
      nrequired++;
    } else {
      optional += "--";
      optional += parg->name;
      if (parg->type != ARG_TYPE_BOOLEAN)  optional += "=...";
      optional += " ";
    }
    nargs++;
    parg++;
  }

  tty_bold_on(); printf("Usage:");tty_bold_off();
  printf("\n    ");
  if (nargs == nrequired) 
    printf("%s %s\n", progname, required.c_str());
  else {
    printf("%s %s [ %s]\n", progname, required.c_str(), optional.c_str());
  }  
}

void print_word_wrapped(char* str, int width, int indent=0) {
  string src(str);
  char** words = split_words(src);

  if (indent) for (int i = 0; i < indent; i++) printf(" ");
  
  int cw = 0;
  char **w = words;
  while (*w != NULL) {
    int n = strlen(*w);
    if (cw+n > width) { 
      printf("\n"); cw = 0; 
      if (indent) for (int i = 0; i < indent; i++) printf(" ");
    }
    printf("%s ", *w);
    cw += n;

    w++;
  }
  printf("\n");

  free_words(words);
}

void arguments_print_help(const char* progname, argument_table_element args[], bool log_level) {
  // display usage 
  arguments_print_usage(progname, args);

  // print more detailed help  
  printf("\nCommand line argument help:\n\n");
  argument_table_element *parg = args;
  while (parg->name != NULL) {
    string title;
    title += "--";
    title += parg->name;
    if (parg->type != ARG_TYPE_BOOLEAN) title += "=...";
    if (!parg->required) title += " [optional]";  

    tty_bold_on();
    printf("%s", title.c_str());
    tty_bold_off();
    printf("\n");

    print_word_wrapped(parg->help, 60, 4);
    printf("\n");

    parg++;
  }

  // print info about log level
  if (log_level) {
    tty_bold_on();
    printf("LOG_LEVEL\n");
    tty_bold_off();
    
    print_word_wrapped("Turn on debugging messages by using the following bash command:", 60, 4);
    print_word_wrapped("$ export LOG_LEVEL=2", 60, 4);
  }

  printf("\n");
}

void parse_argv(char *argv, char** name, char** val) {
  argv += 2; // ignore leading '--'

  *name = argv;
  char *p = argv;
  while (*p != '\0') {
    if (*p == '=') {
      *p = '\0';
      p++;
      break;
    }
    p++;
  }
  *val = p;
}

bool arguments_parse(char** argv, int nargv, argument_table_element args[]) {
  // reset parsed flag for all
  argument_table_element *parg = args;
  while (parg->name != NULL) {
    parg->parsed = false;
    parg++;
  }

  // parse arguments
  for (int i = 0; i < nargv; i++) {
    if ((argv[i][0] == '-') && (argv[i][0] == '-')) {
      char *argname; 
      char* argval;
      parse_argv(argv[i], &argname, &argval);
            
      bool found=false;
      parg = args;
      while (parg->name != NULL) {
	if (strcmp(argname, parg->name) == 0) {
	  if (parg->type == ARG_TYPE_BOOLEAN) {
	    bool* pval = (bool*)parg->var;
	    *pval = true;
	  } else if (parg->type == ARG_TYPE_INTEGER) {
	    int* pval = (int*)parg->var;
	    *pval = atoi(argval);
	  } else if (parg->type == ARG_TYPE_STRING) {
	    char **pval = (char**)parg->var;
	    *pval = argval;
	  }

	  DEBUG << "parsed argv[" << i << "]: '" << argv[i] << "' as: " << argname << "=" << argval;

	  parg->parsed = true;
	  found = true;
	  break;
	}
	parg++;
      }
      if (!found) WARN << "unknown command line argument: " << argv[i];
    } else {
      DEBUG << "ignoring argv[" << i << "]: '" << argv[i] << "'";
    }
  }

  // see if all required arguments were parsed
  parg = args;
  while (parg->name != NULL) {
    if (parg->required && !parg->parsed) return false;
    parg++;
  }
  return true;
}
