#ifndef __NMSTL_RAW_SOCKIO__
#define __NMSTL_RAW_SOCKIO__

#include <sys/socket.h>	
#include <netinet/in.h>	
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>

#include <cstring>

#include <nmstl/netioevent>
#include <ext/hash_set>

using namespace std;
using namespace __gnu_cxx; // for hash_set..

namespace nmstl {
  /**
   * An basic I/O handler for use with IP based
   * (unbound) raw sockets. All headers are
   * included in received packets. Non-blocking,
   * by default.
   */
  class raw_iohandler : public io_handler {
  protected:
    enum protocol_t { RAW=IPPROTO_RAW, UDP=IPPROTO_UDP, TCP=IPPROTO_TCP, ICMP=IPPROTO_ICMP };

    const char* get_proto_string(protocol_t p) {
      switch (p) {
      case RAW:
	return "RAW";
      case UDP:
	return "UDP";
      case TCP:
	return "TCP";
      case ICMP:
	return "ICMP";
      default:
	return "??";
      }
    }
  
  private:
    int         sockfd;
    iohandle    sock_handle;
    protocol_t  protocol;
  public:
    typedef iphdr   header_ip;
    typedef icmphdr header_icmp;
    typedef tcphdr  header_tcp;
    typedef udphdr  header_udp;
  protected:    
    void set_blocking(bool blocking) {
      sock_handle.set_blocking(blocking);
    }
    
    virtual void wavail() {
      want_write(false);
    }

    int getsockfd() { return sockfd; }

    raw_iohandler(io_event_loop& ioloop,
		  protocol_t sock_protocol,
		  inet_address bind_address,
		  bool reading_cb=true, 
		  bool writing_cb=true,
		  bool block=false) : io_handler(ioloop), protocol(sock_protocol) {
      // open the socket...
      sockfd=::socket(AF_INET, SOCK_RAW, protocol);
      
      if (sockfd < 0) {
	CRIT << "Could not create raw socket: " << sockfd;
	CRIT << "Make sure this process has sufficient privileges to create raw sockets.";
	return;
      }

      // bind it
      if (::bind(sockfd, bind_address.data(), bind_address.length()) < 0) {
	WARN << "could not bind socket to: " << bind_address;
      }
      
      // make sure headers are passed to us by the kernel...
      {
	  int dummy = 1;
	  const int *pdummy = &dummy;
	  if (::setsockopt (sockfd, IPPROTO_IP, IP_HDRINCL, pdummy, sizeof (dummy)) < 0)
	    WARN << "cannot set HDRINCL on raw socket.";
      }
            
      DEBUG << "raw socket created:" << get_proto_string(protocol) << ", bind address: " << bind_address;
      
      // create an iohandle for the socket
      sock_handle = iohandle(sockfd);
      set_loop(ioloop);
      set_ioh(sock_handle);
      
      // blocking/non-blocking
      sock_handle.set_blocking(block);    
      
      // set up which callbacks are called when things are ready to happen
      want_read(reading_cb);
      want_write(writing_cb);
    }
  public:
    bool is_valid() { return sockfd >= 0; }

    // closes the associated socket...
    void close() {
      if (is_valid()) 
	::close(sockfd);
    }

    // sends an IP packet
    ssize_t send(dynbuf& ip_packet, int offset=4) {
      // TODO: don't know if this will work...
      //return ::send(sockfd, ip_packet.data(), ip_packet.length(), 0);      

      char *pdata = ip_packet.offset_data() + offset;
      header_ip *ip_h = (header_ip*)(pdata);
      int nbytes = ip_packet.offset_length() - offset;

      struct sockaddr_in addr;
      addr.sin_family = AF_INET;
      addr.sin_addr.s_addr = ip_h->daddr;
      addr.sin_port = 0; // ip_h->protocol;
      
      return ::sendto(sockfd, pdata, nbytes, 0, (struct sockaddr*)&addr, sizeof(struct sockaddr_in));
    }

    static u_int32_t address_nmstl_to_int32(const inet_address& addr) {
      // ugh. inefficient hack: nmstl -> string -> int32
      string addrstr = addr.get_host();
      u_int32_t addrint = inet_addr(addrstr.c_str());
      return addrint;
    }

    static string address_int32_to_string(u_int32_t addr) {
      //DEBUG << "enter: address_int32_to_string : " << addr;
      
      struct in_addr saddr;
      saddr.s_addr = addr;
      //swab(&addr, &(saddr.s_addr), 4);

      char buf[INET_ADDRSTRLEN+1];
      inet_ntop(AF_INET, &(saddr), buf, INET_ADDRSTRLEN);
      //DEBUG << "exit: address_int32_to_string : " << buf;
      return string(buf);
    }

    // produces a string, useful for debugging.
    static string packet_header_to_string(dynbuf& buf) {
      return packet_header_to_string(buf.offset_data(), buf.offset_length());
    }

    static int swab_short(short x) {
      int ret = 0;
      swab(&x, &ret, sizeof(short));
      return ret;
    }

    static int swab_int(int x) {
      int ret = 0;
      swab(&x, &ret, sizeof(int));
      return ret;
    }

    static string packet_header_to_string(char *pdata, size_t data_length) {
      //DEBUG << "enter: packet_header_to_string";

      pdata += 4;

      const int WORK_BUF_LEN = 1024;
      char workbuf[WORK_BUF_LEN];

      string retval("IP ");      

      // IP header     
      header_ip* ip_h = (header_ip*)pdata;

      snprintf(workbuf, WORK_BUF_LEN,
	       "[id=%d, src=%s, dest=%s, protocol=%d, tot_len=%d, ttl=%d]", 
	       (int)(swab_short(ip_h->id)), 
	       address_int32_to_string(ip_h->saddr).c_str(),
	       address_int32_to_string(ip_h->daddr).c_str(),
	       (int)(ip_h->protocol),
	       (int)(swab_short(ip_h->tot_len)),
	       (int)(ip_h->ttl));
      retval.append(workbuf);
      //DEBUG << "checkpoint: " << retval;
      
      if (ip_h->protocol == IPPROTO_TCP) {
	header_tcp* tcp_h = (header_tcp*)(pdata + sizeof(header_ip));

	snprintf(workbuf, WORK_BUF_LEN, 
		 ", TCP [seq=%d, port-src=%d, port-dst=%d",
		 (int)(swab_int(tcp_h->seq)),
		 (int)(swab_short(tcp_h->source)),
		 (int)(swab_short(tcp_h->dest)));

	retval.append(workbuf);

	if (tcp_h->urg) retval.append(", URG");
	if (tcp_h->ack) retval.append(", ACK");
	if (tcp_h->psh) retval.append(", PSH");
	if (tcp_h->rst) retval.append(", RST");
	if (tcp_h->syn) retval.append(", SYN");
	if (tcp_h->fin) retval.append(", FIN");
	
	retval.append("]");
	
      } else if (ip_h->protocol == IPPROTO_UDP) {
	header_udp* udp_h = (header_udp*)(pdata + sizeof(header_ip));

	snprintf(workbuf, WORK_BUF_LEN, 
		 ", UDP [port-src=%d, port-dst=%d]",
		 (int)(swab_short(udp_h->source)),
		 (int)(swab_short(udp_h->dest)));
	retval.append(workbuf);

      } else if (ip_h->protocol == IPPROTO_ICMP) {
	header_icmp* icmp_h = (header_icmp*)(pdata + sizeof(header_ip));
	snprintf(workbuf, WORK_BUF_LEN, 
		 ", ICMP [echo_seq=%d]", (int)(icmp_h->un.echo.sequence));
	retval.append(workbuf);
      } else { // unknown protocol
	//retval.append("");
      }
      
      snprintf(workbuf, WORK_BUF_LEN,
	       ", len=%d", 
	       (int)data_length);
      retval.append(workbuf); 

      //DEBUG << "exit: packet_header_to_string : " << retval;
      return retval;
    }
  };

  // By default all packets are filtered, unless they are
  // destined to an IP address that is 'allowed'. Packets
  // that pass through the filter are passed on, IP headers
  // and all, to the implementation of handle_packet().
  class filtered_raw_iohandler : public raw_iohandler {
  private:
    hash_set<u_int32_t> allowed_dests;

  public:
    // adds a destination that will not be filtered out.
    void add_allowed_dest(const inet_address& dest_addr) {
      allowed_dests.insert(address_nmstl_to_int32(dest_addr));
    }

    // removes a destination from the set of allowed dests.  
    void remove_allowed_dest(const inet_address& dest_addr) {
      hash_set<u_int32_t>::iterator iter = allowed_dests.find(address_nmstl_to_int32(dest_addr)); 
      if (iter != allowed_dests.end()) 
	allowed_dests.erase(iter);
    }

    bool is_allowed_dest(const inet_address& dest_addr) {
      u_int32_t addr = address_nmstl_to_int32(dest_addr);
      return (allowed_dests.find(addr) != allowed_dests.end());
    }

  protected:
    bool is_allowed_dest(u_int32_t addr) {     
      bool allowed = (allowed_dests.find(addr) != allowed_dests.end());      
      return allowed;
    }

    // packet read buffer's maximum length.
    static const unsigned int READ_BUF_MAXLEN = 2048;    

    void ravail() {
      //DEBUG << "ravail() called.";

      // read the packet.
      dynbuf buf(READ_BUF_MAXLEN);
      char *read_buf = buf.data() + 4;
      ssize_t len = ::recv(getsockfd(), read_buf, READ_BUF_MAXLEN - 4, 0);

      //DEBUG << "recv: " << len;

      if (len > 0) { // no error
	buf.set_length(len + 4); // size buffer properly.
	short *pdata = (short*)buf.data();
	pdata[0] = 0x0000;
	pdata[1] = 0x0008;

	// do filtering:
	//    if header IP matches an allowed dest, 
        //    pass it on to handle_packet(...)
        iphdr* hdr = (iphdr*)read_buf;
	
	if (is_allowed_dest(hdr->daddr)) { // pass packet along...
	  //DEBUG << "calling handle_packet";
	  handle_packet(buf);
	}
	else {
	  DEBUG(3) << "dropping: " << packet_header_to_string(buf);
	}
      }      
    }

    virtual void handle_packet(dynbuf& ip_packet) = 0;

    filtered_raw_iohandler(io_event_loop& ioloop,
			   protocol_t sock_protocol,
			   inet_address& bind_address,
			   bool reading_cb=true, bool writing_cb=true, bool block=false) 
      : raw_iohandler(ioloop, sock_protocol, bind_address, reading_cb, writing_cb, block), allowed_dests() 
      { /* don't need to do anything else  */ }
  };
}

#endif
