#include <stdlib.h>
#include <time.h>
#include <string>
#include <nmstl/ioevent>
#include <nmstl/net>

#include "util.h"
#include "tribe_msg.h"
#include "tunnel.h"
#include "poolman.h"
#include "ifmanager.h"

using namespace std;
using namespace nmstl;

void event_connected(tribe_message_iohandler* src);
void event_disconnected(tribe_message_iohandler* src);
void event_recv_msg(tribe_message_iohandler* src, string msg);
void event_recv_tunnel(dynbuf& packet, inet_address& addr);
void event_ifmanager_exited();
void event_iface(string iface, inet_address addr, bool up);

void event_stolen_pkt_recv(unsigned char* data, size_t len);

//
// packet stealer declarations
//
bool packet_stealer_start(void (*packet_recv_cb)(unsigned char* /*data*/, size_t /*len*/), const char* proto_list);
void packet_stealer_stop();
void packet_stealer_ifup(string iface);
void packet_stealer_ifdown(string iface);

string packet_header_to_string(char *pdata, size_t data_length, int offset=4);

//
// global variables
//

io_event_loop io_loop;

inet_address leader_address;
inet_address bind_address;

inet_address msg_leader_address;
inet_address msg_bind_address;
inet_address tunnel_bind_address;
inet_address tunnel_leader_address;
vector<string> managed_interfaces;
vector<string> monitored_interfaces;

interface_pool_manager  *poolman = NULL;
tunnel_iohandler        *tunnel_ioh = NULL;
tribe_message_iohandler *msg_ioh = NULL;

// command line stuff

char *ifaces = "";
char *if_linux = "ppp0,ppp1,ppp2,ppp3";
char *protocols="*";
int retry_sleep = 10;
char *leader = NULL;
char *bind_addr = NULL;
int tunnel_port = TUNNEL_DEFAULT_PORT;
int message_port = MESSAGE_DEFAULT_PORT;

//bool disable_tcp = false;
//bool disable_udp = false;
//bool disable_icmp = false;

bool enable_reboot = false;

int if_max_retry = 0;
int if_retry_reset_period = 600;
int if_retry_sleep = 10;
bool if_reset_all = false;

//
// command line arguments
//
bool help_requested = false;

argument_table_element argument_table[] = {
  { ARG_TYPE_BOOLEAN, "help", &help_requested, false, false, "Print out this help." },
  { ARG_TYPE_STRING, "ifaces", &ifaces, false, false, 
    "A list of interfaces to manage. If there are multiple interfaces in the list, the "
    "interface ID's should be separated by commas but with no spaces. "
    "These are the kind of interface identifiers used by ifmanager. "
    "For example: '--ifaces=vzw_5220,vzw_kpc650'. See the ifmanager help for more info." },

  { ARG_TYPE_STRING, "if_linux", &if_linux, false, false, 
    "A list of linux interfaces to monitor."
    "For example: '--if_linux=ppp0,ppp1,ppp2'." },

  { ARG_TYPE_STRING, "protocols", &protocols, false, false, 
    "The IPv4 transport protocols to manage on the given linux interfaces."
    "For example: '--protocols=tcp,udp,icmp'. '*' denotes all protocols (default)."},

  { ARG_TYPE_STRING, "leader", &leader, true, false, 
    "The address of the leader, e.g. '128.30.76.73' or 'slice.csail.mit.edu'"},

  { ARG_TYPE_STRING, "bind", &bind_addr, true, false, 
    "The local IP address to bind to."},

  { ARG_TYPE_INTEGER, "tunnel", &tunnel_port, false, false, 
    "Port used for IP-in-UDP tunneling between leaders and followers." },

  { ARG_TYPE_INTEGER, "message", &message_port, false, false, 
    "Port used for message passing between leaders and followers."},

  { ARG_TYPE_INTEGER, "retry-sleep", &retry_sleep, false, false,
    "Delay in seconds between  attempts to connect to Tribe leader."
    "Defaults to 10."},

  // { ARG_TYPE_BOOLEAN, "disable-udp", &disable_udp, false, false, "Disable UDP forwarding."},
  // { ARG_TYPE_BOOLEAN, "disable-tcp", &disable_tcp, false, false, "Disable TCP forwarding."},
  // { ARG_TYPE_BOOLEAN, "disable-icmp", &disable_icmp, false, false, "Disable ICMP forwarding."},

  { ARG_TYPE_BOOLEAN, "enable-reboot", &enable_reboot, false, false, "Enable remote reboots."},

  { ARG_TYPE_INTEGER, "if-max-retry", &if_max_retry, false, false, 
    "For spawned ifmanager: Maximum number of reconnection attempts during "
    "a long period of time. If 0, then an infinite number of retries are assumed. "
    "Default is 0." },
  { ARG_TYPE_INTEGER, "if-retry-reset-period", &if_retry_reset_period, false, false,
    "For spawned ifmanager: "
    "Period in seconds used by the max-retry logic. After this many seconds, the "
    "reconnection counter is reset to zero. Default is 600."},
  { ARG_TYPE_INTEGER, "if-retry-sleep", &if_retry_sleep, false, false,
    "For spawned ifmanager: "
    "Delay in seconds between a disconnection and the attempt to reconnect. "
    "Defaults to 10."},
  { ARG_TYPE_BOOLEAN, "if-reset-all", &if_reset_all, false, false, 
    "For spawned ifmanager: "
    "Whether or not 'scripts/reset_all' should be called before each reconnection attempt. "},
  END_OF_TABLE
};

//
// startup and shutdown
//
void shutdown() {           // shuts down everything but msg_ioh
  INFO << "shutting down...";
  
  if (poolman) {
    poolman->stop();
    delete poolman;
    poolman  = NULL;
  }

  if (tunnel_ioh) {
    tunnel_ioh->close();
    delete tunnel_ioh;
    tunnel_ioh = NULL;
  }

  packet_stealer_stop();

  ifmanagers_kill_all();   // kill any alive ifmanagers...
}

void startup() {            // assumes msg_ioh is up and starts up everything else
  INFO << "starting up...";

  // tunnel
  tunnel_ioh = new tunnel_iohandler(tunnel_bind_address, tunnel_leader_address, event_recv_tunnel);

  // packet stealer
  if (!packet_stealer_start(event_stolen_pkt_recv, protocols)) {
    CRIT << "could not start NETFILTER packet stealer! you will not receive any packets.";
  }

  // poolman
  poolman = new interface_pool_manager(monitored_interfaces, POOLMAN_SLEEP_INTERVAL, event_iface);
  poolman->start();

  // start up all ifmanagers
  ifmanagers_start_all(managed_interfaces, event_ifmanager_exited);
  // if start_all fails part-way, it'll call kill_all
}

void create_message_ioh() { // assumes msg_ioh is NULL
  tcpsocket leadersock;
  int i = 1;
  
  while(true) {
    INFO << "attempting to connect to " << leader_address << ", attempt " << i++;
    leadersock = tcpsocket(iohandle(::socket(AF_INET, SOCK_STREAM, 0)));
    status stat = leadersock.bind(msg_bind_address);
    if (stat) {
      stat = leadersock.connect(msg_leader_address);
      if (!stat) {
	INFO << "TCP connect failed: " << stat;
      } 
    } else {
      WARN << "TCP socket bind failed: " << stat;
    }
    
    if (leadersock.stat()) break;    
    else {
      // sleep for a bit
      timespec sleep_period = { retry_sleep, 0 };
      nanosleep(&sleep_period, NULL);
    }
  }  
  
  // only reach here if the tcp socket is connected
  INFO << "tcp socket connected, starting messaging...";
  msg_ioh = new tribe_message_iohandler(leadersock, event_recv_msg, event_connected, event_disconnected);
}
    
void destroy_message_ioh() {
  if (msg_ioh) {
    delete msg_ioh;
    msg_ioh = NULL;
  }
}

void restart() {
  destroy_message_ioh();
  shutdown();
  
  create_message_ioh();
  // message ioh's event_connected will call startup()
}

//
// event handlers
//

void event_connected(tribe_message_iohandler* src) {
  INFO << "connected to server: " << src->get_socket().getpeername();

  startup();
}

void event_disconnected(tribe_message_iohandler* src) {
  WARN << "disconnected from server.";

  restart();
}

void event_recv_msg(tribe_message_iohandler* src, string msg) {
  if (msg == "REBOOT") {
    CRIT << "received remote REBOOT request.";
    shutdown(); // kills all ifmanagers
    exit(0);
  } else {
    INFO << "unexpected message from leader '" << msg << "'";
  }
}

void event_recv_tunnel(dynbuf& packet, inet_address& addr) {
  if (poolman) poolman->send_raw(packet);
}

void event_stolen_pkt_recv(unsigned char* data, size_t len) {
  if (tunnel_ioh) {
    dynbuf pkt(len+16);           // TODO get rid of this stoopid copy etc...
    pkt.set_length(len + 4);
    short *hdr = (short*)pkt.data();
    hdr[0] = 0x0000; hdr[1] = 0x0008;
    memcpy(pkt.data()+4, data, len);

    DEBUG(3) << "stolen_pkt_recv: " << packet_header_to_string(pkt.data(), pkt.length(), 4);
    
    tunnel_ioh->tunnel_send(pkt);
  }
}

void event_ifmanager_exited() {
  CRIT << "ifmanager exited. exiting follower.";
  shutdown(); // kills all active ifmanagers
  exit(EXIT_FAILURE);  
}

void event_iface(string iface, inet_address addr, bool up) {
  // tell packet stealer to react to this...
  if (up) packet_stealer_ifup(iface);     
  else packet_stealer_ifdown(iface);

  // notify my peer about this interface event
  string msg = "";
  if (up) msg += TRIBE_MESSAGE_IFACE_UP; else msg += TRIBE_MESSAGE_IFACE_DOWN;
  msg += " ";
  msg += iface;
  msg += " ";
  msg += addr.as_string();
  
  if (msg_ioh) msg_ioh->send_message(msg);
}

//
// signal handler
//
static void handle_signal(int s) {
  string sigstr;
  switch (s) {
  case SIGINT: sigstr = "SIGINT"; break;
  case SIGTERM: sigstr = "SIGTERM"; break;
  case SIGHUP: sigstr = "SIGHUP"; break;
  case SIGSEGV: sigstr = "SIGSEGV"; break;
  case SIGKILL: sigstr = "SIGKILL"; break;
  default: sigstr = "#" + s;
  }

  INFO << "received signal: " << sigstr;
  if (s==SIGINT || s==SIGTERM || s==SIGHUP || s==SIGKILL || s==SIGSEGV) {
    shutdown();
    io_loop.terminate(); // should eventually exit...
  }
}

// 
// main
//
int main (int argc, char** argv) {
  bool ok = arguments_parse(argv, argc, argument_table);

  if (help_requested) { // detailed help was requested
    arguments_print_help(argv[0], argument_table,true);
  } else if (!ok) { // not all requirements were present
    arguments_print_usage(argv[0], argument_table);
  } else {
    // parse command line
    managed_interfaces = parse_list(ifaces);
    monitored_interfaces = parse_list(if_linux);

    leader_address = inet_address(leader);
    bind_address = inet_address(bind_addr);

    msg_leader_address = inet_address(leader_address.get_host(), message_port);
    msg_bind_address = inet_address(bind_address.get_host());
    tunnel_leader_address = inet_address(leader_address.get_host(), tunnel_port);
    tunnel_bind_address = inet_address(bind_address.get_host(), tunnel_port);

    INFO << "binding message protocol: " << msg_bind_address 
	 << " [local] <-> " << msg_leader_address << " [remote]";

    INFO << "binding tunnel protocol: " << tunnel_bind_address 
	 << " [local] <-> " << tunnel_leader_address << " [remote]";

    // set up signal handlers
    signal(SIGINT, handle_signal);
    signal(SIGTERM, handle_signal);
    signal(SIGHUP, handle_signal);
    signal(SIGKILL, handle_signal);
    
    siginterrupt(SIGINT, 1);
    siginterrupt(SIGTERM, 1);
    siginterrupt(SIGHUP, 1);
    siginterrupt(SIGKILL, 1);

    // startup
    create_message_ioh();
    io_loop.run();
    WARN << "I/O loop exited.";
  }
} 
