#include <sys/socket.h>
#include <net/if.h> 		
#include <ifaddrs.h>
#include <netinet/in.h>	
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>

#include <nmstl/net>
#include <nmstl/ntime>

#include "poolman.h"
#include "ifmanager.h"

// --------- raw I/O --------------------
typedef iphdr   header_ip;
typedef icmphdr header_icmp;
typedef tcphdr  header_tcp;
typedef udphdr  header_udp;

u_int32_t address_nmstl_to_int32(const inet_address& addr) {
  // ugh. inefficient hack: nmstl -> string -> int32
  string addrstr = addr.get_host();
  u_int32_t addrint = inet_addr(addrstr.c_str());
  return addrint;
}

string address_int32_to_string(u_int32_t addr) {      
  struct in_addr saddr;
  saddr.s_addr = addr;
  //swab(&addr, &(saddr.s_addr), 4);
  
  char buf[INET_ADDRSTRLEN+1];
  inet_ntop(AF_INET, &(saddr), buf, INET_ADDRSTRLEN);
  return string(buf);
}

int raw_create_sender_socket(inet_address& bind_address) {
  // open the socket...
  int sockfd=::socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
  
  if (sockfd < 0) {
    CRIT << "Could not create raw socket: " << sockfd;
    CRIT << "Make sure this process has sufficient privileges to create raw sockets.";
    return -1;
  }
  
  // bind it
  if (::bind(sockfd, bind_address.data(), bind_address.length()) < 0) {
    WARN << "could not bind socket to: " << bind_address;
  }
  
  // make sure headers are passed to us by the kernel...
  {
    int dummy = 1;
    const int *pdummy = &dummy;
    if (::setsockopt (sockfd, IPPROTO_IP, IP_HDRINCL, pdummy, sizeof (dummy)) < 0)
      WARN << "cannot set HDRINCL on raw socket.";
  }
  
  return sockfd;
}

ssize_t raw_send(int socketfd, dynbuf& ip_packet, int offset=4) {
  char *pdata = ip_packet.offset_data() + offset;
  header_ip *ip_h = (header_ip*)(pdata);
  int nbytes = ip_packet.offset_length() - offset;
  
  struct sockaddr_in addr;
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = ip_h->daddr;
  addr.sin_port = 0; // ip_h->protocol;
  
  return ::sendto(socketfd, pdata, nbytes, 0, (struct sockaddr*)&addr, sizeof(struct sockaddr_in));
}

// --------- utility functions ----------

// bridge between nmstl and socket structs
inet_address parse_sockaddr_in(sockaddr_in* addr) { 
  char buf[INET_ADDRSTRLEN+1];  
  inet_ntop(AF_INET, &(addr->sin_addr), buf, INET_ADDRSTRLEN);

  return inet_address(string(buf), ntohs(addr->sin_port));
}

// looks up an inet4 interface.
// @return true if an interface with the given name was found; false otherwise.
// @param addr if the iface_name was found, copies contents of
//             address of first record that matched to this param.
bool lookup_interface(string iface_name, inet_address& addr) {
  bool retval=false;

  struct ifaddrs *iface, *ifaces;
  getifaddrs(&iface);
  ifaces = iface;

  for (; iface != NULL; iface = iface->ifa_next) {
    if ((iface->ifa_addr != NULL) && (iface->ifa_addr->sa_family == AF_INET)) {
      if (iface_name == iface->ifa_name) { // found it...
	// copy contents
	addr = parse_sockaddr_in((sockaddr_in*)iface->ifa_addr);
	retval=true; 
	break;
      }
    }
  }

  freeifaddrs(ifaces); // avoid mem leak.
  return retval;
}

// --------- start / stop --------------

void
interface_pool_manager::start() {
  if (!running) {
    running = true;
    string ifacestr("");
    for (vector<string>::iterator iter = ifaces.begin(); iter != ifaces.end(); iter++) {
      ifacestr += *iter;
      ifacestr += " ";
    }
    DEBUG << "poolman started, monitoring: " << ifacestr;
    wakeup();
  } // else: wakeup will be called soon, so do nothing...
}

void
interface_pool_manager::stop() {

  // disarm wakeup timer
  if (wakeup_timer != NULL) {
    wakeup_timer->disarm();
    wakeup_timer = NULL;
  }

  // shut down every active interface...
  if (running)
    for (int i = 0; i < ifaces.size(); i++) {
      if (iface_addresses.find(ifaces[i]) != iface_addresses.end()) 
	iface_down(i);
    }
    //    for (int i = 0; i < ifaces.size(); i++)
    //      iface_down(i);

  DEBUG << "poolman stopped.";
  running = false;
}

// ---------- wakeup --------------------
void
interface_pool_manager::wakeup() {
  if (running) {        
    int idx = 0;
    for (vector<string>::const_iterator iter = ifaces.begin();
	 iter != ifaces.end(); iter++, idx++) {
      string iface = *iter;

      // find last round IP of this interface...
      STRING_HASHMAP(inet_address)::iterator if_addr = iface_addresses.find(iface);
      bool was_active = (if_addr != iface_addresses.end());
      
      // is interface up?
      inet_address address;
      if (lookup_interface(iface, address)) {
	// interface is currently up
	if (was_active) { // was this active in last round?
	  // we only care if the address has changed
	  if (if_addr->second != address) { 
	    DEBUG << "old interface changed address, rebooting: " << iface;
	    iface_down(idx);
	    iface_up(idx, address);
	  }
	} else {
	  // newly activated interface...
	  DEBUG << "interface came up: " << iface;
	  iface_up(idx, address);
	}
      } else {
	// interface is not currently up, but was active...
	if (was_active) {
	  DEBUG << "interface went down: " << iface;
	  iface_down(idx); // active in last round, so clean up...
	}
      }     
    }
    
    // reschedule wakeup...
    wakeup_timer = (new callback_timer
		    (io_loop, wrap(this, &interface_pool_manager::wakeup)));

    wakeup_timer->arm(ntime::now_plus_msecs(period_msecs));

    // test whether any ifmanagers have exited
    ifmanagers_poll(); // HACK: this should really be elsewhere...
  }
}

// ------------- iface_up / iface_down ---------------
void
interface_pool_manager::iface_up(int idx, inet_address address) {
  string iface = ifaces[idx];

  // add to set of active addresses
  iface_addresses[iface] = address;

  // create a raw sender socket
  int sock = raw_create_sender_socket(address);
  int intaddr = address_nmstl_to_int32(address);

  iface_iohandlers[iface] = sock;
  addr_to_iohandler_map[intaddr] = sock;   

  // maybe somebody cares about this event
  if (if_event_handler) if_event_handler(iface, address, true);  
}

void
interface_pool_manager::iface_down(int idx) {
  string iface = ifaces[idx];
  //iface_down(iface);
  //}

  //void
  //interface_pool_manager::iface_down(string iface) {
  u_int32_t addr = 0;
  inet_address iface_addr;

  // find address for this interface
  STRING_HASHMAP(inet_address)::iterator if_addr = iface_addresses.find(iface);
  if (if_addr != iface_addresses.end()) { // if found...
    iface_addr = if_addr->second;
    //follower_ignore_address(if_addr->second); // tell socket I/O layer to ignore
    addr = address_nmstl_to_int32(if_addr->second);
    iface_addresses.erase(if_addr); // and, remove from set of active addresses
  }

  // find raw I/O handler for this interface
  STRING_HASHMAP(int)::iterator j = iface_iohandlers.find(iface);
  if (j != iface_iohandlers.end()) { // if found...
    iface_iohandlers.erase(j); // remove
  }

  // remove from addr_to_iohandler_map
  hash_map< int, int >::iterator i = addr_to_iohandler_map.find(addr);
  if (i != addr_to_iohandler_map.end()) { // if found...
    addr_to_iohandler_map.erase(i);//remove
  }

  // tell info server that an iface_down event occurred.
  // maybe somebody cares about this event
  if (if_event_handler) if_event_handler(iface, iface_addr, false);
}

// -------------- send_raw --------------
void
interface_pool_manager::send_raw(dynbuf& packet) {
  iphdr* hdr = (iphdr*)(packet.offset_data() + 4);
  hash_map< int, int >::iterator i = addr_to_iohandler_map.find(hdr->saddr);
  if (i != addr_to_iohandler_map.end()) { // if found...
    raw_send(i->second, packet);
  } else {
    WARN << "ignoring packet with source: " << address_int32_to_string(hdr->saddr);
  }
}
