#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <sys/time.h>
#include <stdarg.h>
#include <stdio.h>

#include <string>
#include <vector>
using namespace std;

#include "util.h"

#define _THREADS 0

//
// logging stuff
//
int log_level = INIT_LOG_LEVEL;
bool log_trace = INIT_LOG_TRACE;
bool log_thread = INIT_LOG_THREAD;
bool log_timestamp = INIT_LOG_TIMESTAMP;

#ifdef _THREADS
pthread_mutex_t log_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

log_msg::log_msg(const char* event_id, const char* file, int line, const char* function) {
#if _THREADS
  pthread_mutex_lock(&lock_mutex);
#endif
  
 if (log_timestamp) {
   timeval t;
   gettimeofday(&t, NULL);
   tm *x = localtime(&t.tv_sec);
   char buffer[64];
   snprintf(buffer, 64, "%02d:%02d:%02d.%3d\t", 
	    x->tm_hour, x->tm_min, x->tm_sec, t.tv_usec/1000);
   msg << buffer;
 }
 
#if _THREADS
 if (log_thread) {
   msg << pthread_self() << "\t";
 }
#endif
 
 msg << "\033[1;1m" << event_id << "\033[0m\t";
 
 if (log_trace) post << " (at " << file << ':' << line << ':' << function << ')';

#if _THREADS
  pthread_mutex_unlock(&lock_mutex);
#endif
}

static char __fmt_buffer[256];
string fmtstr(const char* fmt, ... ) {
  va_list ap;
  va_start (ap, fmt);
  vsnprintf (__fmt_buffer, 256, fmt, ap);
  va_end (ap);
  return string(__fmt_buffer);
}

//
// spawn stuff
//

int spawn(char *command, char* argv[], char* envp[], int redirect_fd) {
  pid_t child = fork();
  
  if (child == 0) { 
    // (in child)    
    if (redirect_fd) {
      close(1);
      close(2);
      dup(redirect_fd);
      dup(redirect_fd);
    }
    
    int ret = execve(command, argv, envp); // run the command.

    if (redirect_fd) {
      close(1);
      close(2);
      close(redirect_fd);
    }
  
    WARN << "spawn: failed to exec '" << strerror(errno) << "' [" << errno << "]";    
    ::exit(0);
  } else if (child > 0) { 
    // (in parent)
    return child;
  } else { 
    // (in parent)
    WARN << "spawn: failed to fork '" << strerror(errno) << "' [" << errno << "]";
    return -1; 
  }
}

int spawn(string command) {
  DEBUG << "running command: '" << command << "'";
  char **argv = split_words(command);
  char **envp = {NULL};

  int ret = spawn(argv[0], argv, envp, 0);
  free_words(argv);

  //int ret = system(command.c_str());

  return ret;
}

char** split_words(string& src) {
  // parse string, splitting it on space boundaries
  vector<string> elements;
  int n = 0;
  string current="";
  bool in_space = true;
  for (int i = 0; i < src.size(); i++) {
    bool is_whitespace = (src[i] == ' ') || (src[i] == '\t') || (src[i] == '\n') || (src[i] == '\r');
    if (is_whitespace && !in_space) { // first space in a space-block
      in_space = true;
      elements.push_back(current);
      current = "";
    } else {
      current += src[i];

      if (in_space) { // space-block just ended
	in_space = false;
      }
    }
  }
  if (!in_space && (current != "")) {
    elements.push_back(current);
  }

  // convert to a char[]
  n = elements.size();
  char** argv = (char**)calloc(sizeof(char*), n+1);
  for (int i = 0; i < n; i++) {
    size_t sz = elements[i].size()+1;
    argv[i] = new char[sz];
    strncpy(argv[i], elements[i].c_str(), sz);    
  }
  argv[n] = NULL;

  return argv;
}

int count_words(char** x) {
  int n = 0;
  while (x[n] != NULL) n++;
  return n;
}

vector<string> parse_list(const char* liststr, char separator) {
  vector<string> list;

  char* str = strdup(liststr);
  
  char *comma = NULL;
  char *remains = str;
  while ((comma = index(remains, separator)) != NULL) {
    // element := remains to comma-1
    *comma = '\0';
    list.push_back(string(remains));
    
    // remains := everything after this comma
    remains = comma+1;
  }

  if (strlen(remains) > 0) list.push_back(string(remains));
  
  free(str);
  return list;
}

void free_words(char** words) {
  char **x = words;
  while (*x != NULL) {
    delete [] *x;
    x++;
  }

  free(words);
}

//
// command line arguments
//
argument_table_element END_OF_TABLE = { ARG_TYPE_NONE, NULL, NULL, false, false, NULL };

void arguments_print_usage(const char* progname, argument_table_element args[], FILE* out) {
  int nargs = 0, nrequired = 0;
  string required, optional, positional, required2;
  argument_table_element *parg = args;
  while (parg->name != NULL) {
    if (parg->required) {
      required += "--";
      required += parg->name;
      if (parg->type != ARG_TYPE_BOOLEAN) required += "=...";
      required += " ";
      nrequired++;

      positional += "<";
      positional += parg->name;
      positional += "> ";
    } else {
      optional += "--";
      optional += parg->name;
      if (parg->type != ARG_TYPE_BOOLEAN)  optional += "=...";
      optional += " ";

      positional += "[";
      positional += parg->name;
      positional += "] ";
    }

    nargs++;
    parg++;
  }

  tty_bold_on(out); fprintf(out,"Usage:");tty_bold_off(out);
  fprintf(out,"\n     ");
  if (nargs == nrequired) 
    fprintf(out,"%s %s\n", progname, required.c_str());
  else
    fprintf(out,"%s %s [ %s]\n", progname, required.c_str(), optional.c_str()); 
  tty_bold_on(out); fprintf(out," or  ");tty_bold_off(out);

  fprintf(out,"%s %s\n", progname, positional.c_str());
}

void print_word_wrapped(char* str, int width, int indent, FILE* out) {
  string src(str);
  char** words = split_words(src);

  if (indent) for (int i = 0; i < indent; i++) fprintf(out," ");
  
  int cw = 0;
  char **w = words;
  while (*w != NULL) {
    int n = strlen(*w);
    if (cw+n > width) { 
      fprintf(out,"\n"); cw = 0; 
      if (indent) for (int i = 0; i < indent; i++) fprintf(out," ");
    }
    fprintf(out,"%s ", *w);
    cw += n;

    w++;
  }
  fprintf(out,"\n");

  free_words(words);
}

void arguments_print_help(const char* progname, argument_table_element args[], FILE* out) {
  // display usage 
  arguments_print_usage(progname, args, out);

  // print more detailed help  
  fprintf(out,"\nCommand line argument help:\n\n");
  argument_table_element *parg = args;
  while (parg->name != NULL) {
    string title;
    title += "--";
    title += parg->name;
    if (parg->type != ARG_TYPE_BOOLEAN) title += "=...";
    if (!parg->required) title += " [optional]";  

    tty_bold_on(out);
    fprintf(out,"%s", title.c_str());
    tty_bold_off(out);
    fprintf(out,"\n");

    print_word_wrapped(parg->help, 60, 4, out);
    fprintf(out,"\n");

    parg++;
  }

  fprintf(out,"\n");
}

void parse_argv(char *argv, char** name, char** val) {
  argv += 2; // ignore leading '--'

  *name = argv;
  char *p = argv;
  while (*p != '\0') {
    if (*p == '=') {
      *p = '\0';
      p++;
      break;
    }
    p++;
  }
  *val = p;
}

argument_table_element* get_first_req_unparsed(argument_table_element args[]) {
  argument_table_element *parg = args;
  while (parg->name != NULL) {
    if (parg->required && !parg->parsed) return parg;
    parg++;
  }
  return NULL;
}

argument_table_element* get_first_unparsed(argument_table_element args[]) {
  argument_table_element *parg = args;
  while (parg->name != NULL) {
    if (!parg->parsed) return parg;
    parg++;
  }
  return NULL;
}

void arg_assign_value(argument_table_element* parg, char* argval) {
  if (parg->type == ARG_TYPE_BOOLEAN) {
    bool* pval = (bool*)parg->var;
    *pval = true;
  } else if (parg->type == ARG_TYPE_INTEGER) {
    int* pval = (int*)parg->var;
    *pval = atoi(argval);
  } else if (parg->type == ARG_TYPE_STRING) {
    char **pval = (char**)parg->var;
    *pval = argval;
  } else if (parg->type == ARG_TYPE_DOUBLE) {
    double *pval = (double*)parg->var;
    *pval = atof(argval);
  }

  parg->parsed=true;
};

bool arguments_parse(char** argv, int nargv, argument_table_element args[]) {
  // reset parsed flag for all
  argument_table_element *parg = args;
  while (parg->name != NULL) {
    parg->parsed = false;
    parg++;
  }

  bool pflags[nargv];
  for (int i = 0; i < nargv; i++) pflags[i] = false;

  // parse arguments
  for (int i = 1; i < nargv; i++) { // look at argv[1:]
    if ((argv[i][0] == '-') && (argv[i][0] == '-')) {
      char *argname; 
      char* argval;
      parse_argv(argv[i], &argname, &argval);
            
      bool found=false;
      parg = args;
      while (parg->name != NULL) {
	if (strcmp(argname, parg->name) == 0) {
	  arg_assign_value(parg, argval);
	  DEBUG << "parsed argv[" << i << "]: '" << argv[i] << "' as: " << argname << "=" << argval;
	  pflags[i] = true;
	  found = true;
	  break;
	}
	parg++;
      }
      if (!found) WARN << "unknown command line argument: " << argv[i];
    }
  }

  // if any reqv not parsed, use positional info to get them.
  while (parg = get_first_unparsed(args)) {
    bool found=false;
    for (int i = 1; i < nargv; i++) {
      if (!pflags[i]) {
	arg_assign_value(parg, argv[i]);
	DEBUG << "parsed argv[" << i << "] by position, as: " << parg->name << "=" << argv[i];
	pflags[i] = true;
	found=true;
	break;
      }
    }
    if (!found) break;
  }

  // see if all required arguments were parsed
  parg = args;
  while (parg->name != NULL) {
    if (parg->required && !parg->parsed) return false;
    parg++;
  }

  // see if all provided arguments were parsed
  for (int i = 1; i < nargv; i++)
    if (!pflags[i]) {
      DEBUG << "argv[" << i << "] was not parsed as anything.";
      return false;
    }
  

  return true;
}


#if 0 // debug code
char *a="<null>";
int b=0;
int c=1;
int d=2;
int e=3;

argument_table_element argument_table[] = {
  { ARG_TYPE_STRING, "a", &a, true, false, ""},
  { ARG_TYPE_INTEGER, "b", &b, true, false, ""},
  { ARG_TYPE_INTEGER, "c", &c, false, false, ""},
  { ARG_TYPE_INTEGER, "d", &d, false, false, ""},
  { ARG_TYPE_INTEGER, "e", &e, false, false, ""},
  END_OF_TABLE
};


int main(int argc, char **argv) {
  log_level = 0;
  bool ok = arguments_parse(argv, argc, argument_table);

  cerr << "ok=" << ok << endl;

  cerr << a << endl;
  cerr << b << endl;
  cerr << c << endl;
  cerr << d << endl;
  cerr << e << endl;

  arguments_print_usage(argv[0], argument_table);
}
#endif
