#ifndef __UTIL_H__ 
#define __UTIL_H__

#include <stdio.h>
#include <string.h>
#include <vector>
#include <string>
#include <sstream>
#include <iostream>
#include <sys/time.h>
using namespace std;

//
// debug
//
extern int log_level;
extern bool log_trace;
extern bool log_thread;
extern bool log_timestamp;

class hexstr {
private:
  char buffer[33];
public:
  hexstr(char x) { snprintf(buffer, 33, "0x%02x", ((unsigned int)x &0xff)); }
  hexstr(short x) { snprintf(buffer, 33, "0x%04x", ((unsigned int)x &0xffff)); }
  hexstr(int x) { snprintf(buffer, 33, "0x%08x", x); }
  hexstr(long x) { snprintf(buffer, 33, "0x%08lx", x);; }
  hexstr(long long x) { snprintf(buffer, 33, "0x%016llx", x); }
  
  template<typename T>
  hexstr(T* ptr) { 
    long x = (long)ptr;
    if (sizeof (ptr) == 4) { // 32-bit mode
      snprintf(buffer, 33, "ptr:0x%08x", (int)(x)); 
    } else { // 64-bit mode (pointers have 48-bits of info)
      snprintf(buffer, 33, "ptr:0x%012lx", (long)(x)); 
    }
  }
    
  string str() const { return string(buffer); }
};

string fmtstr(const char* fmt, ...) __attribute__ ((format (printf, 1, 2)));

class log_msg {
private:
  stringstream msg;
  stringstream post;
public:
  log_msg(const char* event_id, const char* file, int line, const char* function);

  template<typename T>
  log_msg& operator << (const T& t) { msg << t; return *this; }

  log_msg& operator << (const hexstr& h) { msg << h.str(); return *this; }

  string str() { return msg.str() + post.str(); }
};

inline bool operator ^(ostream& out, log_msg& msg) {
  out << msg.str() << endl << flush;
}

#define INIT_LOG_LEVEL 1
#define INIT_LOG_TRACE true
#define INIT_LOG_THREAD false
#define INIT_LOG_TIMESTAMP false

#define LOG(event, level) if (log_level <= level) cerr ^ log_msg(event, __FILE__, __LINE__, __FUNCTION__)

#define DEBUG LOG("debug", 0)
#define INFO LOG("info", 1)
#define WARN LOG("warning", 2)
#define ERROR LOG("error", 3)

/// a timestamp class with microsecond precision
class utimestamp {
private:
  timeval tv;  
public:
  static utimestamp now() {
    utimestamp x;
    gettimeofday(&x.tv, NULL);
    return x;
  }  

  utimestamp() { }

  /// returns microsecond difference
  long operator-(const utimestamp& other) {
    long delta_usec = tv.tv_usec - other.tv.tv_usec;
    long delta_sec = tv.tv_sec - other.tv.tv_sec;
    return delta_sec * 1000000 + delta_usec;
  }
};

//
// misc
//

/** Splits a string into a string array on space boundaries.
    The array is terminated with a NULL pointer.
    You'll need to call free_words on the returned array. */
char** split_words(string& src);
int count_words(char** warray); ///< counts num of elements before NULL.
void free_words(char** words);

/// parses lists of the form "a,b,c,d" into vectors
vector<string> parse_list(const char* liststr, char separator=',');

inline void tty_bold_on(FILE* tty=stderr) { fprintf(tty,"\033[1;1m"); } ///< makes printf print in bold/bright
inline void tty_bold_off(FILE* tty=stderr) { fprintf(tty,"\033[0m"); }  ///< stops printf from printing in bold/bright

//
// spawning processes
//

/**
 * Fork and then exec the given command. Optionally, redirect the stdout and stderr
 * of the child process to the given file descriptor.
 *
 * @return process id > 0 on success, or < 0 on error.
 */
int spawn(char* command, char *argv[], char *envp[], int stdout_redirect_fd=0);

int spawn(string command);

//
// command line arguments
//

enum argument_type { 
  ARG_TYPE_NONE,
  ARG_TYPE_INTEGER,
  ARG_TYPE_DOUBLE,
  ARG_TYPE_BOOLEAN,
  ARG_TYPE_STRING
};

struct argument_table_element {
  argument_type type;
  char*         name;
  void*         var;
  bool          required;
  bool          parsed;
  char*         help;
};

extern argument_table_element END_OF_TABLE;

/**
 * Parses argv and writes any command line arguments to the right
 * place in the given argument table. The table must be terminated
 * by a END_OF_TABLE element.
 *
 * Returns false if not all required arguments were present in argv;
 * or true otherwise.
 */
bool arguments_parse(char** argv, int nargv, argument_table_element args[]);

void arguments_print_usage(const char* progname, argument_table_element args[], FILE* tty=stderr);
void arguments_print_help(const char* progname, argument_table_element args[], FILE* tty=stderr);

#endif
