#include <stdlib.h>
#include <time.h>
#include <string>
#include <nmstl/ioevent>
#include <nmstl/net>

#include "string_hashmap.h"
#include "util.h"
#include "tribe_msg.h"
#include "tunnel.h"
#include "tun_manager.h"

using namespace std;
using namespace nmstl;

void event_accepted(tribe_message_iohandler* src);
void event_connected(tribe_message_iohandler* src);
void event_disconnected(tribe_message_iohandler* src);
void event_recv_msg(tribe_message_iohandler* src, string msg);

void event_recv_tunnel(dynbuf& packet, inet_address& tunnel_src);
void event_recv_tundev(dynbuf& packet, inet_address& tunnel_dest);

//
// context for each follower
//
struct follower_context {
  bool connected;
  tribe_message_iohandler *msg_ioh;
  STRING_HASHMAP(tun_device_iohandler*) devices;

  follower_context(tribe_message_iohandler *m) : connected(false), msg_ioh(m), devices() {}
};

//
// global variables
//

io_event_loop io_loop;

inet_address bind_address;

inet_address msg_bind_address;
inet_address tunnel_bind_address;

tribe_message_acceptor  *msg_ioh = NULL;
tunnel_iohandler        *tunnel_ioh = NULL;

typedef map<tribe_message_iohandler*, follower_context> followers_t;
followers_t followers;


// command line stuff

char *bind_addr = NULL;
int tunnel_port = TUNNEL_DEFAULT_PORT;
int message_port = MESSAGE_DEFAULT_PORT;

//
// command line arguments
//
bool help_requested = false;

argument_table_element argument_table[] = {
  { ARG_TYPE_BOOLEAN, "help", &help_requested, false, false, "Print out this help." },

  { ARG_TYPE_STRING, "bind", &bind_addr, true, false, 
    "The local IP address to bind to."},

  { ARG_TYPE_INTEGER, "tunnel", &tunnel_port, false, false, 
    "Port used for IP-in-UDP tunneling between leaders and followers." },

  { ARG_TYPE_INTEGER, "message", &message_port, false, false, 
    "Port used for message passing between leaders and followers."},

  END_OF_TABLE
};

//
// utility functions
//

void startup() {
  // start message-protocol 
  msg_ioh = new tribe_message_acceptor(msg_bind_address, event_accepted,
				       event_recv_msg, event_connected, event_disconnected);

  // start tunnel without a default sendto address
  tunnel_ioh = new tunnel_iohandler(tunnel_bind_address, inet_address(), event_recv_tunnel);
}

void remote_iface_up(tribe_message_iohandler* src, string ifname, string ifaddr) {
  followers_t::iterator i = followers.find(src);
  if (i != followers.end()) {
    if (i->second.devices.find(ifname) == i->second.devices.end()) {
      // up
      string name = src->get_peer_host() + "." + ifname;
      inet_address tunnel_dest(src->get_peer_host(), tunnel_port);
      tun_device_iohandler* devhandler = create_dev_handler(name, ifaddr, tunnel_dest, event_recv_tundev);
      if (devhandler != NULL) i->second.devices[ifname] = devhandler;
    } else {
      WARN << "remote iface up message, but interface is already up '" << ifname << "'";
    }
  } else {
    WARN << "Probably a bug.";
  }
}

void remote_iface_down(tribe_message_iohandler* src, string ifname) {
  followers_t::iterator i = followers.find(src);
  if (i != followers.end()) {
    STRING_HASHMAP(tun_device_iohandler*)::iterator dev = i->second.devices.find(ifname);
    if (dev != i->second.devices.end()) {
      // delete from table
      tun_device_iohandler* ioh = dev->second;
      i->second.devices.erase(dev);

      // down
      destroy_dev_handler(ioh);      
    } else {
      WARN << "remote iface down message, but bad interface name '" << ifname << "'";
    }
  } else {
    WARN << "Probably a bug.";
  }
}

//
// event handlers
//

void event_accepted(tribe_message_iohandler* src) {

  if (followers.find(src) == followers.end()) {
    followers.insert(followers_t::value_type(src, follower_context(src)));
  } else {
    WARN << "Accepted a follower that already exists. This may be a bug.";
  } 
}

void event_connected(tribe_message_iohandler* src) {
  followers_t::iterator i = followers.find(src);
  if (i != followers.end()) {
    i->second.connected = true;
  } else {
    WARN << "Probably a bug.";
  }
}

void event_disconnected(tribe_message_iohandler* src) {
  followers_t::iterator i = followers.find(src);
  if (i != followers.end()) {
    for (STRING_HASHMAP(tun_device_iohandler*)::iterator dev = i->second.devices.begin(); 
	 dev != i->second.devices.end();
	 dev++) {

      destroy_dev_handler(dev->second);
    }

    string peer = src->get_peer_host();

    INFO << "peer disconnected and all proxy-devices destroyed: '" << peer << "'";

    followers.erase(i);
    delete src;    
  } else {
    WARN << "Probably a bug.";
  }
}

void event_recv_msg(tribe_message_iohandler* src, string msg) {
  // respond to ifup/ifdown messages which are of the following form:
  // "+up eth0 128.30.76.30"
  // "-down eth0 128.30.76.30"

  char** el = split_words(msg);
  if (count_words(el) == 3) {
    string com = el[0];
    string ifname = el[1];
    string ifaddr = el[2];
    free_words(el);
    
    if (com == TRIBE_MESSAGE_IFACE_UP) {
      remote_iface_up(src, ifname, ifaddr);
      return;
    } else if (com == TRIBE_MESSAGE_IFACE_DOWN) {
      remote_iface_down(src, ifname);
      return;
    }
    
  } else {
    free_words(el);
  }

  // if we reached here, bad message
  WARN << "Bad message from " << src->get_peer_host() << " '" << msg << "'";
}

// called when a packet was received at a tunnel
void event_recv_tunnel(dynbuf& packet, inet_address& tunnel_src) {
  // route to the correct tun interface
  route_to_tun_device(packet);
}

// called when a packet was received at a tun device
void event_recv_tundev(dynbuf& packet, inet_address& tunnel_dest) {
  // send out over tunnel
  tunnel_ioh->tunnel_send(packet, tunnel_dest);
}


//
// main
//
int main (int argc, char** argv) {
  bool ok = arguments_parse(argv, argc, argument_table);

  if (help_requested) { // detailed help was requested
    arguments_print_help(argv[0], argument_table, true);    
  } else if (!ok) { // not all requirements were present
    arguments_print_usage(argv[0], argument_table);
  } else {
    // parse command line
    bind_address = inet_address(bind_addr);

    msg_bind_address = inet_address(bind_address.get_host(), message_port);
    tunnel_bind_address = inet_address(bind_address.get_host(), tunnel_port);

    INFO << "binding message protocol to: " << msg_bind_address;
    INFO << "binding tunnel protocol to: " << tunnel_bind_address;

    // startup
    startup();
    io_loop.run();
    WARN << "I/O loop exited.";
  }
}
