/*
 * linux TUN device utility functions.
 *
 * Some of the code is taken from the linux kernel 
 * documentation on the TUN/TAP network driver. 
 *
 */
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <linux/if.h>
#include <linux/if_tun.h>

//
// INTERNAL FUNCTIONS
//

// Creates a new TUN device and returns a file descriptor for
// the TUN interface.
//
// return: valid file descriptor; or <0 on error.
// param: dev -- on success, filled in with device name (e.g., 'tun0') 
//               must be large enough to hold the name.
int tun_create(char *dev) {
  struct ifreq ifr;
  int fd, err;
  
  if( (fd = open("/dev/net/tun", O_RDWR)) < 0 ) {
    fprintf(stderr, "error: couldn't open /dev/net/tun\n");
    return fd; 
  }

  memset(&ifr, 0, sizeof(ifr));
  
  /* Flags: IFF_TUN   - TUN device (no Ethernet headers) 
   *        IFF_TAP   - TAP device  
   *
   *        IFF_NO_PI - Do not provide packet information  
   */ 
  ifr.ifr_flags = IFF_TUN; 
  if( *dev )
    strncpy(ifr.ifr_name, dev, IFNAMSIZ);
  
  if( (err = ioctl(fd, TUNSETIFF, (void *) &ifr)) < 0 ) {
    close(fd);
    return err;
  }
  strcpy(dev, ifr.ifr_name);
  return fd;
}    

// Destroys a TUN device, using the given file descriptor.
int tun_destroy(const char* dev, int fd) {
  return close(fd);
}

// Brings up a TUN device that has been created.
// runs a number of shell commands to configure iproute etc
void tun_device_up(const char* dev, const char* ip_address, int mtu) {
  char buf[1024];
  size_t bufsz = 1024;
  int ret = 0;

  // flag device as being up and configure it
  snprintf(buf, bufsz, "/sbin/ifconfig %s %s", dev, ip_address);
  ret = system(buf);

  snprintf(buf, bufsz, "/sbin/ifconfig %s mtu %d", dev, mtu);
  ret = system(buf);

  snprintf(buf, bufsz, "/sbin/ifconfig %s up", dev);
  ret = system(buf);

  if (ret < 0) {
    printf("WARNING: Failed to execute ifconfig for %s.\n", dev);
    return;
  }

  // Enable iproute2: add outgoing routes through the interface
  // requires that /etc/iproute2/rt_tables has a table defined for the device
  snprintf(buf, bufsz, "/sbin/ip ro add default scope global dev %s table %s", dev, dev);
  ret = system(buf);

  snprintf(buf, bufsz, "/sbin/ip ru add from %s/32 table %s", ip_address, dev);
  ret = system(buf);

  if (ret < 0) {
    printf("WARNING: Failed to set up iproute for %s. Make sure /etc/iproute2/rt_tables has a table defined for %s.\n", dev, dev);
    return;    
  }
  
  snprintf(buf, bufsz, "/sbin/ip ro flush cache"); // flush routing cache
  ret = system(buf);
}

// 
// EXPORTED FUNCTIONS
//

// Create a new tun device, set the ip address, and return
// the file descriptor for this new device. dev is filled in
// with the device name; should be big enough.
int tun_device_activate(char* dev, const char* ip_address, int mtu) {
  int fd = tun_create(dev);
  
  if (fd >= 0)
    tun_device_up(dev, ip_address, mtu);
  
  return fd;
}

// Deactive the given tun device.
int tun_device_deactivate(const char* dev, int fd) {
  tun_destroy(dev,fd);
}

/* --test code--
int main (char** argv, int argc) {
  char dev[128] = "tun0\0";
  int fd = tun_device_activate(dev, "192.168.1.100");

  printf("fd = %d\n", fd);
  if (fd >= 0) {
    printf("devname = %s\n", dev);
    //getchar();

    for (;;) {
      char buf[2048];
      ssize_t n = read(fd, buf, 2048);    
      printf("n=%d\n", n);
      if (n > 0) 
	cout << nmstl::raw_iohandler::packet_header_to_string(buf, n) << endl;
    }

    tun_device_deactivate(dev, fd);
  }  
}
*/
